/*
 * $Id: algo_impl.cpp 498 2007-03-06 08:44:28Z satoken $
 *
 * PHMMTS -- an implementation of Pair Hidden Markov Models on Tree Structures,
 *           which is based on "Pair Hidden Markov Models on Tree Structures",
 *           Yasubumi Sakakibara, ISMB 2003.
 *
 * Copyright (C) 2003-2006 Kengo Sato
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifndef __TMPL_ALGO_IMPL_CPP__
#define __TMPL_ALGO_IMPL_CPP__

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <iomanip>
#include <vector>
#include <algorithm>
#include <cassert>
#include "algo_impl.h"
#include "log_value.h"

using namespace PHMMTS;

template <class CostTable_, template<class> class DPTable_>
AlgorithmImpl<CostTable_,DPTable_>::
AlgorithmImpl(const CostTable_& cost_table, int local_alignment,
	      int max_length /*=0*/)
  : cost_table_(cost_table), local_alignment_(local_alignment),
    max_length_(max_length), EPS(zero<cost_type>())
{
}

/*************************************************************************/

template <class CostTable_, template<class> class DPTable_>
template <class UpdateFunc, class STree, class Sequence>
void
AlgorithmImpl<CostTable_,DPTable_>::
inside_empty(const STree& tree, int height, 
	     const Sequence& rna, int from, int to,
	     UpdateFunc& update) const
{
  // empty nodes are matched to the only bottom cells.
  cost_type emit = static_cast<cost_type>(1.0);
  for (int i=from; i<=to; ++i) {
    for (uint s=0; s<tree[height].size(); ++s) {
      update(EMPTY, emit, Position(height, s, i, i), tree, rna);
    }
  }
}

template <class CostTable_, template<class> class DPTable_>
template <class UpdateFunc, class STree, class Sequence>
void
AlgorithmImpl<CostTable_,DPTable_>::
inside_normal(const STree& tree, int height,
	      const Sequence& rna, int from, int to,
	      UpdateFunc& update) const
{
  int max_len = max_length_>0 ? std::min(to-from, max_length_) : to-from;
  bool unfolded_lstart = local_alignment_ & UNFOLDED_LSTART;
  bool folded_lstart =
    local_alignment_ & FOLDED_LSTART || height==static_cast<int>(tree.size())-1;
  for (int j=from; j<to; ++j) {
    for (int i=j; i>=from && j-i<max_len; --i) {
      // for each substring
      for (uint s=0; s<tree[height].size(); ++s) {
	// for each parent state
	process_normal_state(tree, height, s, rna, i, j, update);
	assert(tree[height].property(s).rchildren.empty());
      }
      if (folded_lstart && (unfolded_lstart || i==from && j==to-1))
	process_local_start(tree, height, rna, i, j, update);
    }
  }
}

template <class CostTable_, template<class> class DPTable_>
template <class UpdateFunc, class STree, class Sequence>
void
AlgorithmImpl<CostTable_,DPTable_>::
inside_branch(const STree& tree, int height,
	      const Sequence& rna, int from, int to,
	      UpdateFunc& update) const
{
  int max_len = max_length_>0 ? std::min(to-from, max_length_) : to-from;
  bool unfolded_lstart = local_alignment_ & UNFOLDED_LSTART;
  bool folded_lstart =
    local_alignment_ & FOLDED_LSTART || height==static_cast<int>(tree.size())-1;
  for (int j=from; j<to; ++j) {
    for (int i=j; i>=from && j-i<max_len; --i) {
      // for each substring
      for (uint s=0; s<tree[height].size(); ++s) {
	// for each parent state
	if (tree[height].property(s).rchildren.empty()) {
	  process_normal_state(tree, height, s, rna, i, j, update);
	} else {
	  process_branch_state(tree, height, s, rna, i, j, update);
	}
      }
      if (folded_lstart && (unfolded_lstart || i==from && j==to-1))
	process_local_start(tree, height, rna, i, j, update);
    }
  }
}

/************************************************************************/

template <class CostTable_, template <class> class DPTable_>
template <class UpdateFunc, class STree, class Sequence>
void
AlgorithmImpl<CostTable_,DPTable_>::
outside_empty(const STree& tree, int height,
	      const Sequence& rna, int from, int to,
	      UpdateFunc& update) const
{
#if 0
  int max_len = max_length_>0 ? std::min(to-from, max_length_) : to-from;
  for (int j=to-1; j>=from; --j) {
    for (int i=from; i<=j && j-i<max_len; i++) {

    }
  }
#endif
}

template <class CostTable_, template <class> class DPTable_>
template <class UpdateFunc, class STree, class Sequence>
void
AlgorithmImpl<CostTable_,DPTable_>::
outside_normal(const STree& tree, int height,
	       const Sequence& rna, int from, int to,
	       UpdateFunc& update) const
{
  int max_len = max_length_>0 ? std::min(to-from, max_length_) : to-from;
  bool unfolded_lstart = local_alignment_ & UNFOLDED_LSTART;
  bool folded_lstart =
    local_alignment_ & FOLDED_LSTART || height==static_cast<int>(tree.size())-1;
  for (int j=to-1; j>=from; --j) {
    for (int i=from; i<=j && j-i<max_len; i++) {
      /// for each substring
      for (int s=tree[height].size()-1; s>=0; --s) {
	// for each parent state
	process_normal_state(tree, height, s, rna, i, j, update);
	assert(tree[height].property(s).rchildren.empty());
      }
      if (folded_lstart && (unfolded_lstart || i==from && j==to-1))
	process_local_start(tree, height, rna, i, j, update);
    }
  }
}

template <class CostTable_, template <class> class DPTable_>
template <class UpdateFunc, class STree, class Sequence>
void
AlgorithmImpl<CostTable_,DPTable_>::
outside_branch(const STree& tree, int height,
	       const Sequence& rna, int from, int to,
	       UpdateFunc& update) const
{
  int max_len = max_length_>0 ? std::min(to-from, max_length_) : to-from;
  bool unfolded_lstart = local_alignment_ & UNFOLDED_LSTART;
  bool folded_lstart =
    local_alignment_ & FOLDED_LSTART || height==static_cast<int>(tree.size())-1;
  for (int j=to-1; j>=from; --j) {
    for (int i=from; i<=j && j-i<max_len; i++) {
      // for each substring
      for (int s=tree[height].size()-1; s>=0; --s) {
	// for each parent state
	if (tree[height].property(s).rchildren.empty()) {
	  process_normal_state(tree, height, s, rna, i, j, update);
	} else {
	  process_branch_state(tree, height, s, rna, i, j, update);
	}
      }
      if (folded_lstart && (unfolded_lstart || i==from && j==to-1))
	process_local_start(tree, height, rna, i, j, update);
    }
  }
}

/*************************************************************************/

template <class CostTable_, template<class> class DPTable_>
template <class UpdateFunc, class STree, class Sequence>
void
AlgorithmImpl<CostTable_,DPTable_>::
scan_empty(const STree& tree, int height, 
	   const Sequence& rna, int from, int to, int pos,
	   UpdateFunc& update) const
{
  // empty nodes are matched to the only bottom cells.
  cost_type emit = static_cast<cost_type>(1.0);
  for (int i=pos; i<=to; ++i) {
    for (uint s=0; s<tree[height].size(); ++s) {
      update(EMPTY, emit, Position(height, s, i, i), tree, rna);
    }
  }
}

template <class CostTable_, template<class> class DPTable_>
template <class UpdateFunc, class STree, class Sequence>
void
AlgorithmImpl<CostTable_,DPTable_>::
scan_normal(const STree& tree, int height,
	    const Sequence& rna, int from, int to, int pos,
	    UpdateFunc& update) const
{
  bool unfolded_lstart = local_alignment_ & UNFOLDED_LSTART;
  bool folded_lstart =
    local_alignment_ & FOLDED_LSTART || height==static_cast<int>(tree.size())-1;
  for (int j=pos; j<to; ++j) {
    for (int i=j; i>=from && j-i<max_length_; --i) {
      for (uint s=0; s<tree[height].size(); ++s) {
	process_normal_state(tree, height, s, rna, i, j, update);
	assert(tree[height].property(s).rchildren.empty());
      }
      if (folded_lstart && (unfolded_lstart || i==from && j==to-1))
	process_local_start(tree, height, rna, i, j, update);
    }
  }
}

template <class CostTable_, template<class> class DPTable_>
template <class UpdateFunc, class STree, class Sequence>
void
AlgorithmImpl<CostTable_,DPTable_>::
scan_branch(const STree& tree, int height,
	    const Sequence& rna, int from, int to, int pos,
	    UpdateFunc& update) const
{
  bool unfolded_lstart = local_alignment_ & UNFOLDED_LSTART;
  bool folded_lstart =
    local_alignment_ & FOLDED_LSTART || height==static_cast<int>(tree.size())-1;
  for (int j=pos; j<to; ++j) {
    for (int i=j; i>=from && j-i<max_length_; --i) {
      for (uint s=0; s<tree[height].size(); ++s) {
	if (tree[height].property(s).rchildren.empty()) {
	  process_normal_state(tree, height, s, rna, i, j, update);
	} else {
	  process_branch_state(tree, height, s, rna, i, j, update);
	}
      }
      if (folded_lstart && (unfolded_lstart || i==from && j==to-1))
	process_local_start(tree, height, rna, i, j, update);
    }
  }
}

/*******************************************************************/

template <class CostTable_, template<class> class DPTable_>
template <class UpdateFunc, class STree, class Sequence>
void
AlgorithmImpl<CostTable_,DPTable_>::
process_normal_state(const STree& tree, int height, int s,
		     const Sequence& rna, int i, int j, 
		     UpdateFunc& update) const
{
  typedef typename Sequence::value_type rna_t;
  typedef typename STree::node_type::Property Property;
  int child = tree[height].left();
  assert(child>=0);
  assert(child<height);
  const rna_t GAP = static_cast<rna_t>(RNASymbol<rna_t>::GAP);
  const Property& st = tree[height].property(s);

  int ii = i+st.u_left;
  int jj = j-st.u_right;
  if (ii+st.loop>jj+1) return;

  cost_type emit = static_cast<cost_type>(1.0);
  if (st.f_left!=0 || st.u_left!=0 || st.f_right!=0 || st.u_right!=0)
    emit =
      cost_table_.emit_cost(st.f_left ? tree[height].left_rna() : GAP,
			    st.f_right ? tree[height].right_rna() : GAP,
			    st.u_left ? rna[i] : GAP,
			    st.u_right ? rna[j] : GAP);
  
  if (emit>EPS) {
    typename Property::trans_list_t::const_iterator t;
    for (t=st.children.begin(); t!=st.children.end(); ++t) {
      // for each child state
      int ch = t->first;
      assert(ch==child || ch==height);
      const Property& tt = tree[ch].property(t->second);
      cost_type state = cost_table_.state_cost(st.state, tt.state);
      if (state>EPS) {
	update(tree[height].type(), emit, state,
	       Position(height, s, i, j+1),
	       Position(ch, t->second, ii, jj+1), tree, rna);
      }
    }

    // L_END for local alignments
    if (st.local_end &&
	(((local_alignment_ & FOLDED_LEND) && tree[child].type()!=EMPTY) ||
	 (local_alignment_ & UNFOLDED_LEND) && ii < jj+1)) {
      cost_type state = cost_table_.state_cost(st.state, L_END);
      if (state>EPS) {
	update(tree[height].type(), emit, state,
	       Position(height, s, i, j+1),
	       Position(child, -1, ii, jj+1), tree, rna);
      }
    }
  }
}

template <class CostTable_, template<class> class DPTable_>
template <class UpdateFunc, class STree, class Sequence>
void
AlgorithmImpl<CostTable_,DPTable_>::
process_local_start(const STree& tree, int height,
		    const Sequence& rna, int i, int j, 
		    UpdateFunc& update) const
{
  typedef typename Sequence::value_type rna_t;
  typedef typename STree::node_type::Property Property;
  int ch = tree[height].left();
  // L_START for local alignments
  for (uint t=0; t<tree[ch].size(); ++t) {
    // for each child state
    const Property& tt = tree[ch].property(t);
    if (tt.local_start) {
      cost_type state = cost_table_.state_cost(L_START, tt.state);
      if (state>EPS) {
	update(START, static_cast<cost_type>(1.0), state,
	       Position(height, -1, i, j+1),
	       Position(ch, t, i, j+1), tree, rna);
      }
    }
  }
}

template <class CostTable_, template<class> class DPTable_>
template <class UpdateFunc, class STree, class Sequence>
void
AlgorithmImpl<CostTable_,DPTable_>::
process_branch_state(const STree& tree, int height, int s,
		     const Sequence& rna, int i, int j, 
		     UpdateFunc& update) const
{
  typedef typename STree::node_type::Property Property;
#ifndef NDEBUG
  int lchild = tree[height].left();
  int rchild = tree[height].right();
#endif
  assert(lchild>=0);
  assert(lchild<height);
  assert(rchild>=0);
  assert(rchild<height);
  const Property& st = tree[height].property(s);
  
  cost_type emit = static_cast<cost_type>(1.0);
  typename Property::trans_list_t::const_iterator t,u;
  for (t=st.children.begin(); t!=st.children.end(); ++t) {
    int lch = t->first;
    const Property& tt=tree[lch].property(t->second);
    assert(lch==lchild || lch==height);
    for (u=st.rchildren.begin(); u!=st.rchildren.end(); ++u) {
      int rch = u->first;
      const Property& ut=tree[rch].property(u->second);
      cost_type state = cost_table_.state_cost(st.state, tt.state, ut.state);
      if (state>EPS) {
	for (int l=i+1; l<=j; ++l) {
	  update(BRANCH, emit, state,
		 Position(height, s, i, j+1),
		 Position(lch, t->second, i, l),
		 Position(rch, u->second, l, j+1), tree, rna);
	}
      }
    }
  }
}

#endif // __TMPL_ALGO_IMPL_CPP__
