/*
 * $Id: algo_impl.h 482 2006-04-12 09:00:23Z satoken $
 *
 * PHMMTS -- an implementation of Pair Hidden Markov Models on Tree Structures,
 *           which is based on "Pair Hidden Markov Models on Tree Structures",
 *           Yasubumi Sakakibara, ISMB 2003.
 *
 * Copyright (C) 2003-2006 Kengo Sato
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifndef __INC_ALGO_IMPL_H__
#define __INC_ALGO_IMPL_H__

#include <list>
#include <vector>
#include "defines.h"

namespace PHMMTS {
  template <class> class CostTable;
  template <class> class DPTable;

  template <class CostTable_, template <class> class DPTable_ = DPTable>
  class AlgorithmImpl
  {
    typedef typename CostTable_::cost_type cost_type;

    const CostTable_& cost_table_;
    int local_alignment_;
    int max_length_;

    cost_type EPS;

  public:
    enum {
      FOLDED_LSTART = 1<<0,
      FOLDED_LEND = 1<<1,
      UNFOLDED_LSTART = 1<<2,
      UNFOLDED_LEND = 1<<3
    };

  public:
    AlgorithmImpl(const CostTable_& cost_table, int enable_local_alignment,
		  int max_length=0);

    int max_length() const { return max_length_; }
    
    const CostTable_& cost_table() const { return cost_table_; }
    
    // inside algorithm
    template <class UpdateFunc, class STree, class Sequence>
    inline
    void inside_empty(const STree& tree, int height,
		      const Sequence& rna, int from, int to,
		      UpdateFunc& update) const;

    template <class UpdateFunc, class STree, class Sequence>
    inline
    void inside_normal(const STree& tree, int height,
		       const Sequence& rna, int from, int to,
		       UpdateFunc& update) const;

    template <class UpdateFunc, class STree, class Sequence>
    inline
    void inside_branch(const STree& tree, int height,
		       const Sequence& rna, int from, int to,
		       UpdateFunc& update) const;

    // outside algorithm
    template <class UpdateFunc, class STree, class Sequence>
    inline
    void outside_empty(const STree& tree, int height,
		       const Sequence& rna, int from, int to,
		       UpdateFunc& update) const;

    template <class UpdateFunc, class STree, class Sequence>
    inline
    void outside_normal(const STree& tree, int height,
			const Sequence& rna, int from, int to,
			UpdateFunc& update) const;

    template <class UpdateFunc, class STree, class Sequence>
    inline
    void outside_branch(const STree& tree, int height,
			const Sequence& rna, int from, int to,
			UpdateFunc& update) const;

    // inside algorithm for scanning
    template <class UpdateFunc, class STree, class Sequence>
    inline
    void scan_empty(const STree& tree, int height,
		    const Sequence& rna, int from, int to, int pos,
		    UpdateFunc& update) const;

    template <class UpdateFunc, class STree, class Sequence>
    inline
    void scan_normal(const STree& tree, int height,
		     const Sequence& rna, int from, int to, int pos,
		     UpdateFunc& update) const;

    template <class UpdateFunc, class STree, class Sequence>
    inline
    void scan_branch(const STree& tree, int height,
		     const Sequence& rna, int from, int to, int pos, 
		     UpdateFunc& update) const;

  private:
    template <class UpdateFunc, class STree, class Sequence>
    inline
    void process_normal_state(const STree& tree, int height, int state,
			      const Sequence& rna, int i, int j, 
			      UpdateFunc& update) const;

    template <class UpdateFunc, class STree, class Sequence>
    inline
    void process_branch_state(const STree& tree, int height, int state,
			      const Sequence& rna, int i, int j, 
			      UpdateFunc& update) const;

    template <class UpdateFunc, class STree, class Sequence>
    inline
    void process_local_start(const STree& tree, int height,
			     const Sequence& rna, int i, int j, 
			     UpdateFunc& update) const;
  };
};

#include "algo_impl.cpp"

#endif // __INC_ALGO_IMPL_H__

// Local Variables:
// mode: C++
// End:
