/*
 * $Id: background.cpp 482 2006-04-12 09:00:23Z satoken $
 *
 * PHMMTS -- an implementation of Pair Hidden Markov Models on Tree Structures,
 *           which is based on "Pair Hidden Markov Models on Tree Structures",
 *           Yasubumi Sakakibara, ISMB 2003.
 *
 * Copyright (C) 2003-2006 Kengo Sato
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "background.h"

using namespace PHMMTS;

template <class CT>
struct TraverseFunc {
  typedef typename CT::cost_type cost_type;
  std::vector<cost_type>& del_table_;
  const CT& cost_table_;

  TraverseFunc(std::vector<cost_type>& del_table_table,
	       const CT& cost_table)
    : del_table_(del_table_table), cost_table_(cost_table)
  {
  }

  template< class STree >
  void operator()(const STree& tree, int n)
  {
    switch (tree[n].type()) {
    case EMPTY:
      del_table_[n] = static_cast<cost_type>(1.0);
      break;
    case LSINGLE:
      del_table_[n] =
	del_table_[tree[n].left()] *
	cost_table_.emit_cost(tree[n].left_rna());
      break;
    case RSINGLE:
      del_table_[n] =
	del_table_[tree[n].left()] *
	cost_table_.emit_cost(tree[n].right_rna());
      break;
    case STEM:
      del_table_[n] =
	del_table_[tree[n].left()] *
	cost_table_.emit_cost(tree[n].left_rna()) *
	cost_table_.emit_cost(tree[n].right_rna());
      break;
    case START:
      del_table_[n] = del_table_[tree[n].left()];
      break;
    case BRANCH:
      del_table_[n] =
	del_table_[tree[n].left()] *
	del_table_[tree[n].right()];
      break;
    default:
      break;
    }
  }
};

template < class cost_type >
template < class CT, class STree, class Sequence >
void
Background<cost_type>::
initialize(const CT& cost_table,
	   const STree& tree, const Sequence& rna, int from, int to)
{
  from_ = from;
  ins_table_.resize(to-from);
  ins_table_[0] = static_cast<cost_type>(1.0);
  for (size_t i=1; i<ins_table_.size(); ++i) {
    ins_table_[i] =ins_table_[i-1] * cost_table.emit_cost(rna[from+i-1]);
  }
  
  del_table_.resize(tree.size());
  struct TraverseFunc<CT> func(del_table_, cost_table);
  STree::traverse(tree, func);
}

template < class cost_type >
template < class CT, class Sequence >
void
Background<cost_type>::
update_position(const CT& cost_table,
		const Sequence& rna, int from, int to)
{
  from_ = from;
  ins_table_.resize(to-from);
  ins_table_[0] = static_cast<cost_type>(1.0);
  for (size_t i=1; i<ins_table_.size(); ++i) {
    ins_table_[i] =ins_table_[i-1] * cost_table.emit_cost(rna[from+i-1]);
  }
}

template < class cost_type >
cost_type
Background<cost_type>::
inside(const Position& pos) const
{
  return
    // for local gap insertion
    ins_table_[pos.right-1-from_] / ins_table_[pos.left-from_] *
    // for local gap deletion
    del_table_[pos.height];
}

template < class cost_type >
cost_type
Background<cost_type>::
outside(const Position& pos) const
{
  return
    // for local gap insertion
    ins_table_[ins_table_.size()-1] / ins_table_[pos.right-1-from_] *
    ins_table_[pos.left-from_] * 
    // for local gap deletion
    del_table_[del_table_.size()-1] / del_table_[pos.height];
}

// template instantiation for 'cost_t'
#include <string>
#include "costtable.h"
#include "log_value.h"
#include "skeletal.h"

template
class Background<cost_t>;

#if 0
template
void
Background<cost_t>::
initialize(const CostTable<cost_t>&,
	   const SkeletalTree&, const std::string&, int, int);
#endif

template
void
Background<cost_t>::
initialize(const CostTable<cost_t>&,
	   const SkeletalTreeS&, const RNASequence&, int, int);

template
void
Background<cost_t>::
update_position(const CostTable<cost_t>&,
		const std::string&, int, int);

template
void
Background<cost_t>::
update_position(const CostTable<cost_t>&,
		const RNASequence&, int, int);

// template instantiation for 'LogValue<short>'
#include "defaultcosttable.h"

template
class Background< LogValue<short> >;

#if 0
template
void
Background< LogValue<short> >::
initialize(const DefaultCostTable&,
	   const SkeletalTree&, const std::string&, int, int);
#endif

template
void
Background< LogValue<short> >::
initialize(const DefaultCostTable&,
	   const SkeletalTreeS&, const RNASequence&, int, int);

template
void
Background< LogValue<short> >::
update_position(const DefaultCostTable&,
		const std::string&, int, int);

template
void
Background< LogValue<short> >::
update_position(const DefaultCostTable&,
		const RNASequence&, int, int);

