/*
 * $Id: bintree.h 482 2006-04-12 09:00:23Z satoken $
 *
 * PHMMTS -- an implementation of Pair Hidden Markov Models on Tree Structures,
 *           which is based on "Pair Hidden Markov Models on Tree Structures",
 *           Yasubumi Sakakibara, ISMB 2003.
 *
 * Copyright (C) 2003-2006 Kengo Sato
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */


#ifndef __INC_BIN_TREE_H__
#define __INC_BIN_TREE_H__

#include <queue>

#ifndef NULL
#define NULL 0
#endif

/**
 * @class BinTree
 * @brief A class of the binary tree containar.
 */
template <class T>
class BinTree
{
  BinTree<T>* parent_; ///< a pointer to the parent node.
  BinTree<T>* left_; ///< a pointer to the left child node.
  BinTree<T>* right_; ///< a pointer to the right child node.
  T data_; ///< a contained data.

public:
  /**
   * @brief A constructor.
   * @param data a contained data.
   * @param left a pointer to the left child node.
   * @param right a pointer to the right child node.
   * @param 
   */
  BinTree(const T& data, BinTree<T>* left=NULL, BinTree<T>* right=NULL)
    : parent_(NULL), data_(data)
  {
    set_left(left);
    set_right(right);
  }

  /**
   * @return a pointer to the parent node.
   */
  BinTree<T>* parent() const { return parent_; }

  /** 
   * @return a pointer to the left child node.
   */
  BinTree<T>* left() const { return left_; }

  /** 
   * @return a pointer to the right child node.
   */
  BinTree<T>* right() const { return right_; }

  /**
   * @return a contained data.
   */
  T& data() { return data_; }

  /**
   * @return a contained data.
   */
  const T& data() const { return data_; }

  /**
   * @brief set the parent node.
   * @param parent a pointer to the parent node.
   */
  void set_parent(BinTree<T>* parent) { parent_ = parent; }

  /**
   * @brief set the left child node.
   * @param left a pointer to the left child node.
   */
  void set_left(BinTree<T>* left)
  {
    left_ = left;
    if (left!=NULL)
      left_->set_parent(this);
  }

  /**
   * @brief set the right child node.
   * @param left a pointer to the right child node.
   */
  void set_right(BinTree* right)
  {
    right_ = right;
    if (right!=NULL)
      right_->set_parent(this);
  }

  /**
   * @brief set the contained data.
   * @param data a contained data.
   */
  void set_data(const T& data) { data = data_; }

  /**
   * @return true if the node has no child.
   */
  bool is_leaf() const { return left_==NULL && right_==NULL; }

  /**
   * @return true if the node is root node.
   */
  bool is_root() const { return parent_==0; }

  template <class Node, class Func>
  static Func traverse_width_first(Node* root, Func f)
  {
    std::queue<BinTree<T> const*> queue;
    queue.push(root);
    while (!queue.empty()) {
      Node* t = (Node*)queue.front();
      queue.pop();
      if (!f(t)) break;
      if (t->left()!=NULL) queue.push(t->left());
      if (t->right()!=NULL) queue.push(t->right());
    }
    return f;
  }

  template <class Node, class Func>
  static Func traverse_depth_first(Node* root, Func f)
  {
    traverse_depth_first_sub(root, f);
    return f;
  }
  
private:
  template <class Node, class Func>
  static bool traverse_depth_first_sub(Node* node, Func& f)
  {
    if (node->left()!=NULL) {
      if (!traverse_depth_first_sub((Node*)node->left(), f))
	return false;
    }
    if (node->right()!=NULL) {
      if (!traverse_depth_first_sub((Node*)node->right(), f))
	return false;
    }
    return f(node);
  }
};

#endif // __INC_BIN_TREE_H__

// Local Variables:
// mode:C++
// End:
