/*
 * $Id: cell.h 482 2006-04-12 09:00:23Z satoken $
 *
 * PHMMTS -- an implementation of Pair Hidden Markov Models on Tree Structures,
 *           which is based on "Pair Hidden Markov Models on Tree Structures",
 *           Yasubumi Sakakibara, ISMB 2003.
 *
 * Copyright (C) 2003-2006 Kengo Sato
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */


#ifndef __INC_CELL_H__
#define __INC_CELL_H__

#include <cassert>
#include "defines.h"
#include "log_value.h"

namespace PHMMTS {
  
  template <class cost_type_>
  class CYKCell
  {
  public:
    typedef cost_type_ cost_type;

  private:
    cost_type cost_;
    int state_ : 8;
    bool self_trans_ : 8;

  public:
    CYKCell()
      : cost_(zero<cost_type>()), state_(-1), self_trans_(false) { }

    CYKCell(const CYKCell& n)
      : cost_(n.cost_), state_(n.state_), self_trans_(n.self_trans_) { }

    CYKCell(int state, bool self_trans, const cost_type& cost)
      : cost_(cost), state_(state), self_trans_(self_trans) { }

    CYKCell& operator=(const CYKCell& n)
    {
      if (this!=&n) {
	cost_=n.cost_;
	state_=n.state_;
	self_trans_=n.self_trans_;
      }
      return *this;
    }

    const cost_type& cost() const { return cost_; }
    int state() const { return state_; }
    bool self_trans() const { return self_trans_; }
    bool use() const { return !zerop(cost_); }
    bool operator!() const { return !use(); }

    bool update(int state, bool self_trans, const cost_type& cost)
    {
      if (!use() || cost>cost()) {
	cost_ = cost;
	state_ = state;
	self_trans_ = self_trans;
	return true;
      }
      return false;
    }

    bool update(const CYKCell& cell)
    {
      if (!use() || cell.cost()>cost()) {
	*this = cell;
	return true;
      }
      return false;
    }
  };

  class BrCell
  {
    uint pos_;			/**< the postion of branching point */
    int state_ : 8;		/**< the state id of right side */

  public:
    BrCell() { }
    BrCell(int pos, int state) : pos_(pos), state_(state) { }
    BrCell(const BrCell& n) : pos_(n.pos_), state_(n.state_) { }
    BrCell& operator=(const BrCell& n)
    {
      if (this!=&n) {
	pos_ = n.pos_;
	state_ = n.state_;
      }
      return *this;
    }
    int pos() const { return pos_; }
    int state() const { return state_; }
  };

  template <class cost_type_>
  class CYKnoTraceCell
  {
  public:
    typedef cost_type_ cost_type;
  private:
    cost_type cost_;
  public:
    CYKnoTraceCell() : cost_(static_cast<cost_type>(0.0)) { }
    CYKnoTraceCell(const CYKnoTraceCell& n) : cost_(n.cost_) { }
    CYKnoTraceCell(const cost_type& cost) : cost_(cost) { }

    CYKnoTraceCell& operator=(const CYKnoTraceCell& n)
    {
      if (this!=&n) cost_=n.cost_;
      return *this;
    }
    const cost_type& cost() const { return cost_; }
    bool use() const
    {
      return !zerop(cost_);
    }
    bool update(const cost_type& c)
    {
      if (!use() || cost_ < c) {
	cost_ = c;
	return true;
      }
      return false;
    }
  };

  template <class cost_type_>
  class IOCell
  {
  public:
    typedef cost_type_ cost_type;
  private:
    cost_type cost_;
  public:
    IOCell() : cost_(static_cast<cost_type>(0.0)) { }
    IOCell(const IOCell& n) : cost_(n.cost_) { }
    IOCell(const cost_type& cost) : cost_(cost) { }

    IOCell& operator=(const IOCell& n)
    {
      if (this!=&n) cost_=n.cost_;
      return *this;
    }
    const cost_type& cost() const { return cost_; }
    bool use() const
    {
      return !zerop(cost_);
    }
    bool update(const cost_type& c)
    {
      cost_ += c;
      return true;
    }
  };

}

#endif // __INC_CELL_H__

// Local Variables:
// mode: C++
// End:
