/*
 * $Id: costtable.cpp 490 2006-07-24 08:38:21Z satoken $
 *
 * PHMMTS -- an implementation of Pair Hidden Markov Models on Tree Structures,
 *           which is based on "Pair Hidden Markov Models on Tree Structures",
 *           Yasubumi Sakakibara, ISMB 2003.
 *
 * Copyright (C) 2003-2006 Kengo Sato
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <iostream>
#include <sstream>
#include <algorithm>
#include <vector>
#include <string>
#include <iterator>
#include <cstdlib>
#include <cctype>
#include <cassert>
#include <cmath>
#include "log_value.h"
#include "costtable.h"
#include "rnatree.h"

using namespace PHMMTS;

static
const int
DEFAULT_INDEX_TABLE[26] = {
  // a,  b,  c,     d,  e, f, g,     h,  i, j, k, l, m,  n, 
  RNA_A, 12, RNA_C, 14, 4, 4, RNA_G, 11, 4, 4, 8, 4, 7, 15,
  // o, p, q, r, s, t,     u,     v,  w, x, y, z
     4, 4, 4, 5, 9, RNA_U, RNA_U, 13, 10, 4, 6, 4
};

static
const char
RNA_SYMBOL[] = {
  'A', 'C', 'G', 'U', '-'
};

template <class CostType>
template <class RNA>
CostType
CostTable<CostType>::
emit_cost(const RNA& i, const RNA& j, const RNA& k, const RNA& l) const
{
  return *emit_count_[i][j][k][l];
}

template <class CostType>
CostType
CostTable<CostType>::
emit_cost(char i, char j, char k, char l) const
{
  assert(emit_count_[index(i)][index(j)][index(k)][index(l)]);
  return *emit_count_[index(i)][index(j)][index(k)][index(l)];
}

template <class CostType>
template <class RNA>
CostType
CostTable<CostType>::
emit_cost(const RNA& i) const
{
  assert(l_emit_count_[i]);
  return *l_emit_count_[i];
}

template <class CostType>
CostType
CostTable<CostType>::
emit_cost(char i) const
{
  assert(l_emit_count_[index(i)]);
  return *l_emit_count_[index(i)];
}

template <class CostType>
template <class RNA>
void
CostTable<CostType>::
set_emit_count(const RNA& i, const RNA& j,
	       const RNA& k, const RNA& l, const cost_type& cnt)
{
  assert(emit_count_[i][j][k][l]);
  *emit_count_[i][j][k][l] = cnt;
}

template <class CostType>
void
CostTable<CostType>::
set_emit_count(char i, char j, char k, char l, const cost_type& cnt)
{
  assert(emit_count_[index(i)][index(j)][index(k)][index(l)]);
  *emit_count_[index(i)][index(j)][index(k)][index(l)] = cnt;
}

template <class CostType>
template <class RNA>
void
CostTable<CostType>::
set_emit_count(const RNA& i, const cost_type& cnt)
{
  assert(l_emit_count_);
  *l_emit_count_[i] = cnt;
}

template <class CostType>
void
CostTable<CostType>::
set_emit_count(char i, const cost_type& cnt)
{
  assert(l_emit_count_[index(i)]);
  *l_emit_count_[index(i)] = cnt;
}

template <class CostType>
template <class RNA>
void
CostTable<CostType>::
add_emit_count(const RNA& i, const RNA& j,
	       const RNA& k, const RNA& l, const cost_type& cnt)
{
  assert(emit_count_[i][j][k][l]);
  *emit_count_[i][j][k][l] += cnt;
}

template <class CostType>
void
CostTable<CostType>::
add_emit_count(char i, char j, char k, char l, const cost_type& cnt)
{
  assert(emit_count_[index(i)][index(j)][index(k)][index(l)]);
  *emit_count_[index(i)][index(j)][index(k)][index(l)] += cnt;
}

template <class CostType>
template <class RNA>
void
CostTable<CostType>::
add_emit_count(const RNA& i, const cost_type& cnt)
{
  assert(l_emit_count_[i]);
  *l_emit_count_[i] += cnt;
}

template <class CostType>
void
CostTable<CostType>::
add_emit_count(char i, const cost_type& cnt)
{
  assert(l_emit_count_[index(i)]);
  *l_emit_count_[index(i)] += cnt;
}

template <class CostType>
CostType
CostTable<CostType>::
state_cost(match_state_t from, match_state_t to) const
{
  assert(state_count_[from][to]);
  return *state_count_[from][to];
}

template <class CostType>
CostType
CostTable<CostType>::
state_cost(match_state_t from,
	   match_state_t to_left, match_state_t to_right) const
{
  if (from==MATCH && to_left==MATCH && to_right==MATCH) {
    return branch_count_[0];
  } else {
    return zero_;
  }
}

template <class CostType>
void
CostTable<CostType>::
set_state_count(match_state_t from, match_state_t to, const cost_type& cnt)
{
  assert(state_count_[from][to]);
  *state_count_[from][to] = cnt;
}

template <class CostType>
void
CostTable<CostType>::
add_state_count(match_state_t from, match_state_t to, const cost_type& cnt)
{
  assert(state_count_[from][to]);
  *state_count_[from][to] += cnt;
}

template <class CostType>
void
CostTable<CostType>::
set_state_count(match_state_t from, match_state_t to_left,
		match_state_t to_right, const cost_type& cnt)
{
  assert(from==MATCH);
  assert(to_left==MATCH);
  assert(to_right==MATCH);
  branch_count_[0] = cnt;
}

template <class CostType>
void
CostTable<CostType>::
add_state_count(match_state_t from, match_state_t to_left,
		match_state_t to_right, const cost_type& cnt)
{
  assert(from==MATCH);
  assert(to_left==MATCH);
  assert(to_right==MATCH);
  branch_count_[0] += cnt;
}

template <class CostType>
void
CostTable<CostType>::
store(std::ostream& os, bool logged, const cost_type& eps) const
{
  if (logged) os << "logged" << std::endl;

  // for single emits
  for (int i=0; i<N_RNA-1; ++i) {
    for (int j=i; j<N_RNA-1; ++j) {
      assert(emit_count_[i][RNA_GAP][j][RNA_GAP]);
      if ((logged && log(*emit_count_[i][RNA_GAP][j][RNA_GAP])!=0.0) ||
	  (!logged && *emit_count_[i][RNA_GAP][j][RNA_GAP]>eps)) {
	os << RNA_SYMBOL[i] << ' ' << RNA_SYMBOL[RNA_GAP] << ' '
	   << RNA_SYMBOL[j] << ' ' << RNA_SYMBOL[RNA_GAP] << ' ';
	if (logged)
	  os << log(*emit_count_[i][RNA_GAP][j][RNA_GAP]) << std::endl;
	else
	  os << *emit_count_[i][RNA_GAP][j][RNA_GAP] << std::endl;
      }
    }
    assert(emit_count_[i][RNA_GAP][RNA_GAP][RNA_GAP]);
    if ((logged && log(*emit_count_[i][RNA_GAP][RNA_GAP][RNA_GAP])!=0.0) ||
	(!logged && *emit_count_[i][RNA_GAP][RNA_GAP][RNA_GAP])) {
      os << RNA_SYMBOL[i] << ' ' << RNA_SYMBOL[RNA_GAP] << ' '
	 << RNA_SYMBOL[RNA_GAP] << ' ' << RNA_SYMBOL[RNA_GAP] << ' ';
      if (logged)
	os << log(*emit_count_[i][RNA_GAP][RNA_GAP][RNA_GAP]) << std::endl;
      else
	os << *emit_count_[i][RNA_GAP][RNA_GAP][RNA_GAP] << std::endl;
    }
  }

  // for stem emits
  for (int i=0; i<N_RNA-1; ++i) {
    for (int j=0; j<N_RNA-1; ++j) {
      for (int k=0; k<N_RNA-1; ++k) {
	for (int l=0; l<N_RNA-1; ++l) {
	  if (i>k || i==k && j>l) continue;
	  assert(emit_count_[i][j][k][l]);
	  if ((logged && log(*emit_count_[i][j][k][l])!=0.0) ||
	      (!logged && *emit_count_[i][j][k][l]>eps)) {
	    os //<< "emit "
	      << RNA_SYMBOL[i] << ' ' << RNA_SYMBOL[j] << ' '
	      << RNA_SYMBOL[k] << ' ' << RNA_SYMBOL[l] << ' ';
	    if (logged)
	      os << log(*emit_count_[i][j][k][l]) << std::endl;
	    else
	      os << *emit_count_[i][j][k][l] << std::endl;
	  }
	}
      }
      assert(emit_count_[i][j][RNA_GAP][RNA_GAP]);
      if ((logged && log(*emit_count_[i][j][RNA_GAP][RNA_GAP])!=0.0) ||
	  (!logged && *emit_count_[i][j][RNA_GAP][RNA_GAP]>eps)) {
	os //<< "emit "
	  << RNA_SYMBOL[i] << ' ' << RNA_SYMBOL[j] << ' '
	  << RNA_SYMBOL[RNA_GAP] << ' ' << RNA_SYMBOL[RNA_GAP] << ' ';
	if (logged)
	  os << log(*emit_count_[i][j][RNA_GAP][RNA_GAP]) << std::endl;
	else
	  os << *emit_count_[i][j][RNA_GAP][RNA_GAP] << std::endl;
      }
    }
  }

  // for local emits
  for (int i=0; i<N_RNA-1; ++i) {
    if ((logged && log(*l_emit_count_[i])!=0.0) ||
	(!logged && *l_emit_count_[i]>eps)) {
      if (logged)
	os << RNA_SYMBOL[i] << ' ' << log(*l_emit_count_[i]) << std::endl;
      else 
	os << RNA_SYMBOL[i] << ' ' << *l_emit_count_[i] << std::endl;
    }
  }

  // for state transtions
  for (int s=0; s<n_normal_states; ++s) {
    for (int t=0; t<n_normal_states; ++t) {
      if ((logged && log(*state_count_[s][t])!=0.0) ||
	  (!logged && *state_count_[s][t]>eps)) {
	os //<< "state "
	  << state_symbol(s) << ' ' << state_symbol(t) << ' ';
	if (logged)
	  os << log(*state_count_[s][t]) << std::endl;
	else
	  os << *state_count_[s][t] << std::endl;

      }
    }
  }

  {
    struct {
      const char* str;
      cost_type* cost;
    } cc[] = {
      { "L M", state_count_[L_START][MATCH] },
      { "M L", state_count_[MATCH][L_END] },
      //{ "L L", state_count_[L_START][L_START] },
      { "M M M", &branch_count_[0] },
      { NULL, NULL }
    };

    for (int i=0; cc[i].str!=NULL; ++i) {
      if ((logged && log(*cc[i].cost)!=0.0) ||
	  (!logged && *cc[i].cost>eps)) {
	os << cc[i].str << ' ';
	if (logged)
	  os << log(*cc[i].cost) << std::endl;
	else
	  os << *cc[i].cost << std::endl;
      }
    }
  }
}

template <class CostType>
void
CostTable<CostType>::
restore(std::istream& is)
{
  bool logged=false;
  initialize();

  std::string line;
  while (std::getline(is, line)) {
    if (line.empty() || line[0]=='#') continue;
    std::istringstream s(line);
    std::vector<std::string> v;
    float p;
    CostType val;
    int i,j,k,l;
    std::copy(std::istream_iterator<std::string>(s),
	      std::istream_iterator<std::string>(),
	      std::back_inserter(v));
    switch (v.size()) {
    case 1:			// "logged" flag
      if (v[0]=="logged") {
	logged=true;
	for (size_t i=0; i<size(); ++i)
	  params_[i] = exp(0.0);
      }
      break;
    case 2:			// l_emit
      p = std::atof(v[1].c_str());
      val = logged ? exp(p) : p;
      i = index(v[0][0]);
      *l_emit_count_[i] = val;
      break;
    case 3:			// state or evalue
      if (v[0]=="E-value") {
	has_evalue_parameters_=true;
	evalue_lambda_ = std::atof(v[1].c_str());
	evalue_const_ = std::atof(v[2].c_str());
      } else {
	p = std::atof(v[2].c_str());
	val = logged ? exp(p) : p;
	i = state_index(v[0]);
	j = state_index(v[1]);
	set_state_count((match_state_t)i,(match_state_t)j,val);
      }
      break;
    case 4:			// state (branch)
      p = std::atof(v[3].c_str());
      val = logged ? exp(p) : p;
      i = state_index(v[0]);
      j = state_index(v[1]);
      k = state_index(v[2]);
      set_state_count((match_state_t)i,(match_state_t)j,(match_state_t)k,val);
      break;
    case 5:			// emit
      p = std::atof(v[4].c_str());
      val = logged ? exp(p) : p;
      i = index(v[0][0]);
      j = index(v[1][0]);
      k = index(v[2][0]);
      l = index(v[3][0]);
      *emit_count_[i][j][k][l] = val;
      break;
    default:
      assert(!"parse error");
      break;
    }
  }
}

template <class CostType>
void
CostTable<CostType>::
initialize() 
{
  int x = 0;
  
  for (int i=0; i<N_RNA; ++i)
    for (int j=0; j<N_RNA; ++j)
      for (int k=0; k<N_RNA; ++k)
	for (int l=0; l<N_RNA; ++l)
	  emit_count_[i][j][k][l] = NULL;

  // for single emits
  for (int i=0; i<N_RNA-1; ++i) {
    for (int j=0; j<N_RNA-1; ++j) {
      if (i>j) {
	emit_count_[i][RNA_GAP][j][RNA_GAP]
	  = emit_count_[j][RNA_GAP][i][RNA_GAP];
      } else {
	emit_count_[i][RNA_GAP][j][RNA_GAP] = &params_[x++];
      }
      emit_count_[RNA_GAP][i][RNA_GAP][j]
	= emit_count_[i][RNA_GAP][j][RNA_GAP];
    }
    emit_count_[i][RNA_GAP][RNA_GAP][RNA_GAP] = &params_[x];
    emit_count_[RNA_GAP][i][RNA_GAP][RNA_GAP] = &params_[x];
    emit_count_[RNA_GAP][RNA_GAP][i][RNA_GAP] = &params_[x];
    emit_count_[RNA_GAP][RNA_GAP][RNA_GAP][i] = &params_[x++];
  }
  assert(x==n_single_emits);

  // for stem emits
  for (int i=0; i<N_RNA-1; ++i) {
    for (int j=0; j<N_RNA-1; ++j) {
      for (int k=0; k<N_RNA-1; ++k) {
	for (int l=0; l<N_RNA-1; ++l) {
	  if (i>k || i==k && j>l)
	    emit_count_[i][j][k][l] = emit_count_[k][l][i][j];
	  else
	    emit_count_[i][j][k][l] = &params_[x++];
	}
      }
      emit_count_[i][j][RNA_GAP][RNA_GAP] = &params_[x];
      emit_count_[RNA_GAP][RNA_GAP][i][j] = &params_[x++];
    }
  }
  assert(x==n_single_emits+n_stem_emits);
  
  // for local emits
  for (int i=0; i<N_RNA-1; ++i) {
    l_emit_count_[i] = &params_[x++];
  }

  // for state transitions
  for (int s=0; s<N_STATES; ++s) {
    for (int t=0; t<N_STATES; ++t) {
      state_count_[s][t] = &zero_;
    }
  }
  for (int s=0; s<n_normal_states; ++s) {
    for (int t=0; t<n_normal_states; ++t) {
      state_count_[s][t] = &params_[x++];
    }
  }
  state_count_[L_START][MATCH] = &params_[x++];
  state_count_[L_END][MATCH] = state_count_[L_START][MATCH];
  state_count_[MATCH][L_END] = &params_[x++];
  state_count_[MATCH][L_START] = state_count_[MATCH][L_END];
  //state_count_[L_START][L_START] = &params_[x++];
  //state_count_[L_END][L_END] = state_count_[L_START][L_START];

  // for branch transtion
  branch_count_ = &params_[x];
  x += n_branches;

  assert(x==n_params);

  clear();
}

template <class CostType>
void
CostTable<CostType>::
clear() 
{
  for (int i=0; i<n_params; ++i)
    params_[i] = zero_;
}

template <class CostType>
void
CostTable<CostType>::
normalize()
{
  cost_type state_total[N_STATES];
  for (int i=0; i<N_STATES; ++i)
    state_total[i] = static_cast<cost_type>(0.0);
  cost_type br_match_total(0.0);
  cost_type l_total(0.0);
  cost_type stem_total(0.0);
  cost_type single_total(0.0);
  cost_type local_total(0.0);

  // for single emits
  for (int i=0; i<N_RNA-1; ++i) {
    for (int j=i; j<N_RNA-1; ++j) {
      single_total += *emit_count_[i][RNA_GAP][j][RNA_GAP];
    }
    single_total += *emit_count_[i][RNA_GAP][RNA_GAP][RNA_GAP];
  }
  for (int i=0; i<N_RNA-1; ++i) {
    for (int j=i; j<N_RNA-1; ++j) {
      if (single_total>0.0)
	*emit_count_[i][RNA_GAP][j][RNA_GAP] /= single_total;
    }
    if (single_total>0.0)
      *emit_count_[i][RNA_GAP][RNA_GAP][RNA_GAP] /= single_total;
  }

  // for stem emits
  for (int i=0; i<N_RNA-1; ++i) {
    for (int j=0; j<N_RNA-1; ++j) {
      for (int k=0; k<N_RNA-1; ++k) {
	for (int l=0; l<N_RNA-1; ++l) {
	  if ((i<=k && (i!=k || j<=l)))
	    stem_total += *emit_count_[i][j][k][l];
	}
      }
      stem_total += *emit_count_[i][j][RNA_GAP][RNA_GAP];
    }
  }
  for (int i=0; i<N_RNA-1; ++i) {
    for (int j=0; j<N_RNA-1; ++j) {
      for (int k=0; k<N_RNA-1; ++k) {
	for (int l=0; l<N_RNA-1; ++l) {
	  if ((i<=k && (i!=k || j<=l)) && stem_total>0.0)
	    *emit_count_[i][j][k][l] /= stem_total;
	}
      }
      if (stem_total>0.0)
	*emit_count_[i][j][RNA_GAP][RNA_GAP] /= stem_total;
    }
  }

  // for local emits
  for (int i=0; i<N_RNA-1; ++i)
    local_total += *l_emit_count_[i];
  for (int i=0; i<N_RNA-1; ++i)
    if (local_total>0.0)
      *l_emit_count_[i] /= local_total;

  // count state transisions
  for (int s=0; s<n_normal_states; ++s) {
    for (int t=0; t<n_normal_states; ++t) {
      state_total[s] += *state_count_[s][t];
    }
  }
  state_total[MATCH] += *state_count_[MATCH][L_END];

  for (int s=0; s<n_normal_states; ++s) {
    for (int t=0; t<n_normal_states; ++t) {
      if (state_total[s]>0.0) {
	*state_count_[s][t] /= state_total[s];
      }
    }
  }
  if (state_total[MATCH]>0.0)
    *state_count_[MATCH][L_END] /= state_total[MATCH];

  l_total += *state_count_[L_START][MATCH];
  //l_total += *state_count_[L_START][L_START];

  if (l_total>0.0) {
    *state_count_[L_START][MATCH] /= l_total;
  //*state_count_[L_START][L_START] /= l_total;
  }

  for (int i=0; i<1; ++i)
    br_match_total += branch_count_[i];
  for (int i=0; i<1; ++i)
    if (br_match_total>0.0)
      branch_count_[i] /= br_match_total;
}

// private inline
template <class CostType>
template <class RNA>
int
CostTable<CostType>::
index(const RNA& rna) const
{
  int ret=RNA_GAP;
  if (isupper(rna))
    ret=DEFAULT_INDEX_TABLE[rna-'A'];
  else if (islower(rna))
    ret=DEFAULT_INDEX_TABLE[rna-'a'];
  if (ret>RNA_GAP /*&& !_allow_expand*/)
    ret=RNA_GAP;
  if (RNA_SYMBOL[ret]!=rna) {
    std::cout << rna << std::endl;
  }
  assert(RNA_SYMBOL[ret]==rna);
  return ret;
}

static struct {
  const char* symbol;
  match_state_t state;
} state_map[] = {
  { "M", MATCH },
  { "I", INSERT },
  { "D", DELETE },
  { "L", L_START },
  { NULL, MATCH }
};

// private inline
template <class CostType>
match_state_t
CostTable<CostType>::
state_index(const std::string& s) const
{
  for (int i=0; state_map[i].symbol!=NULL; ++i) {
    if (s==state_map[i].symbol) return state_map[i].state;
  }
  return (match_state_t)-1;
}

// private inline
template <class CostType>
const char*
CostTable<CostType>::
state_symbol(int t) const
{
  for (int i=0; state_map[i].symbol!=NULL; ++i) {
    if (state_map[i].state==t) return state_map[i].symbol;
  }
  assert(false);
  return NULL;
}

// template instantiation for 'cost_t::value_type'
template
class CostTable<cost_t::value_type>;

template
CostTable<cost_t::value_type>::cost_type
CostTable<cost_t::value_type>::
emit_cost(const rna_t& i, const rna_t& j,
	  const rna_t& k, const rna_t& l) const;

template
CostTable<cost_t::value_type>::cost_type
CostTable<cost_t::value_type>::
emit_cost(const rna_t& i) const;

template
void 
CostTable<cost_t::value_type>::
add_emit_count(const rna_t& i, const rna_t& j,
	       const rna_t& k, const rna_t& l, const cost_t::value_type& cnt);

template
void 
CostTable<cost_t::value_type>::
add_emit_count(const rna_t& i, const cost_t::value_type& cnt);

template
void 
CostTable<cost_t::value_type>::
set_emit_count(const rna_t& i, const rna_t& j,
	       const rna_t& k, const rna_t& l, const cost_t::value_type& cnt);

template
void 
CostTable<cost_t::value_type>::
set_emit_count(const rna_t& i, const cost_t::value_type& cnt);

// template instantiation for 'cost_t'
#include "log_value.h"

template
class CostTable<cost_t>;

template
cost_t
CostTable<cost_t>::
emit_cost(const rna_t& i, const rna_t& j,
	  const rna_t& k, const rna_t& l) const;

template
cost_t
CostTable<cost_t>::
emit_cost(const rna_t& i) const;

template
void 
CostTable<cost_t>::
add_emit_count(const rna_t& i, const rna_t& j,
	       const rna_t& k, const rna_t& l, const cost_t& cnt);

template
void 
CostTable<cost_t>::
add_emit_count(const rna_t& i, const cost_t& cnt);

template
void 
CostTable<cost_t>::
set_emit_count(const rna_t& i, const rna_t& j,
	       const rna_t& k, const rna_t& l, const cost_t& cnt);


// template instantiation for 'LogValue<short>'
template
class CostTable< LogValue<short> >;

template
LogValue<short>
CostTable< LogValue<short> >::
emit_cost(const rna_t& i, const rna_t& j,
	  const rna_t& k, const rna_t& l) const;

template
LogValue<short>
CostTable< LogValue<short> >::
emit_cost(const rna_t& i) const;

template
void 
CostTable< LogValue<short> >::
set_emit_count(const rna_t& i, const rna_t& j,
	       const rna_t& k, const rna_t& l, const LogValue<short>& cnt);

template
void 
CostTable< LogValue<short> >::
set_emit_count(const rna_t& i, const LogValue<short>& cnt);

