/*
 * $Id: costtable.h 487 2006-07-22 16:24:57Z satoken $
 *
 * PHMMTS -- an implementation of Pair Hidden Markov Models on Tree Structures,
 *           which is based on "Pair Hidden Markov Models on Tree Structures",
 *           Yasubumi Sakakibara, ISMB 2003.
 *
 * Copyright (C) 2003-2006 Kengo Sato
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifndef __INC_COSTTABLE_H__
#define __INC_COSTTABLE_H__

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <iostream>
#include <vector>
#include <list>
#include "defines.h"
#ifdef HAVE_MPI
#include <mpi.h>
#endif

namespace PHMMTS {

  template <class CostType>
  class CostTable
  {
  public:
    typedef CostType cost_type;
    
  private:
    static const int n_normal_states = 3;
    static const int n_single_emits = N_RNA*(N_RNA-1)/2+4;
    static const int n_stems = 4*4+1;
    static const int n_stem_emits = n_stems*(n_stems-1)/2+n_stems-1;
    static const int n_emits = n_single_emits + n_stem_emits;
    static const int n_l_emits = N_RNA-1;
    static const int n_states = n_normal_states*n_normal_states+2;
    static const int n_branches = 1;

  public:
    static const int n_params = n_emits + n_l_emits + n_states + n_branches;

  private:
    cost_type* emit_count_[N_RNA][N_RNA][N_RNA][N_RNA];
    cost_type* l_emit_count_[N_RNA-1];
    cost_type* state_count_[N_STATES][N_STATES];
    cost_type* branch_count_;
    cost_type params_[n_params];
    cost_type zero_;
    bool has_evalue_parameters_;
    float evalue_lambda_;
    float evalue_const_;

  public:
    /** 
     * A default constructor
     * 
     */
    CostTable() : zero_(0.0), has_evalue_parameters_(false) { initialize(); }

    /** 
     * A copy constructor
     * 
     * @param x  a source object
     */
    CostTable(const CostTable& x) : zero_(0.0)
    {
      assert(size()==x.size());
      initialize();
      for (size_t i=0; i<size(); ++i)
	(*this)[i] = x[i];
    }

    /** 
     * A copy constructor
     * 
     * @param x  a source object
     */
    template <class T>
    CostTable(const T& x) : zero_(0.0)
    {
      assert(size()==x.size());
      initialize();
      for (size_t i=0; i<size(); ++i)
	(*this)[i] = x[i];
    }

    /** 
     * An accessor with an index
     * 
     * @param i an index of an element
     * 
     * @return the i-th value
     */
    cost_type& operator[](size_t i)
    {
      return params_[i];
    }

    /** 
     * An accessor with an index
     * 
     * @param i an index of an element
     * 
     * @return the i-th value
     */
    const cost_type& operator[](size_t i) const
    {
      return params_[i];
    }

    /** 
     * An assignment operator
     * 
     * @param x  a source object
     * 
     * @return the result of the assignment
     */
    CostTable& operator=(const CostTable& x)
    {
      assert(size()==x.size());
      for (size_t i=0; i<size(); ++i)
	(*this)[i] = x[i];
      return *this;
    }

    /** 
     * An assignment operator
     * 
     * @param x  a source object
     * 
     * @return the result of the assignment
     */
    template <class T>
    CostTable& operator=(const T& x)
    {
      assert(size()==x.size());
      for (size_t i=0; i<size(); ++i)
	(*this)[i] = x[i];
      return *this;
    }

    /** 
     * Return the size of this object
     * 
     * @return the size of this object
     */
    size_t size() const
    {
      return n_params;
    }

    /** 
     * Make this object be a probability distribution
     * 
     */
    void normalize();

    template < class RNA >
    cost_type emit_cost(const RNA& i, const RNA& j,
			const RNA& k, const RNA& l) const;

    cost_type emit_cost(char i, char j,char k, char l) const;

    template < class RNA >
    cost_type emit_cost(const RNA& i) const;

    cost_type emit_cost(char i) const;

    template < class RNA >
    void add_emit_count(const RNA& i, const RNA& j,
			const RNA& k, const RNA& l, const cost_type& cnt);

     void add_emit_count(char i, char j, char k, char l, const cost_type& cnt);

    template < class RNA >
    void add_emit_count(const RNA& i, const cost_type& cnt);

    void add_emit_count(char i, const cost_type& cnt);

    template < class RNA >
    void set_emit_count(const RNA& i, const RNA& j,
			const RNA& k, const RNA& l, const cost_type& cnt);

    void set_emit_count(char i, char j, char k, char l, const cost_type& cnt);

    template < class RNA >
    void set_emit_count(const RNA& i, const cost_type& cnt);

    void set_emit_count(char i, const cost_type& cnt);

    cost_type state_cost(match_state_t from, match_state_t to) const;
    cost_type state_cost(match_state_t from,
			 match_state_t to_left, match_state_t to_right) const;

    void add_state_count(match_state_t from, match_state_t to,
			 const cost_type& cnt);
    void add_state_count(match_state_t from, match_state_t to_left,
			 match_state_t to_right, const cost_type& cnt);
    void set_state_count(match_state_t from, match_state_t to,
			 const cost_type& cnt);
    void set_state_count(match_state_t from, match_state_t to_left,
			 match_state_t to_right, const cost_type& cnt);

    void store(std::ostream& os, bool logged=false,
	       const cost_type& eps=static_cast<cost_type>(1.0e-50)) const;

    void restore(std::istream& is);

    void clear();

    void add(const CostTable& t)
    {
      for (int i=0; i!=n_params; ++i)
	params_[i] += t.params_[i];
    }

    bool has_evalue_parameters() const { return has_evalue_parameters_; }
    float evalue_lambda() const { return evalue_lambda_; }
    float evalue_const() const { return evalue_const_; }

#ifdef HAVE_MPI
    void send(uint dest, int tag) const
    {
      MPI::COMM_WORLD.Send(params_, n_params, MPI::FLOAT, dest, tag);
    }

    void recv(uint src, int tag)
    {
      MPI::COMM_WORLD.Recv(params_, n_params, MPI::FLOAT, src, tag);
    }
#endif    

  private:
    template <class RNA>
    inline int index(const RNA& rna) const;
    inline match_state_t state_index(const std::string& s) const;
    inline const char* state_symbol(int t) const;
    void initialize();
  };
}

#endif // __INC_COSTTABLE_H__

// Local Variables:
// mode: C++
// End:
