/*
 * $Id: crf_train.cpp 487 2006-07-22 16:24:57Z satoken $
 *
 * PHMMTS -- an implementation of Pair Hidden Markov Models on Tree Structures,
 *           which is based on "Pair Hidden Markov Models on Tree Structures",
 *           Yasubumi Sakakibara, ISMB 2003.
 *
 * Copyright (C) 2003-2006 Kengo Sato
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <iostream>
#include <fstream>
#include <list>
#include <algorithm>
#include "skeletal.h"
#include "multiseq.h"
#include "costtable.h"
#include "log_value.h"
#include "lbfgs.h"
#include "example.h"
#include "dptable.h"
#include "cyk.h"
#include "inside.h"
#include "outside.h"
#include "inside_outside.h"
#ifdef HAVE_MPI
#include <mpi.h>
#endif

using namespace PHMMTS;

typedef CostTable<float> Expectation;
typedef CostTable<cost_t> LCostTable;

template <class CostTable, class Expl>
class ExpectationCRFFunc : public std::unary_function<const Expl&, void>
{
  typedef typename CostTable::cost_type cost_type;
  
  const CostTable& cost_table_;
  Expectation& ex_;
  search_type_t search_type_;
  float& log_likelihood_;
  int verbose_;
  size_t cur_;
  size_t stage_;
  size_t num_;

public:
  ExpectationCRFFunc(const CostTable& cost_table, Expectation& ex,
		     search_type_t search_type,
		     float& ll, int verbose=2, size_t stage=1, size_t num=0)
    : cost_table_(cost_table), ex_(ex), 
      search_type_(search_type), log_likelihood_(ll),
      verbose_(verbose), cur_(0), stage_(stage), num_(num)
  {
    log_likelihood_ = 0.0;
  }

  void operator()(const Expl& x)
  {
    cost_type p, c;
#ifndef HAVE_MPI
    if (verbose_>=2 && num_>0) {
      std::cout << "- Stage: " << stage_ << ", "
		<< ++cur_ << '/' << num_ << ": "
		<< std::flush;
    }
#endif

    Inside<CostTable> inside(cost_table_, search_type_);
    p = inside.run(x.skeletal_tree(), x.unfolded_seq());
    c = x.cost(cost_table_);
    float ll = log(c)-log(p);
    if (ll>0.0) ll=0.0;
    log_likelihood_ += ll;
    
    Outside<CostTable> outside(cost_table_, inside, search_type_);
    outside.run(x.skeletal_tree(), x.unfolded_seq());

    InsideOutside<CostTable,Expectation>
      inside_outside(cost_table_, inside, outside, ex_, search_type_);
    inside_outside.run(x.skeletal_tree(), x.unfolded_seq());
    
#ifndef HAVE_MPI
    if (verbose_>=2 && num_>0) {
      std::cout << "LL= " << ll;
      if (log(c)-log(p)>0.0 && verbose_>=3) {
	std::cout << " (error "
		  << log(c) << "-" << log(p) << ")";
	CYK<CostTable> cyk(cost_table_, search_type_);
	cost_type q = cyk.run(x.skeletal_tree(), x.unfolded_seq());
	std::cout << " " << log(q);
      }
      std::cout << std::endl;
    }
#endif
  }
};

template < class Expl >
static
void
object_func(const std::vector<double>& x, double& f, std::vector<double>& g,
	    const std::list<Expl>& train, const Expectation& cf,
	    float sigma, int verbose, size_t stage, search_type_t search_type)
{
  // copy current values into cost_table to use as cost of the parser
  LCostTable cost_table;
  for (size_t i=0; i<x.size(); ++i)
    cost_table[i].set_log_value(x[i]);

  // parse each training data and calc expectations
#ifdef HAVE_MPI
  int my_rank = MPI::COMM_WORLD.Get_rank();
  int num_procs = MPI::COMM_WORLD.Get_size();
  Expectation ex, lex;
  float log_likelihood;
  ExpectationCRFFunc<LCostTable, Expl>
    func(cost_table, lex, search_type, log_likelihood,
	 verbose, stage, train.size());
  typename std::list<Expl>::const_iterator p;
  int cnt=0;
  for (p=train.begin(); p!=train.end(); ++p) {
    if (cnt++%num_procs==my_rank) func(*p);
  }
  if (my_rank==0) {
    ex.add(lex);
    for (int t=1; t!=num_procs; ++t) {
      float v;
      MPI::COMM_WORLD.Recv(&v, 1, MPI::FLOAT, t, 0);
      log_likelihood += v;
      lex.recv(t, 1);
      ex.add(lex);
    }
    for (int t=1; t!=num_procs; ++t) {
      MPI::COMM_WORLD.Send(&log_likelihood, 1, MPI::FLOAT, t, 0);
      ex.send(t, 1);
    }
  } else {
    MPI::COMM_WORLD.Send(&log_likelihood, 1, MPI::FLOAT, 0, 0);
    lex.send(0, 1);
    MPI::COMM_WORLD.Recv(&log_likelihood, 1, MPI::FLOAT, 0, 0);
    ex.recv(0, 1);
  }
#else  //  HAVE_MPI
  Expectation ex;
  float log_likelihood;
  ExpectationCRFFunc<LCostTable, Expl>
    func(cost_table, ex, search_type, log_likelihood,
	 verbose, stage, train.size());
  std::for_each(train.begin(), train.end(), func);
#endif	//  HAVE_MPI

  // object func
  float s=0.0;
  for (size_t i=0; i<x.size(); ++i)
    s += x[i]*x[i];
  f = -log_likelihood + s/(2*sigma*sigma);
  // gradients
  for (size_t i=0; i<g.size(); ++i)
    g[i] = -(cf[i] - ex[i]) + x[i]/(sigma*sigma);
}

template < class Expl >
void
crf_training(const std::list<Expl>& train,
	     std::vector<double>& x,
	     const Expectation& cf,
	     size_t max_it, float eps, float sigma, int verbose,
	     size_t& it, float& ll, search_type_t search_type)
{
  double f,f_prev;
  std::vector<double> g(cf.size());

  LBFGS lbfgs;
  lbfgs.initialize(cf.size(), 5);

  // the first time
  {
    object_func(x, f, g, train, cf, sigma, verbose, 1, search_type);
    lbfgs.update(&x[0], &f, &g[0]);
    if (verbose>=1)
      std::cout << "== Stage 1: sum of LL= " << -f << std::endl;
    f_prev = f;
  }

  // iterate until the convergence of the object func has come.
  for (it=1; it<max_it; ++it) {
    float conv_rate;
    int ret;
    object_func(x, f, g, train, cf, sigma, verbose, it+1, search_type);
    conv_rate = std::abs((f-f_prev)/f_prev);
    if (conv_rate < eps) {
      if (verbose>=1)
	std::cout << "== Stage " << it+1
		  << ": sum of LL= " << -f << ", "
		  << "conv rate= " << conv_rate /*<< ", "*/
		  << /*"state: " << ret <<*/ std::endl;
      break;
    }
    ret = lbfgs.update(&x[0], &f, &g[0]);

    if (verbose>=1)
      std::cout << "== Stage " << it+1
		<< ": sum of LL= " << -f << ", "
		<< "conv rate= " << conv_rate /*<< ", "*/
		<< /*"state: " << ret <<*/ std::endl;
    if (ret==0) break;
    else if (ret<0) std::cout << "** An error has happened!" << std::endl;
    f_prev = f;
  }
  ll = -f;
}

#if 0
template
void
crf_training(const std::list<Example>& train,
	     std::vector<double>& x,
	     const Expectation& cf,
	     size_t max_it, float eps, float sigma, int verbose,
	     size_t& it, float& ll, bool l_opt);
#endif

template
void
crf_training(const std::list<ExampleS>& train,
	     std::vector<double>& x,
	     const Expectation& cf,
	     size_t max_it, float eps, float sigma, int verbose,
	     size_t& it, float& ll, search_type_t search_type);
