/*
 * $Id: cyk.cpp 482 2006-04-12 09:00:23Z satoken $
 *
 * PHMMTS -- an implementation of Pair Hidden Markov Models on Tree Structures,
 *           which is based on "Pair Hidden Markov Models on Tree Structures",
 *           Yasubumi Sakakibara, ISMB 2003.
 *
 * Copyright (C) 2003-2006 Kengo Sato
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "algo_impl.h"
#include "cyk.h"
#include "background.h"
#include "progress.h"

using namespace PHMMTS;

template <class CostTable_, template<class> class DPTable_>
CYK<CostTable_,DPTable_>::
CYK(const CostTable_& cost_table, search_type_t search_type,
    ProgressBar* progress /*= NULL*/, int max_length /*=0*/)
  : pcyk_(), bcyk_(), bk_(), max_pos_(), max_(zero<cost_type>()),
    impl_(NULL), search_type_(search_type), progress_(progress)
{
  int l=0;
  switch (search_type_) {
  case GLOBAL_ALIGNMENT:
    break;
  case LOCAL_ALIGNMENT:
    l = Impl::UNFOLDED_LSTART;
    break;
  case LOCAL_STRUCTURE:
    l = (Impl::FOLDED_LSTART | Impl::FOLDED_LEND |
	 Impl::UNFOLDED_LSTART | Impl::UNFOLDED_LEND );
    break;
  default:
    assert(!"no such search type.");
    break;
  }
  impl_ = new Impl(cost_table, l, max_length);
}

template <class CostTable_, template<class> class DPTable_>
CYK<CostTable_,DPTable_>::
~CYK()
{
  delete impl_;
}

template <class CostTable_, template<class> class DPTable_>
template <class STree, class Sequence>
typename CYK<CostTable_,DPTable_>::cost_type
CYK<CostTable_,DPTable_>::
run(const STree& tree, const Sequence& rna, int from, int to)
{
  // initialize CYK tables
  int tree_size = tree.size();

  max_pos_ = Position();
  max_ = static_cast<cost_type>(0.0);

  pcyk_.set_size(tree_size, from, to);
  bcyk_.set_size(tree_size, from, to);

  if (progress_) progress_->start(0, tree_size);
    
  pcyk_.create();
  bcyk_.create();
  bk_.initialize(impl_->cost_table(), tree, rna, from, to);
  
  // run CYK algorithm for each table
  for (int h=0; h<tree_size; ++h) {
    pcyk_.create(h, tree[h].size());

    switch(tree[h].type()) {
    case EMPTY:
      impl_->inside_empty(tree, h, rna, from, to, *this);
      break;
    case LSINGLE:
    case RSINGLE:
    case STEM:
    case START:
      impl_->inside_normal(tree, h, rna, from, to, *this);
      break;
    case BRANCH:
      bcyk_.create(h, 1);
      impl_->inside_branch(tree, h, rna, from, to, *this);
      break;
    default:
      assert(!"no such node type");
      break;
    }
    if (progress_) progress_->update();
  }
  if (progress_) progress_->finish();

  if (search_type_==GLOBAL_ALIGNMENT) {
    for (uint s=0; s<tree[tree_size-1].size(); ++s) {
      if (pcyk_(tree_size-1, s, from, to).use() &&
	  pcyk_(tree_size-1, s, from, to).cost() > max_) {
	max_pos_ = Position(tree_size-1, s, from, to);
	max_ = pcyk_(tree_size-1, s, from, to).cost();
      }
    }
  }
  return max_;
}

template <class CostTable_, template<class> class DPTable_>
template <class STree, class Sequence>
typename CYK<CostTable_,DPTable_>::cost_type
CYK<CostTable_,DPTable_>::
scan_first(const STree& tree, const Sequence& rna, int from, int to)
{
  // initialize CYK tables
  int win_size = to-from+1;
  assert(impl_->max_length()<=win_size);
  int tree_size = tree.size();
  max_pos_ = Position();
  max_ = static_cast<cost_type>(0.0);
  pcyk_.set_size(tree_size, win_size);
  bcyk_.set_size(tree_size, win_size);
  bk_.initialize(impl_->cost_table(), tree, rna, from, to);

  if (progress_) progress_->start(0, tree_size);

  // run CYK algorithm for each table
  for (int h=0; h<tree_size; ++h) {
    pcyk_.create(h, tree[h].size(), cell_type());

    switch(tree[h].type()) {
    case EMPTY:
      impl_->inside_empty(tree, h, rna, from, to, *this);
      break;
    case LSINGLE:
    case RSINGLE:
    case STEM:
    case START:
      impl_->inside_normal(tree, h, rna, from, to, *this);
      break;
    case BRANCH:
      bcyk_.create(h, 1, BrCell());
      impl_->inside_branch(tree, h, rna, from, to, *this);
      break;
    default:
      assert(!"no such node type");
      break;
    }
    if (progress_) progress_->update();
  }
  if (progress_) progress_->finish();

  if (search_type_==GLOBAL_ALIGNMENT) {
    for (uint s=0; s<tree[tree_size-1].size(); ++s) {
      if (pcyk_(tree_size-1, s, from, to).use() &&
	  pcyk_(tree_size-1, s, from, to).cost() > max_) {
	max_pos_ = Position(tree_size-1, s, from, to);
	max_ = pcyk_(tree_size-1, s, from, to).cost();
      }
    }
  }
  return max_;
}

template <class CostTable_, template<class> class DPTable_>
template <class STree, class Sequence>
typename CYK<CostTable_,DPTable_>::cost_type
CYK<CostTable_,DPTable_>::
scan_next(const STree& tree, const Sequence& rna, int from, int to, int pos)
{
  assert(impl_->max_length()<=pcyk_.get_length());
  int tree_size = tree.size();
  bk_.update_position(impl_->cost_table(), rna, from, to);
  if (max_pos_.left<from) {
    max_pos_ = Position();
    max_ = static_cast<cost_type>(0.0);
  }

  if (progress_) progress_->start(0, tree_size);
  
  // run CYK algorithm for each table
  for (int h=0; h<tree_size; ++h) {
    // clear previous results
    for (int j=pos; j<to; ++j) {
      for (int i=j; i>=from && j-i<impl_->max_length(); --i) {
	for (uint s=0; s<tree[h].size(); ++s) {
	  pcyk_(h,s,i,j+1)=cell_type();
	}
      }
    }
    
    switch(tree[h].type()) {
    case EMPTY:
      impl_->scan_empty(tree, h, rna, from, to, pos, *this);
      break;
    case LSINGLE:
    case RSINGLE:
    case STEM:
    case START:
      impl_->scan_normal(tree, h, rna, from, to, pos, *this);
      break;
    case BRANCH:
      impl_->scan_branch(tree, h, rna, from, to, pos, *this);
      break;
    default:
      assert(!"no such node type");
      break;
    }
    if (progress_) progress_->update();
  }
  if (progress_) progress_->finish();

  if (search_type_==GLOBAL_ALIGNMENT) {
    for (uint s=0; s<tree[tree_size-1].size(); ++s) {
      if (pcyk_(tree_size-1, s, from, to).use() &&
	  pcyk_(tree_size-1, s, from, to).cost() > max_) {
	max_pos_ = Position(tree_size-1, s, from, to);
	max_ = pcyk_(tree_size-1, s, from, to).cost();
      }
    }
  }
  return max_;
}

template < class CostTable_, template <class> class DPTable_ >
template < class SkeletalTree_ >
typename CYK<CostTable_,DPTable_>::cost_type
CYK<CostTable_,DPTable_>::
local_cost(const SkeletalTree_& skel) const
{
  if (pcyk_(max_pos_).state()==-1) // L_END
    return max_;

  if (skel[max_pos_.height].left()<0) // EMPTY node
    return local_cost(skel, max_pos_);

  return local_cost(skel, max_pos_) * max_ / pcyk_(max_pos_).cost();
}

template < class CostTable_, template <class> class DPTable_ >
template < class SkeletalTree_ >
typename CYK<CostTable_,DPTable_>::cost_type
CYK<CostTable_,DPTable_>::
local_cost(const SkeletalTree_& skel, const Position& pos) const
{
  if (pcyk_(pos).state()==-1) {	// L_END
    return pcyk_(pos).cost();
  }

  if (skel[pos.height].left()<0) { // EMPTY node
    assert(pos.left==pos.right);
    return static_cast<cost_type>(0.0);
  }

  const cell_type& node = pcyk_(pos);
  typedef typename SkeletalTree_::node_type::Property Property;
  const Property& p = skel[pos.height].property(pos.state);
  if (p.rchildren.empty()) {	// not BRANCH node
    int ch = node.self_trans() ? pos.height : skel[pos.height].left();
    return
      local_cost(skel, 
		 Position(ch, node.state(),
			  pos.left+p.u_left, pos.right-p.u_right));
  } else {			// BRANCH node
    int mid = bcyk_(pos).pos();
    int rstate = bcyk_(pos).state();
    return
      local_cost(skel, 
		 Position(skel[pos.height].left(),
			  node.state(), pos.left, mid)) *
      local_cost(skel, 
		 Position(skel[pos.height].right(),
			  rstate, mid, pos.right));
  }
}

template < class CostTable_, template <class> class DPTable_ >
template < class SkeletalTree_, class Sequence_>
size_t
CYK<CostTable_,DPTable_>::
traceback(const SkeletalTree_& skel, const Sequence_& rna, 
	  typename SkeletalTree_::Sequence& folded, Sequence_& unfolded,
	  std::string& state_seq, std::string& paren) const
{
  return traceback(skel, rna, pcyk_, bcyk_,
		   max_pos_, folded, unfolded, state_seq, paren);
}

template < class CostTable_, template <class> class DPTable_ >
template < class SkeletalTree_, class Sequence_>
size_t
CYK<CostTable_,DPTable_>::
traceback(const SkeletalTree_& skel, const Sequence_& rna,
	  const pcyk_table& pcyk, const bcyk_table& bcyk,
	  const Position& pos,
	  typename SkeletalTree_::Sequence& folded, Sequence_& unfolded,
	  std::string& state_seq, std::string& paren) const
{
  size_t ret=0;
  typedef typename SkeletalTree_::Sequence::value_type fc_t;
  typedef typename Sequence_::value_type uc_t;
  fc_t fc;
  uc_t uc;

  if (pos.state==-1) {		// L_END
    typename SkeletalTree_::Sequence f;
    Sequence_ u;
    std::string p;
    skel.to_string(f, p, pos.height);
    for (int i=pos.left; i<pos.right; ++i) u.push_back(rna[i]);
    while (f.size() < u.size()) {
      f.push_back(static_cast<fc_t>(RNASymbol<fc_t>::GAP));
      p += ' ';
    }
    while (f.size() > u.size()) {
      u.push_back(static_cast<uc_t>(RNASymbol<uc_t>::GAP));
    }
    {
      typename SkeletalTree_::Sequence::iterator x;
      for (x=f.begin(); x!=f.end(); ++x) {
	folded.push_back(*x);
      }
    }
    {
      typename Sequence_::iterator x;
      for (x=u.begin(); x!=u.end(); ++x) {
	unfolded.push_back(*x);
      }
    }
    paren += p;
    state_seq.append(f.size(), ' ');
    return f.size();
  }

  typedef typename SkeletalTree_::node_type::Property Property;
  const Property& p = skel[pos.height].property(pos.state);
  int n=0;
  
  const cell_type& node = pcyk(pos);
  if (skel[pos.height].left()<0) {
    // EMPTY node
    assert(pos.left==pos.right);
    return 0;

  } else {
    // output left RNA
    if (p.f_left || p.u_left) {
      n++;
      fc = p.f_left ?
	skel[pos.height].left_rna() :
	static_cast<fc_t>(RNASymbol<fc_t>::GAP);
      folded.push_back(fc);
      uc = p.u_left ?
	rna[pos.left] :
	static_cast<uc_t>(RNASymbol<uc_t>::GAP);
      unfolded.push_back(uc);
      state_seq += p.f_left && p.u_left && fc==uc ? p.match : p.mismatch;
      paren += p.f_left && p.f_right || p.u_left && p.u_right ? '(' : ' ';
    }

    // recursive call for children nodes
    if (p.rchildren.empty()) {	// not BRANCH node
      int ch = node.self_trans() ? pos.height : skel[pos.height].left();
      ret = traceback(skel, rna, pcyk, bcyk,
		      Position(ch, node.state(),
			       pos.left+p.u_left, pos.right-p.u_right),
		      folded, unfolded, state_seq, paren);
    } else {			// BRANCH node
      int mid = bcyk(pos).pos();
      int rstate = bcyk(pos).state();
      ret = traceback(skel, rna, pcyk, bcyk,
		      Position(skel[pos.height].left(),
			       node.state(), pos.left, mid),
		      folded, unfolded, state_seq, paren);
      ret+= traceback(skel, rna, pcyk, bcyk,
		      Position(skel[pos.height].right(),
			       rstate, mid, pos.right),
		      folded, unfolded, state_seq, paren);
    }

    // output right RNA
    if (p.f_right || p.u_right) {
      n++;
      fc = p.f_right ?
	skel[pos.height].right_rna() :
	static_cast<fc_t>(RNASymbol<fc_t>::GAP);
      folded.push_back(fc);
      uc = p.u_right ?
	rna[pos.right-1] :
	static_cast<uc_t>(RNASymbol<uc_t>::GAP);
      unfolded.push_back(uc);
      state_seq += p.f_right && p.u_right && fc==uc ? p.match : p.mismatch;
      paren += p.f_left && p.f_right || p.u_left && p.u_right ? ')' : ' ';
    }

    return ret+n;
  }
}


template < class CostTable_, template <class> class DPTable_ >
template < class CostType, class Position, class STree, class Seq >
void
CYK<CostTable_,DPTable_>::
operator()(node_type_t type, const CostType& emit, const Position& parent,
	   const STree& tree, const Seq& rna)
{
  assert(type==EMPTY);
  cell_type n(-1, false, emit);
  pcyk_(parent).update(n);
}

template < class CostTable_, template <class> class DPTable_ >
template < class CostType, class Position, class STree, class Seq >
void
CYK<CostTable_,DPTable_>::
operator()(node_type_t type,
	   const CostType& emit, const CostType& state,
	   const Position& parent, const Position& child,
	   const STree& tree, const Seq& rna)
{
  if (child.state==-1) {	// L_END
    CostType cost = bk_.inside(child) * emit * state;
    cell_type n(child.state, false, cost);
    pcyk_(parent).update(n);
  } else if (parent.state==-1) { // L_START
    if (pcyk_(child).use()) {
      CostType cost = pcyk_(child).cost() *  state * bk_.outside(parent);
      if (max_ < cost) {
	max_ = cost;
	max_pos_ = child;
      }
    }
  } else if (pcyk_(child).use()) {
    CostType cost = pcyk_(child).cost() * emit * state;
    cell_type n(child.state, parent.height==child.height, cost);
    pcyk_(parent).update(n);
  }
}

template < class CostTable_, template <class> class DPTable_ >
template < class CostType, class Position, class STree, class Seq >
void
CYK<CostTable_,DPTable_>::
operator()(node_type_t type,
	   const CostType& emit, const CostType& state, 
	   const Position& parent,
	   const Position& lchild, const Position& rchild,
	   const STree& tree, const Seq& rna)
{
  assert(type==BRANCH);
  if (pcyk_(lchild).use() && pcyk_(rchild).use()) {
    CostType cost = pcyk_(lchild).cost()*pcyk_(rchild).cost()*emit*state;
    cell_type n(lchild.state, false, cost);
    if (pcyk_(parent).update(n)) {
      bcyk_(parent.height, 0, parent.left, parent.right)
	= BrCell(lchild.right, rchild.state);
    }
  }
}



// template instantiation for 'cost_t' and 'DPTable'
#include <string>
#include "costtable.h"
#include "log_value.h"
#include "skeletal.h"
#include "dptable.h"

template
class CYK< CostTable<cost_t>, DPTable >;

template
CYK< CostTable<cost_t>, DPTable >::cost_type
CYK< CostTable<cost_t>, DPTable >::
run(const SkeletalTreeS& tree, const RNASequence& rna, int from, int to);

template
size_t
CYK< CostTable<cost_t>, DPTable >::
traceback(const SkeletalTreeS& skel, const RNASequence& rna, 
	  SkeletalTreeS::Sequence& folded,
	  RNASequence& unfolded,
	  std::string& state_seq, std::string& paren) const;

template
CYK< CostTable<cost_t>, DPTable >::cost_type
CYK< CostTable<cost_t>, DPTable >::
local_cost(const SkeletalTreeS& skel) const;


// template instantiation for 'cost_t' and 'ScanTable'
#include "scantable.h"

template
class CYK< CostTable<cost_t>, ScanTable >;

template
CYK< CostTable<cost_t>, ScanTable >::cost_type
CYK< CostTable<cost_t>, ScanTable >::
scan_first(const SkeletalTreeS& tree, const RNASequence& rna, int from, int to);

template
CYK< CostTable<cost_t>, ScanTable >::cost_type
CYK< CostTable<cost_t>, ScanTable >::
scan_next(const SkeletalTreeS& tree, const RNASequence& rna,
	  int from, int to, int pos);

template
size_t
CYK< CostTable<cost_t>, ScanTable >::
traceback(const SkeletalTreeS& skel, const RNASequence& rna, 
	  SkeletalTreeS::Sequence& folded,
	  RNASequence& unfolded,
	  std::string& state_seq, std::string& paren) const;

template
CYK< CostTable<cost_t>, ScanTable >::cost_type
CYK< CostTable<cost_t>, ScanTable >::
local_cost(const SkeletalTreeS& skel) const;


// template instantiation for 'LogValue<short>' and 'DPTable'
#include "defaultcosttable.h"

template
class CYK< DefaultCostTable, DPTable >;

template
CYK< DefaultCostTable, DPTable >::cost_type
CYK< DefaultCostTable, DPTable >::
run(const SkeletalTreeS& tree, const RNASequence& rna, int from, int to);

template
size_t
CYK< DefaultCostTable, DPTable >::
traceback(const SkeletalTreeS& skel, const RNASequence& rna, 
	  SkeletalTreeS::Sequence& folded,
	  RNASequence& unfolded,
	  std::string& state_seq, std::string& paren) const;

template
CYK< DefaultCostTable, DPTable >::cost_type
CYK< DefaultCostTable, DPTable >::
local_cost(const SkeletalTreeS& skel) const;


// template instantiation for 'LogValue<short>' and 'ScanTable'
template
class CYK< DefaultCostTable, ScanTable >;

template
CYK< DefaultCostTable, ScanTable >::cost_type
CYK< DefaultCostTable, ScanTable >::
scan_first(const SkeletalTreeS& tree, const RNASequence& rna, int from, int to);

template
CYK< DefaultCostTable, ScanTable >::cost_type
CYK< DefaultCostTable, ScanTable >::
scan_next(const SkeletalTreeS& tree, const RNASequence& rna,
	  int from, int to, int pos);

template
size_t
CYK< DefaultCostTable, ScanTable >::
traceback(const SkeletalTreeS& skel, const RNASequence& rna, 
	  SkeletalTreeS::Sequence& folded,
	  RNASequence& unfolded,
	  std::string& state_seq, std::string& paren) const;

template
CYK< DefaultCostTable, ScanTable >::cost_type
CYK< DefaultCostTable, ScanTable >::
local_cost(const SkeletalTreeS& skel) const;
