/*
 * $Id: cyk_no_trace.cpp 482 2006-04-12 09:00:23Z satoken $
 *
 * PHMMTS -- an implementation of Pair Hidden Markov Models on Tree Structures,
 *           which is based on "Pair Hidden Markov Models on Tree Structures",
 *           Yasubumi Sakakibara, ISMB 2003.
 *
 * Copyright (C) 2003-2006 Kengo Sato
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "algo_impl.h"
#include "cyk_no_trace.h"
#include "background.h"
#include "progress.h"

using namespace PHMMTS;

template <class CostTable_, template<class> class DPTable_>
CYKnoTrace<CostTable_,DPTable_>::
CYKnoTrace(const CostTable_& cost_table, search_type_t search_type,
	   ProgressBar* progress /*= NULL*/, int max_length /*=0*/)
  : pcyk_(), bk_(), max_pos_(), max_(zero<cost_type>()),
    impl_(NULL), search_type_(search_type), progress_(progress)
{
  int l=0;
  switch (search_type_) {
  case GLOBAL_ALIGNMENT:
    break;
  case LOCAL_ALIGNMENT:
    l = Impl::UNFOLDED_LSTART;
    break;
  case LOCAL_STRUCTURE:
    l = (Impl::FOLDED_LSTART | Impl::FOLDED_LEND |
	 Impl::UNFOLDED_LSTART | Impl::UNFOLDED_LEND );
    break;
  default:
    assert(!"no such search type.");
    break;
  }
  impl_ = new Impl(cost_table, l, max_length);
}

template <class CostTable_, template<class> class DPTable_>
CYKnoTrace<CostTable_,DPTable_>::
~CYKnoTrace()
{
  delete impl_;
}

template <class CostTable_, template<class> class DPTable_>
template <class STree, class Sequence>
typename CYKnoTrace<CostTable_,DPTable_>::cost_type
CYKnoTrace<CostTable_,DPTable_>::
run(const STree& tree, const Sequence& rna, int from, int to)
{
  // initialize CYK tables
  int tree_size = tree.size();

  max_pos_ = Position();
  max_ = static_cast<cost_type>(0.0);
  pcyk_.set_size(tree_size, from, to);
  bk_.initialize(impl_->cost_table(), tree, rna, from, to);

  if (progress_) progress_->start(0, tree_size);

  // run CYK algorithm for each table
  for (int h=0; h<tree_size; ++h) {
    pcyk_.create(h, tree[h].size(),
		 static_cast<cell_type>(zero<cost_type>()));
    
    switch(tree[h].type()) {
    case EMPTY:
      impl_->inside_empty(tree, h, rna, from, to, *this);
      break;
    case LSINGLE:
    case RSINGLE:
    case STEM:
    case START:
      impl_->inside_normal(tree, h, rna, from, to, *this);
      if (tree[h].left()>=0)
	pcyk_.destroy(tree[h].left());
      break;
    case BRANCH:
      impl_->inside_branch(tree, h, rna, from, to, *this);
      if (tree[h].left()>=0)
	pcyk_.destroy(tree[h].left());
      if (tree[h].right()>=0)
	pcyk_.destroy(tree[h].right());
      break;
    default:
      assert(!"no such node type");
      break;
    }
    if (progress_) progress_->update();
  }

  if (progress_) progress_->finish();
  
  if (search_type_==GLOBAL_ALIGNMENT) {
    for (uint s=0; s<tree[tree_size-1].size(); ++s) {
      if (pcyk_(tree_size-1, s, from, to).use() &&
	  pcyk_(tree_size-1, s, from, to).cost() > max_) {
	max_pos_ = Position(tree_size-1, s, from, to);
	max_ = pcyk_(tree_size-1, s, from, to).cost();
      }
    }
  }
  return max_;
}


template <class CostTable_, template<class> class DPTable_>
template <class STree, class Sequence>
typename CYKnoTrace<CostTable_,DPTable_>::cost_type
CYKnoTrace<CostTable_,DPTable_>::
scan_first(const STree& tree, const Sequence& rna, int from, int to)
{
  // initialize CYK tables
  int win_size = to-from+1;
  assert(impl_->max_length()<=win_size);
  int tree_size = tree.size();
  max_pos_ = Position();
  max_ = static_cast<cost_type>(0.0);
  pcyk_.set_size(tree_size, win_size);
  bk_.initialize(impl_->cost_table(), tree, rna, from, to);

  if (progress_) progress_->start(0, tree_size);

  // run CYK algorithm for each table
  for (int h=0; h<tree_size; ++h) {
    pcyk_.create(h, tree[h].size(), 
		 static_cast<cell_type>(zero<cost_type>()));

    switch(tree[h].type()) {
    case EMPTY:
      impl_->inside_empty(tree, h, rna, from, to, *this);
      break;
    case LSINGLE:
    case RSINGLE:
    case STEM:
    case START:
      impl_->inside_normal(tree, h, rna, from, to, *this);
      break;
    case BRANCH:
      impl_->inside_branch(tree, h, rna, from, to, *this);
      break;
    default:
      assert(!"no such node type");
      break;
    }
    if (progress_) progress_->update();
  }
  if (progress_) progress_->finish();

  if (search_type_==GLOBAL_ALIGNMENT) {
    for (uint s=0; s<tree[tree_size-1].size(); ++s) {
      if (pcyk_(tree_size-1, s, from, to).use() &&
	  pcyk_(tree_size-1, s, from, to).cost() > max_) {
	max_pos_ = Position(tree_size-1, s, from, to);
	max_ = pcyk_(tree_size-1, s, from, to).cost();
      }
    }
  }
  return max_;
}

template <class CostTable_, template<class> class DPTable_>
template <class STree, class Sequence>
typename CYKnoTrace<CostTable_,DPTable_>::cost_type
CYKnoTrace<CostTable_,DPTable_>::
scan_next(const STree& tree, const Sequence& rna, int from, int to, int pos)
{
  assert(impl_->max_length()<=pcyk_.get_length());
  int tree_size = tree.size();
  bk_.update_position(impl_->cost_table(), rna, from, to);
  if (max_pos_.left<from) {
    max_pos_ = Position();
    max_ = static_cast<cost_type>(0.0);
  }

  if (progress_) progress_->start(0, tree_size);
  
  // run CYK algorithm for each table
  for (int h=0; h<tree_size; ++h) {
    // clear previous results
    for (int j=pos; j<to; ++j) {
      for (int i=j; i>=from && j-i<impl_->max_length(); --i) {
	for (uint s=0; s<tree[h].size(); ++s) {
	  pcyk_(h,s,i,j+1)=static_cast<cell_type>(zero<cost_type>());
	}
      }
    }

    switch(tree[h].type()) {
    case EMPTY:
      impl_->scan_empty(tree, h, rna, from, to, pos, *this);
      break;
    case LSINGLE:
    case RSINGLE:
    case STEM:
    case START:
      impl_->scan_normal(tree, h, rna, from, to, pos, *this);
      break;
    case BRANCH:
      impl_->scan_branch(tree, h, rna, from, to, pos, *this);
      break;
    default:
      assert(!"no such node type");
      break;
    }
    if (progress_) progress_->update();
  }
  if (progress_) progress_->finish();

  if (search_type_==GLOBAL_ALIGNMENT) {
    for (uint s=0; s<tree[tree_size-1].size(); ++s) {
      if (pcyk_(tree_size-1, s, from, to).use() &&
	  pcyk_(tree_size-1, s, from, to).cost() > max_) {
	max_pos_ = Position(tree_size-1, s, from, to);
	max_ = pcyk_(tree_size-1, s, from, to).cost();
      }
    }
  }
  return max_;
}

template < class CostTable_, template <class> class DPTable_ >
template < class CostType, class Position, class STree, class Seq >
void
CYKnoTrace<CostTable_,DPTable_>::
operator()(node_type_t type, const CostType& emit, const Position& parent,
	   const STree& tree, const Seq& rna)
{
  assert(type==EMPTY);
  pcyk_(parent).update(emit);
}

template < class CostTable_, template <class> class DPTable_ >
template < class CostType, class Position, class STree, class Seq >
void
CYKnoTrace<CostTable_,DPTable_>::
operator()(node_type_t type,
	   const CostType& emit, const CostType& state,
	   const Position& parent, const Position& child,
	   const STree& tree, const Seq& rna)
{
  if (child.state==-1) {	// L_END
    CostType cost = bk_.inside(child) * emit * state;
    pcyk_(parent).update(cost);
  } else if (parent.state==-1) { // L_START
    if (pcyk_(child).use()) {
      CostType cost = pcyk_(child).cost() *  state * bk_.outside(parent);
      if (max_ < cost) {
	max_ = cost;
	max_pos_ = child;
      }
    }
  } else if (pcyk_(child).use()) {
    CostType cost = pcyk_(child).cost() * emit * state;
    pcyk_(parent).update(cost);
  }
}

template < class CostTable_, template <class> class DPTable_ >
template < class CostType, class Position, class STree, class Seq >
void
CYKnoTrace<CostTable_,DPTable_>::
operator()(node_type_t type,
	   const CostType& emit, const CostType& state, 
	   const Position& parent,
	   const Position& lchild, const Position& rchild,
	   const STree& tree, const Seq& rna)
{
  assert(type==BRANCH);
  if (pcyk_(lchild).use() && pcyk_(rchild).use()) {
    CostType cost = pcyk_(lchild).cost() * pcyk_(rchild).cost() * emit * state;
    pcyk_(parent).update(cost);
  }
}

// template instantiation for 'cost_t'
#include <string>
#include "costtable.h"
#include "log_value.h"
#include "skeletal.h"
#include "dptable.h"
#include "scantable.h"

template
class CYKnoTrace< CostTable<cost_t>, DPTable >;

template
cost_t
CYKnoTrace< CostTable<cost_t>, DPTable >::
run(const SkeletalTreeS& tree, const RNASequence& rna, int from, int to);

template
class CYKnoTrace< CostTable<cost_t>, ScanTable >;

template
cost_t
CYKnoTrace< CostTable<cost_t>, ScanTable >::
scan_first(const SkeletalTreeS& tree, const RNASequence& rna, int from, int to);

template
cost_t
CYKnoTrace< CostTable<cost_t>, ScanTable >::
scan_next(const SkeletalTreeS& tree, const RNASequence& rna,
	  int from, int to, int pos);

// template instantiation for 'LogValue<short>'
#include "defaultcosttable.h"

template
class CYKnoTrace< DefaultCostTable, DPTable >;

template
LogValue<short>
CYKnoTrace< DefaultCostTable, DPTable >::
run(const SkeletalTreeS& tree, const RNASequence& rna, int from, int to);

#if 0
template
class CYKnoTrace< DefaultCostTable, ScanTable >;

template
LogValue<short>
CYKnoTrace< DefaultCostTable , ScanTable >::
scan_first(const SkeletalTreeS& tree, const RNASequence& rna, int from, int to);
template
LogValue<short>
CYKnoTrace< DefaultCostTable, ScanTable >::
scan_next(const SkeletalTreeS& tree, const RNASequence& rna,
	  int from, int to, int pos);
#endif
