/*
 * $Id: dptable.h 482 2006-04-12 09:00:23Z satoken $
 *
 * PHMMTS -- an implementation of Pair Hidden Markov Models on Tree Structures,
 *           which is based on "Pair Hidden Markov Models on Tree Structures",
 *           Yasubumi Sakakibara, ISMB 2003.
 *
 * Copyright (C) 2003-2006 Kengo Sato
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifndef __INC_DPTABLE_H__
#define __INC_DPTABLE_H__

#include "defines.h"
#include "cyktable.h"
#include <vector>

namespace PHMMTS
{
  template <class T>
  class DPTable
  {
  public:
    typedef T value_type;
    typedef value_type node_type;

  private:
    int height_;
    int from_;
    int length_;
    std::vector<CYKTable<T>*>* table_;

  public:
    DPTable() : height_(0), from_(0), length_(0), table_(NULL) { }
    
    DPTable(int height, int length)
      : height_(height), from_(0), length_(length), table_(NULL)
    {
      create();
    }

    DPTable(int height, int from, int to)
      : height_(height), from_(from), length_(to-from+1), table_(NULL)
    {
      create();
    }

    ~DPTable() { destroy_all(); }

    int get_length() const { return length_; }
    int get_height() const { return height_; }

    void set_size(int height, int length)
    {
      if (height_!=height || length_!=length) {
	destroy_all();
	height_ = height;
	length_ = length;
	create();
      }
    }

    void set_size(int height, int from, int to)
    {
      from_ = from;
      set_size(height, to-from+1);
    }
      
    void create()
    {
      destroy_all();
      table_ = new std::vector<CYKTable<T>*>[height_];
    }

    void create(int h, size_t states)
    {
      if (table_==NULL) create();
      if (table_[h].size()!=states) {
	destroy(h);
	table_[h].resize(states);
	for (uint s=0; s<states; ++s)
	  table_[h][s] = new CYKTable<T>(length_);
      }
    }

    void create(int h, int states, const T& val)
    {
      create(h, states);
      fill(h, val);
    }

    void destroy(int h)
    {
      for (uint s=0; s<table_[h].size(); ++s)
	delete table_[h][s];
      table_[h].clear();
    }

    void destroy_all()
    {
      if (table_!=NULL) {
	for (int h=0; h<height_; ++h)
	  destroy(h);
	delete[] table_;
	table_=NULL;
      }
    }

    void fill(int h, const T& val)
    {
      for (uint s=0; s<table_[h].size(); ++s) {
	table_[h][s]->fill(val);
      }
    }

    void fill(const T& val)
    {
      for (int h=0; h<height_; ++h)
	fill(h, val);
    }

  private:
    void put(int h, int s, int i, int j, const T& val)
    {
      table_[h][s]->put(i-from_, j-from_, val);
    }

    void put(const Position& pos, const T& val)
    {
      table_[pos.height][pos.state]->put(pos.left-from_, pos.right-from_, val);
    }

    const T& get(int h, int s, int i, int j) const
    {
      return table_[h][s]->get(i-from_, j-from_);
    }

    const T& get(const Position& pos) const
    {
      return table_[pos.height][pos.state]->get(pos.left-from_, pos.right-from_);
    }

    T& get(int h, int s, int i, int j)
    {
      return table_[h][s]->get(i-from_, j-from_);
    }

    T& get(const Position& pos)
    {
      return table_[pos.height][pos.state]->get(pos.left-from_, pos.right-from_);
    }

  public:
    const T& operator()(int h, int s, int i, int j) const
    {
      return get(h,s,i,j);
    }

    const T& operator()(const Position& pos) const
    {
      return get(pos);
    }

    T& operator()(int h, int s, int i, int j)
    {
      return get(h,s,i,j);
    }

    T& operator()(const Position& pos)
    {
      return get(pos);
    }
  };
}

#endif // __INC_DPTABLE_H__

// Local Variables:
// mode: C++
// End:
