/*
 * $Id: em_train.cpp 492 2006-08-08 01:29:48Z satoken $
 *
 * PHMMTS -- an implementation of Pair Hidden Markov Models on Tree Structures,
 *           which is based on "Pair Hidden Markov Models on Tree Structures",
 *           Yasubumi Sakakibara, ISMB 2003.
 *
 * Copyright (C) 2003-2006 Kengo Sato
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <iostream>
#include <fstream>
#include <list>
#include <algorithm>
#include "skeletal.h"
#include "multiseq.h"
#include "costtable.h"
#include "log_value.h"
#include "example.h"
#include "dptable.h"
#include "inside.h"
#include "outside.h"
#include "inside_outside.h"

using namespace PHMMTS;

typedef CostTable<float> Expectation;
typedef CostTable<cost_t> LCostTable;

template <class CostTable, class Expl>
class ExpectationFunc : public std::unary_function<const Expl&, void>
{
  const CostTable& cost_table_;
  Expectation& ex_;
  float& log_likelihood_;
  search_type_t search_type_;
  int verbose_;
  size_t cur_;
  size_t stage_;
  size_t num_;

public:
  ExpectationFunc(const CostTable& cost_table, Expectation& ex,
		  search_type_t search_type,
		  float& ll, int verbose=2, size_t stage=1, size_t num=0)
    : cost_table_(cost_table), ex_(ex), log_likelihood_(ll),
      search_type_(search_type),
      verbose_(verbose), cur_(0), stage_(stage), num_(num)
  {
    log_likelihood_ = 0.0;
  }

  void operator()(const Expl& x)
  {
    typename CostTable::cost_type p;
#ifndef HAVE_MPI
    if (verbose_>=2 && num_>0) {
      std::cout << "- Stage: " << stage_ << ", "
		<< ++cur_ << '/' << num_ << ": "
		<< std::flush;
    }
#endif

    Inside<CostTable> inside(cost_table_, search_type_);
    p = inside.run(x.skeletal_tree(), x.unfolded_seq());
    float ll = log(p);
    log_likelihood_ += ll;

    Outside<CostTable> outside(cost_table_, inside, search_type_);
    outside.run(x.skeletal_tree(), x.unfolded_seq());

    InsideOutside<CostTable,Expectation>
      inside_outside(cost_table_, inside, outside, ex_, search_type_);
    inside_outside.run(x.skeletal_tree(), x.unfolded_seq());

#ifndef HAVE_MPI
    if (verbose_>=2 && num_>0)
      std::cout << "LL= " << ll << std::endl;
#endif
  }
};

template < class Expl >
static
void
object_func(const LCostTable& cost_table,
	    float& f, Expectation& ex, search_type_t search_type,
	    const std::list<Expl>& train,
	    int verbose, size_t stage)
{
  // parse each training data and calc expectations
#ifdef HAVE_MPI
  int my_rank = MPI::COMM_WORLD.Get_rank();
  int num_procs = MPI::COMM_WORLD.Get_size();
  float log_likelihood;
  Expectation lex;
  ExpectationFunc<LCostTable, Expl>
    func(cost_table, lex, search_type, log_likelihood,
	 verbose, stage, train.size());
  typename std::list<Expl>::const_iterator x;
  int cnt=0;
  for (x=train.begin(); x!=train.end(); ++x) {
    if (cnt++%num_procs==my_rank) func(*x);
  }
  if (my_rank==0) {
    ex.add(lex);
    for (int t=1; t!=num_procs; ++t) {
      float v;
      MPI::COMM_WORLD.Recv(&v, 1, MPI::FLOAT, t, 0);
      log_likelihood += v;
      lex.recv(t, 1);
      ex.add(lex);
    }
    for (int t=1; t!=num_procs; ++t) {
      MPI::COMM_WORLD.Send(&log_likelihood, 1, MPI::FLOAT, t, 0);
      ex.send(t, 1);
    }
  } else {
    MPI::COMM_WORLD.Send(&log_likelihood, 1, MPI::FLOAT, 0, 0);
    lex.send(0, 1);
    MPI::COMM_WORLD.Recv(&log_likelihood, 1, MPI::FLOAT, 0, 0);
    ex.recv(0, 1);
  }
#else  //  HAVE_MPI
  float log_likelihood;
  ExpectationFunc<LCostTable, Expl>
    func(cost_table, ex, search_type, log_likelihood,
	 verbose, stage, train.size());
  std::for_each(train.begin(), train.end(), func);
#endif	//  HAVE_MPI
  f = log_likelihood;
}

template < class Expl >
LCostTable
em_training(const std::list<Expl>& train, const LCostTable& init,
	    size_t max_it, float eps, int verbose,
	    size_t& it, float& f, search_type_t search_type)
{
  LCostTable cost_table(init);
  float f_prev;
  Expectation ex;

  // the first time
  {
    object_func(cost_table, f, ex, search_type, train, verbose, 1);
    if (verbose>=1)
      std::cout << "== Stage 1: sum of LL= " << f << std::endl;
    f_prev = f;
    cost_table = ex;
    cost_table.normalize();
    ex.clear();
  }

  // iterate until the convergence of the object func has come.
  for (it=1; it<max_it; ++it) {
    float conv_rate;
    object_func(cost_table, f, ex, search_type, train, verbose, it+1);
    conv_rate = std::abs((f-f_prev)/f_prev);
    if (verbose>=1) {
      std::cout << "== Stage " << it+1
		<< ": sum of LL= " << f << ", "
		<< "conv rate= " << conv_rate << std::endl;
    }
    if (conv_rate < eps) break;
    f_prev = f;
    cost_table = ex;
    cost_table.normalize();
    ex.clear();
  }

  return cost_table;
}

#if 0
template
LCostTable
em_training(const std::list<Example>& train, const LCostTable& init,
	    size_t max_it, float eps, int verbose,
	    size_t& it, float& f, bool l_opt);
#endif

template
LCostTable
em_training(const std::list<ExampleS>& train, const LCostTable& init,
	    size_t max_it, float eps, int verbose,
	    size_t& it, float& f, search_type_t search_type);
