/*
 * $Id: example.cpp 511 2007-09-18 06:18:08Z satoken $
 *
 * PHMMTS -- an implementation of Pair Hidden Markov Models on Tree Structures,
 *           which is based on "Pair Hidden Markov Models on Tree Structures",
 *           Yasubumi Sakakibara, ISMB 2003.
 *
 * Copyright (C) 2003-2006 Kengo Sato
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <iostream>
#include <vector>
#include <list>
#include <algorithm>
#include "skeletal_base.h"
#include "skeletal_base.cpp"
#include "skeletal.h"
#include "multiseq.h"
#include "costtable.h"
#include "example.h"

using namespace PHMMTS;

namespace PHMMTS {
template < >
void
SimpleSkeletalTree<RNASequence,
		   SkeletalTreeNodeBase<RNASequence::value_type> >::
parse(const RNASequence& str, const std::string& paren)
{
  struct local {
    static bool is_gap(const RNA& val)
    {
      return val==static_cast<rna_t>(RNASymbol<rna_t>::GAP);
    }
  };
  RNATree< RNASequence >* t =
    RNATree< RNASequence >::parse(str, paren, local::is_gap);
  parse(t);
  t->destroy();
}
};

template
class
SkeletalTreeBase<RNASequence, SkeletalTreeNodeBase<RNASequence::value_type> >;

template
class
SimpleSkeletalTree<RNASequence, SkeletalTreeNodeBase<RNASequence::value_type> >;

template < class Seq >
ExampleBase<Seq>::
ExampleBase(const std::string& p, const Seq& f, const Seq& u)
  : multi_skeletal_(MultiSeq(f, u), p, is_gap),
    skeletal_(f, p), unfolded_(u), upstream_(), downstream_()
{
  typename Seq::iterator new_end;
  new_end = std::remove(unfolded_.begin(), unfolded_.end(),
			static_cast<rna_t>(RNASymbol<rna_t>::GAP));
  unfolded_.erase(new_end, unfolded_.end());
}

template < class Seq >
ExampleBase<Seq>::
ExampleBase(const std::string& p, const Seq& f, const Seq& u,
	    const Seq& upstream, const Seq& downstream)
  : multi_skeletal_(MultiSeq(f, u), p, is_gap),
    skeletal_(f, p), unfolded_(), upstream_(upstream), downstream_(downstream)
{
  typename Seq::iterator new_end;
  unfolded_.resize(upstream_.size()+u.size()+downstream_.size());
  new_end = std::copy(upstream_.begin(), upstream_.end(), unfolded_.begin());
  new_end = std::copy(u.begin(), u.end(), new_end);
  std::copy(downstream_.begin(), downstream_.end(), new_end);
  new_end = std::remove(unfolded_.begin(), unfolded_.end(),
			static_cast<rna_t>(RNASymbol<rna_t>::GAP));
  unfolded_.erase(new_end, unfolded_.end());
}

template < class Seq >
ExampleBase<Seq>::
ExampleBase(const std::string& p, const Seq& f, const Seq& u, bool not_align)
  : skeletal_(f, p), unfolded_(u), upstream_(), downstream_()
{
  typename Seq::iterator new_end;
  new_end = std::remove(unfolded_.begin(), unfolded_.end(),
			static_cast<rna_t>(RNASymbol<rna_t>::GAP));
  unfolded_.erase(new_end, unfolded_.end());
}

//static
template < class Seq >
bool
ExampleBase<Seq>::
is_gap(const typename MultiSeq::value_type& val)
{
  
  typename MultiSeq::value_type::const_iterator x;
  for (x=val.begin(); x!=val.end(); ++x) {
    if (*x!=static_cast<rna_t>(RNASymbol<rna_t>::GAP))
      return false;
  }
  return true;
}

//static
template < class Seq >
match_state_t
ExampleBase<Seq>::
get_state(const MultiSkeletalTreeNode& n)
{
  const rna_t GAP = static_cast<rna_t>(RNASymbol<rna_t>::GAP);
  switch (n.type()) {
  case LSINGLE:
    if (n.left_rna()[0]==GAP)
      return INSERT;
    else if (n.left_rna()[1]==GAP)
      return DELETE;
    else
      return MATCH;
    break;

  case RSINGLE:
    if (n.right_rna()[0]==GAP)
      return INSERT;
    else if (n.right_rna()[1]==GAP)
      return DELETE;
    else
      return MATCH;
    break;

  case STEM:
    if (n.left_rna()[0]==GAP ||	n.right_rna()[0]==GAP)
      return INSERT;
    else if (n.left_rna()[1]==GAP || n.right_rna()[1]==GAP)
      return DELETE;
    else
      return MATCH;
    break;

  case BRANCH:
  case START:
  case EMPTY:
  case NOTUSE:
  default:
    return MATCH;
    break;
  }
}

//static
template < class Seq >
int
ExampleBase<Seq>::
check_ill_stem(const MultiSkeletalTreeNode& n)
{
  const rna_t GAP = static_cast<rna_t>(RNASymbol<rna_t>::GAP);
  if (n.type()!=STEM) return LEGAL;
  int x=0;
  if (n.left_rna()[0]!=GAP) x|=0x8;
  if (n.right_rna()[0]!=GAP) x|=0x4;
  if (n.left_rna()[1]!=GAP) x|=0x2;
  if (n.right_rna()[1]!=GAP) x|=0x1;
  switch (x) {
  case 0xf: // LRLR
  case 0xc: // LR--
  case 0x3: // --LR
    return LEGAL;
    break;
  case 0x7: // -RLR
    return IL;
    break;
  case 0xb: // L-LR
    return IR;
    break;
  case 0xd: // LR-R
    return DL;
    break;
  case 0xe: // LRL-
    return DR;
    break;
  case 0x6: // -RL-
    return ILDR;
    break;
  case 0x9: // L--R
    return DLIR;
    break;
  default:
    assert(!"reach to an unreachable condition");
    return NOT_STEM;
    break;
  }
}

template <class CT, class Seq>
struct TRFunc : public CT
{
  typedef ExampleBase<Seq> Ex;

  std::vector<match_state_t> st;
  bool enable_local_alignment_;
  
  TRFunc(size_t sz, bool enable_local_alignment)
    : CT(), st(sz), enable_local_alignment_(enable_local_alignment)
  {
  }

  void state_count(const typename Ex::MultiSkeletalTree& tree,
		   int from, int to)
  {
    const typename Ex::MultiSkeletalTreeNode& from_node = tree[from];
    const typename Ex::MultiSkeletalTreeNode& to_node = tree[to];

    if (to_node.type()!=STEM ||
	Ex::check_ill_stem(to_node)==Ex::LEGAL) {
      if (from_node.type()!=STEM ||
	  Ex::check_ill_stem(from_node)==Ex::LEGAL) {
	// LEGAL -> LEGAL
	if (to_node.type()==EMPTY) {
	  if (enable_local_alignment_) {
	    switch (Ex::get_state(from_node)) {
	    case MATCH:
	      st[from] = MATCH;
	      CT::process_state_count(st[from], st[to]);
	      break;
	    case INSERT:
	      st[from] = st[to] = L_END;
	      //process_state_count(st[from], st[to]);
	      break;
	    case DELETE:
	      st[from] = st[to] = L_END;
	      //process_state_count(st[from], st[to]);
	      break;
	    default:
	      assert(!"no such state.");
	      break;
	    }
	  } else {
	    st[from] = Ex::get_state(from_node);
	    CT::process_state_count(st[from], st[to]);
	  }
	} else {
	  if (enable_local_alignment_) {
	    st[from] = Ex::get_state(from_node);
	    // extend L_END state
	    if ((st[from]==INSERT || st[from]==DELETE) && st[to]==L_END)
	      st[from]=L_END;
	    else
	      CT::process_state_count(st[from], st[to]);
	  } else {
	    st[from] = Ex::get_state(from_node);
	    CT::process_state_count(st[from], st[to]);
	  }
	}
      } else {
	// ILLEGAL -> LEGAL
	st[from] = INSERT;
	CT::process_state_count(INSERT, st[to]);
      }
    } else {
      if (from_node.type()!=STEM ||
	  Ex::check_ill_stem(from_node)==Ex::LEGAL) {
	// LEGAL -> ILLEGAL
	st[from] = Ex::get_state(from_node);
	CT::process_state_count(st[from], DELETE);
	CT::process_state_count(DELETE, INSERT);
      } else {
	// ILLEGAL -> ILLEGAL
	if (Ex::check_ill_stem(from_node)==Ex::check_ill_stem(to_node)) {
	  st[from] = DELETE;
	  CT::process_state_count(DELETE, DELETE);
	  CT::process_state_count(INSERT, INSERT);
	} else {
	  st[from] = INSERT;
	  CT::process_state_count(INSERT, DELETE);
	  CT::process_state_count(DELETE, INSERT);
	}
      }
    }
  }

  void operator()(const typename Ex::MultiSkeletalTree& skeletal, size_t n) 
  {
    typedef typename Ex::rna_t rna_t;
    const rna_t GAP = static_cast<rna_t>(RNASymbol<rna_t>::GAP);
    switch (skeletal[n].type()) {
    case BRANCH:
      st[n] = MATCH;
      assert(st[skeletal[n].left()]==MATCH);
      assert(st[skeletal[n].right()]==MATCH);
      process_state_count(st[n], st[skeletal[n].left()],
			  st[skeletal[n].right()]);
      break;

    case START:
      state_count(skeletal, n, skeletal[n].left());
      break;

    case STEM:
      state_count(skeletal, n, skeletal[n].left());
      switch (Ex::check_ill_stem(skeletal[n])) {
      case Ex::IL:
	process_emit_count(GAP, skeletal[n].right_rna()[0], GAP, GAP);
	process_emit_count(GAP, GAP,
			   skeletal[n].left_rna()[1],
			   skeletal[n].right_rna()[1]);
	break;
      case Ex::IR:
	process_emit_count(skeletal[n].left_rna()[0], GAP, GAP, GAP);
	process_emit_count(GAP, GAP,
			   skeletal[n].left_rna()[1],
			   skeletal[n].right_rna()[1]);
	break;
      case Ex::DL:
	process_emit_count(skeletal[n].left_rna()[0],
			   skeletal[n].right_rna()[0], GAP, GAP);
	process_emit_count(GAP, GAP,
			   GAP, skeletal[n].right_rna()[1]);
	break;
      case Ex::DR:
	process_emit_count(skeletal[n].left_rna()[0],
			   skeletal[n].right_rna()[0], GAP, GAP);
	process_emit_count(GAP, GAP,
			   skeletal[n].left_rna()[1], GAP);
	break;
      case Ex::ILDR:
	process_emit_count(GAP, skeletal[n].right_rna()[0], 
			   GAP, GAP);
	process_emit_count(GAP, GAP,
			   skeletal[n].left_rna()[1], GAP);
	break;
      case Ex::DLIR:
	process_emit_count(skeletal[n].left_rna()[0], GAP, GAP, GAP);
	process_emit_count(GAP, GAP,
			   GAP, skeletal[n].right_rna()[1]);
	break;
      default:
	if (st[n]==L_END && enable_local_alignment_) {
	  if (skeletal[n].left_rna()[0]!=GAP &&
	      skeletal[n].right_rna()[0]!=GAP)
	    process_emit_count(skeletal[n].left_rna()[0],
			       skeletal[n].right_rna()[0]);
	  else if (skeletal[n].left_rna()[1]!=GAP &&
		   skeletal[n].right_rna()[1]!=GAP) 
	    process_emit_count(skeletal[n].left_rna()[1],
			       skeletal[n].right_rna()[1]);
	  else
	    assert(false);
	} else {
	  process_emit_count(skeletal[n].left_rna()[0],
			     skeletal[n].right_rna()[0],
			     skeletal[n].left_rna()[1],
			     skeletal[n].right_rna()[1]);
	}
	break;
      }
      break;

    case LSINGLE:
      state_count(skeletal, n, skeletal[n].left());
      if (st[n]==L_END && enable_local_alignment_) {
	if (skeletal[n].left_rna()[0]!=GAP)
	  process_emit_count(skeletal[n].left_rna()[0], GAP);
	else if (skeletal[n].left_rna()[1]!=GAP)
	  process_emit_count(skeletal[n].left_rna()[1], GAP);
	else
	  assert(false);
      } else {
	process_emit_count(skeletal[n].left_rna()[0], GAP,
			   skeletal[n].left_rna()[1], GAP);
      }
      break;

    case RSINGLE:
      state_count(skeletal, n, skeletal[n].left());
      if (st[n]==L_END && enable_local_alignment_) {
	if (skeletal[n].right_rna()[0]!=GAP)
	  process_emit_count(GAP, skeletal[n].right_rna()[0]);
	else if (skeletal[n].right_rna()[1]!=GAP)
	  process_emit_count(GAP, skeletal[n].right_rna()[1]);
	else
	  assert(false);
      } else {
	process_emit_count(GAP, skeletal[n].right_rna()[0], 
			   GAP, skeletal[n].right_rna()[1]);
      }
      break;

    case EMPTY:
      st[n] = MATCH;
      break;

    case NOTUSE:
      assert(!"NOTUSE doesn't use.");
      break;
    }
  }
};

template <class T>
struct CountFeature 
{
  CostTable<T>* ct_;
  T val_;

  CountFeature() : ct_(NULL), val_(1.0) { }

  void set_costtable(CostTable<T>* ct) { ct_ = ct; }

  void process_state_count(match_state_t from, match_state_t to)
  {
    ct_->add_state_count(from, to, val_);
  }

  void process_state_count(match_state_t from,
			   match_state_t left, match_state_t right)
  {
    ct_->add_state_count(from, left, right, val_);
  }
  
  template <class RNA>
  void process_emit_count(const RNA& i, const RNA& j,
			  const RNA& k, const RNA& l)
  {
    ct_->add_emit_count(i, j, k, l, val_);
  }

  template <class RNA>
  void process_emit_count(const RNA& i, const RNA& j)
  {
    ct_->add_emit_count(i, val_);
    ct_->add_emit_count(j, val_);
  }
};

template < class Seq >
template < class T >
void
ExampleBase<Seq>::
add_count(CostTable<T>& ct, bool l_opt, const T& val /*= 1.0*/) const
{
  TRFunc< CountFeature<T>, Seq > func(multi_skeletal_.size(), l_opt);
  func.set_costtable(&ct);
  MultiSkeletalTree::traverse(multi_skeletal_, func);
  if (upstream_.size()>0 || downstream_.size()>0) {
    for (uint i=0; i!=upstream_.size(); ++i)
      ct.add_emit_count(upstream_[i], val);
    for (uint i=0; i!=downstream_.size(); ++i)
      ct.add_emit_count(downstream_[i], val);
    ct.add_state_count(L_START, MATCH, val);
  }
}

template <class T>
struct CalcScore
{
  const CostTable<T>* ct_;
  T val_;
  CalcScore() : ct_(NULL), val_(1.0) { }

  void set_costtable(const CostTable<T>* ct) { ct_ = ct; }
  
  void process_state_count(match_state_t from, match_state_t to)
  {
    val_ *= ct_->state_cost(from, to);
#if 0
    std::cout << "state "
	      << from << " " << to << " "
	      << log(ct_->state_cost(from, to)) << std::endl;
#endif
  }

  void process_state_count(match_state_t from,
			   match_state_t left, match_state_t right)
  {
    val_ *= ct_->state_cost(from, left, right);
#if 0
    std::cout << "state "
	      << from << " " << left << " " << right << " "
	      << log(ct_->state_cost(from, left, right)) << std::endl;
#endif
  }

  template <class RNA>
  void process_emit_count(const RNA& i, const RNA& j,
			  const RNA& k, const RNA& l)
  {
    val_ *= ct_->emit_cost(i, j, k, l);
#if 0
    std::cout << "emit "
	      << i << " " << j << " "
	      << k << " " << l << " "
	      << log(ct_->emit_cost(i, j, k, l)) << std::endl;
#endif
  }

  template <class RNA>
  void process_emit_count(const RNA& i, const RNA& j)
  {
    val_ *= ct_->emit_cost(i);
    val_ *= ct_->emit_cost(j);
  }

  const T& value() const { return val_; }
};

template < class Seq >
template < class T >
T
ExampleBase<Seq>::
cost(const CostTable<T>& ct, bool l_opt) const
{
  TRFunc< CalcScore<T>, Seq > func(multi_skeletal_.size(), l_opt);
  func.set_costtable(&ct);
  MultiSkeletalTree::traverse(multi_skeletal_, func);
  T val=func.value();
  if (upstream_.size()>0 || downstream_.size()>0) {
    for (uint i=0; i!=upstream_.size(); ++i)
      val *= ct.emit_cost(upstream_[i]);
    for (uint i=0; i!=downstream_.size(); ++i)
      val *= ct.emit_cost(downstream_[i]);
    val *= ct.state_cost(L_START, MATCH);
  }
  return val;
}

//static
template < class Seq >
size_t
ExampleBase<Seq>::
load_examples(std::istream& in, std::list< ExampleBase<Seq> >& train)
{
  size_t c = 0;
  std::string paren, folded, unfolded;
  while (std::getline(in, paren) &&
	 std::getline(in, folded) &&
	 std::getline(in, unfolded)) {
    while (paren.size() < folded.size()) paren.push_back(' ');
    if (static_cast<int>(folded.find_first_not_of("ACGUacgu.-"))>=0 ||
	static_cast<int>(unfolded.find_first_not_of("ACGUacgu.-"))>=0) {
#if 0
      std::cout << "Alignments including illegal charactors are ignored:\n"
		<< folded << std::endl
		<< unfolded << std::endl;
#endif
      continue;
    }

    Example t(paren, folded, unfolded, true);
    train.push_back(t);
    ++c;
  }
  return c;
}

namespace PHMMTS {
//static
template <>
size_t
ExampleBase<RNASequence>::
load_examples(std::istream& in, std::list< ExampleBase<RNASequence> >& train)
{
  size_t c = 0;
  std::string paren, folded, unfolded;
  while (std::getline(in, paren) &&
	 std::getline(in, folded) &&
	 std::getline(in, unfolded)) {
    while (paren.size() < folded.size()) paren.push_back(' ');
    if (static_cast<int>(folded.find_first_not_of("ACGUacgu.-"))>=0 ||
	static_cast<int>(unfolded.find_first_not_of("ACGUacgu.-"))>=0) {
#if 0
      std::cout << "Alignments including illegal charactors are ignored:\n"
		<< folded << std::endl
		<< unfolded << std::endl;
#endif
      continue;
    }

    RNASequence f, u;
    f.resize(folded.size());
    std::transform(folded.begin(), folded.end(), f.begin(), char2rna);
    u.resize(unfolded.size());
    std::transform(unfolded.begin(), unfolded.end(), u.begin(), char2rna);

    ExampleS t(paren, f, u, true);
    train.push_back(t);
    ++c;
  }
  return c;
}
};

//static
template < class Seq >
template < class CT >
size_t
ExampleBase<Seq>::
load_examples(std::istream& in, std::list< ExampleBase<Seq> >& train,
	      CT& count_feature, bool l_opt)
{
  size_t c = 0;
  std::string paren, folded, unfolded;
  while (std::getline(in, paren) &&
	 std::getline(in, folded) &&
	 std::getline(in, unfolded)) {
    assert(folded.size()==unfolded.size());
    std::string upstream, downstream;
    while (paren.size() < folded.size()) paren.push_back(' ');

    uint p=folded.find_first_of("ACGUacgu.-", 0);
    uint q=folded.find_first_not_of("ACGUacgu.-", p);

    if (static_cast<int>(p)>0) {
      std::string t;
      upstream.assign(unfolded, 0, p);
      t.assign(unfolded, p, q-p);
      downstream.assign(unfolded, q, t.size()-q);
      unfolded.swap(t);

      t.clear();
      t.assign(folded, p, q-p);
      folded.swap(t);

      t.clear();
      t.assign(paren, p, q-p);
      paren.swap(t);
    }

    if (static_cast<int>(folded.find_first_not_of("ACGUacgu.-"))>=0 ||
	static_cast<int>(unfolded.find_first_not_of("ACGUacgu.-"))>=0) {
#if 0
      std::cout << "Alignments including illegal charactors are ignored:\n"
		<< folded << std::endl
		<< unfolded << std::endl;
#endif
      continue;
    }

    Example t(paren, folded, unfolded, upstream, downstream);
    t.add_count(count_feature, l_opt);
    train.push_back(t);
    ++c;
  }
  return c;
}

//static
template <>
template < class CT >
size_t
PHMMTS::ExampleBase<RNASequence>::
load_examples(std::istream& in, std::list< ExampleBase<RNASequence> >& train,
	      CT& count_feature, bool l_opt)
{
  size_t c = 0;
  std::string paren, folded, unfolded;
  while (std::getline(in, paren) &&
	 std::getline(in, folded) &&
	 std::getline(in, unfolded)) {
    assert(folded.size()==unfolded.size());
    std::string upstream, downstream;
    while (paren.size() < folded.size()) paren.push_back(' ');

    uint p=folded.find_first_of("ACGUacgu.-", 0);
    uint q=folded.find_first_not_of("ACGUacgu.-", p);

    if (static_cast<int>(p)>0) {
      std::string t;
      upstream.assign(unfolded, 0, p);
      t.assign(unfolded, p, q-p);
      downstream.assign(unfolded, q, t.size()-q);
      unfolded.swap(t);

      t.clear();
      t.assign(folded, p, q-p);
      folded.swap(t);

      t.clear();
      t.assign(paren, p, q-p);
      paren.swap(t);
    }
    
    if (static_cast<int>(folded.find_first_not_of("ACGUacgu.-"))>=0 ||
	static_cast<int>(unfolded.find_first_not_of("ACGUacgu.-"))>=0) {
#if 0
      std::cout << "Alignments including illegal charactors are ignored:\n"
		<< folded << std::endl
		<< unfolded << std::endl;
#endif
      continue;
    }

    RNASequence f, u, up, down;
    f.resize(folded.size());
    std::transform(folded.begin(), folded.end(), f.begin(), char2rna);
    u.resize(unfolded.size());
    std::transform(unfolded.begin(), unfolded.end(), u.begin(), char2rna);
    up.resize(upstream.size());
    std::transform(upstream.begin(), upstream.end(), up.begin(), char2rna);
    down.resize(downstream.size());
    std::transform(downstream.begin(), downstream.end(), down.begin(), char2rna);

    ExampleS t(paren, f, u, up, down);
    t.add_count(count_feature, l_opt);
    train.push_back(t);
    ++c;
  }
  return c;
}

// template instantiation

#include "log_value.h"

#if 0
template
class ExampleBase<std::string>;

template
cost_t
ExampleBase<std::string>::
cost(const CostTable<cost_t>& ct, bool l_opt) const;

template
void
ExampleBase<std::string>::
add_count(CostTable<cost_t>& ct, bool l_opt,
	  const cost_t& val) const;

//static
template
size_t
ExampleBase<std::string>::
load_examples(std::istream& in, std::list<Example>& train,
	      CostTable<float>& count_feature, bool l_opt);
#endif

template
class ExampleBase< RNASequence >;

template
cost_t
ExampleBase< RNASequence >::
cost(const CostTable<cost_t>& ct, bool l_opt) const;

template
void
ExampleBase< RNASequence >::
add_count(CostTable<cost_t>& ct, bool l_opt,
	  const cost_t& val) const;

//static
template
size_t
ExampleBase<RNASequence>::
load_examples(std::istream& in, std::list<ExampleS>& train,
	      CostTable<float>& count_feature, bool l_opt);
