/*
 * $Id: inside.cpp 482 2006-04-12 09:00:23Z satoken $
 *
 * PHMMTS -- an implementation of Pair Hidden Markov Models on Tree Structures,
 *           which is based on "Pair Hidden Markov Models on Tree Structures",
 *           Yasubumi Sakakibara, ISMB 2003.
 *
 * Copyright (C) 2003-2006 Kengo Sato
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "algo_impl.h"
#include "background.h"
#include "inside.h"

using namespace PHMMTS;

template <class CostTable_, template<class> class DPTable_>
Inside<CostTable_,DPTable_>::
Inside(const CostTable_& cost_table, search_type_t search_type,
       int max_length /*=0*/)
  : inside_(), bk_(), cost_(zero<cost_type>()),
    impl_(NULL), search_type_(search_type)
{
  int l=0;
  switch (search_type_) {
  case GLOBAL_ALIGNMENT:
    break;
  case LOCAL_ALIGNMENT:
    l = Impl::UNFOLDED_LSTART;
    break;
  case LOCAL_STRUCTURE:
    l = (Impl::FOLDED_LSTART | Impl::FOLDED_LEND |
	 Impl::UNFOLDED_LSTART | Impl::UNFOLDED_LEND );
    break;
  default:
    assert(!"no such search type.");
    break;
  }
  impl_ = new Impl(cost_table, l, max_length);
}

template <class CostTable_, template<class> class DPTable_>
Inside<CostTable_,DPTable_>::
~Inside()
{
  delete impl_;
}

template <class CostTable_, template<class> class DPTable_>
template <class STree, class Sequence>
typename Inside<CostTable_,DPTable_>::cost_type
Inside<CostTable_,DPTable_>::
run(const STree& tree, const Sequence& rna, int from, int to)
{
  // initialize CYK tables
  inside_.set_size(tree.size(), from, to);
  bk_.initialize(impl_->cost_table(), tree, rna, from, to);
  cost_ = zero<cost_type>();

  // run CYK algorithm for each table
  for (int h=0; h<(int)tree.size(); ++h) {
    inside_.create(h, tree[h].size(), zero<cost_type>());

    switch(tree[h].type()) {
    case EMPTY:
      impl_->inside_empty(tree, h, rna, from, to, *this);
      break;
    case LSINGLE:
    case RSINGLE:
    case STEM:
    case START:
      impl_->inside_normal(tree, h, rna, from, to, *this);
      break;
    case BRANCH:
      impl_->inside_branch(tree, h, rna, from, to, *this);
      break;
    default:
      assert(!"no such node type");
      break;
    }
  }

  if (search_type_==GLOBAL_ALIGNMENT) {
    for (uint s=0; s<tree[tree.size()-1].size(); ++s) {
      cost_ += inside_(tree.size()-1,s,from,to).cost();
    }
  }
  return cost_;
}

template <class CostTable_, template<class> class DPTable_>
template < class CostType, class Position, class STree, class Seq >
void
Inside<CostTable_,DPTable_>::
operator()(node_type_t type, const CostType& emit, const Position& parent,
	   const STree& tree, const Seq& rna)
{
  assert(type==EMPTY);
  inside_(parent).update(emit);
}

template <class CostTable_, template<class> class DPTable_>
template < class CostType, class Position, class STree, class Seq >
void
Inside<CostTable_,DPTable_>::
operator()(node_type_t type,
	   const CostType& emit, const CostType& state,
	   const Position& parent, const Position& child,
	   const STree& tree, const Seq& rna)
{
  if (child.state==-1) {	// L_END
    CostType cost = bk_.inside(child) * emit * state;
    inside_(parent).update(cost);
  } else if (parent.state==-1) { // L_START
    if (inside_(child).use()) {
      CostType cost = inside_(child).cost() * state * bk_.outside(parent);
      cost_ += cost;
    }
  } else if (inside_(child).use()) {
    CostType cost = inside_(child).cost()*emit*state;
    inside_(parent).update(cost);
  }
}

template <class CostTable_, template<class> class DPTable_>
template < class CostType, class Position, class STree, class Seq >
void
Inside<CostTable_,DPTable_>::
operator()(node_type_t type,
	   const CostType& emit, const CostType& state, 
	   const Position& parent,
	   const Position& lchild, const Position& rchild,
	   const STree& tree, const Seq& rna)
{
  assert(type==BRANCH);
  if (inside_(lchild).use() && inside_(rchild).use()) {
    CostType cost = inside_(lchild).cost()*inside_(rchild).cost()*emit*state;
    inside_(parent).update(cost);
  }
}

// template instantiation for 'cost_t'
#include <string>
#include "costtable.h"
#include "log_value.h"
#include "skeletal.h"
#include "dptable.h"

template
class Inside< CostTable<cost_t>, DPTable >;

#if 0
template
cost_t
Inside< CostTable<cost_t>, DPTable >::
run(const SkeletalTree& tree, const std::string& rna, int from, int to);
#endif

template
cost_t
Inside< CostTable<cost_t>, DPTable >::
run(const SkeletalTreeS& tree, const RNASequence& rna, int from, int to);

// template instantiation for 'LogValue<short>'
#include "defaultcosttable.h"

template
class Inside< DefaultCostTable, DPTable >;

#if 0
template
LogValue<short>
Inside< DefaultCostTable, DPTable >::
run(const SkeletalTree& tree, const std::string& rna, int from, int to);
#endif

template
LogValue<short>
Inside< DefaultCostTable, DPTable >::
run(const SkeletalTreeS& tree, const RNASequence& rna, int from, int to);
