/*
 * $Id: inside_outside.cpp 493 2006-08-08 02:05:10Z satoken $
 *
 * PHMMTS -- an implementation of Pair Hidden Markov Models on Tree Structures,
 *           which is based on "Pair Hidden Markov Models on Tree Structures",
 *           Yasubumi Sakakibara, ISMB 2003.
 *
 * Copyright (C) 2003-2006 Kengo Sato
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "inside_outside.h"
#include "inside.h"
#include "outside.h"
#include "algo_impl.h"
#include "background.h"

using namespace PHMMTS;

template < class CostTable_, class Expectation_,
	   template<class> class DPTable_ >
InsideOutside<CostTable_,Expectation_,DPTable_>::
InsideOutside(const CostTable_& cost_table,
	      const Inside<CostTable_,DPTable_>& in,
	      const Outside<CostTable_,DPTable_>& out,
	      Expectation_& ex, search_type_t search_type,
	      int max_length /*=0*/)
  : in_(in), out_(out), inside_(in.table()), outside_(out.table()),
    bk_(in.background()), total_cost_(in.cost()),
    ex_(ex), impl_(NULL), search_type_(search_type)
{
  int l=0;
  switch (search_type_) {
  case GLOBAL_ALIGNMENT:
    break;
  case LOCAL_ALIGNMENT:
    l = Impl::UNFOLDED_LSTART;
    break;
  case LOCAL_STRUCTURE:
    l = (Impl::FOLDED_LSTART | Impl::FOLDED_LEND |
	 Impl::UNFOLDED_LSTART | Impl::UNFOLDED_LEND );
    break;
  default:
    assert(!"no such search type.");
    break;
  }
  impl_ = new Impl(cost_table, l, max_length);
}

template < class CostTable_, class Expectation_,
	   template<class> class DPTable_ >
InsideOutside<CostTable_,Expectation_,DPTable_>::
~InsideOutside()
{
  delete impl_;
}

/************************************************************************/

template <class CostTable_, class Expectation_,
	  template<class> class DPTable_>
template <class STree, class Sequence>
typename InsideOutside<CostTable_,Expectation_,DPTable_>::cost_type
InsideOutside<CostTable_,Expectation_,DPTable_>::
run(const STree& tree, const Sequence& rna, int from, int to)
{
  // run CYK algorithm for each table
  for (int h=0; h<(int)tree.size(); ++h) {
    switch(tree[h].type()) {
    case EMPTY:
      impl_->inside_empty(tree, h, rna, from, to, *this);
      break;
    case LSINGLE:
    case RSINGLE:
    case STEM:
    case START:
      impl_->inside_normal(tree, h, rna, from, to, *this);
      break;
    case BRANCH:
      impl_->inside_branch(tree, h, rna, from, to, *this);
      break;
    default:
      assert(!"no such node type");
      break;
    }
  }

  return total_cost_;
}

template < class CostTable_, class Expectation_,
	   template<class> class DPTable_ >
template < class CostType, class Position, class STree, class Seq >
void
InsideOutside< CostTable_, Expectation_, DPTable_ >::
operator()(node_type_t type, const CostType& emit, const Position& parent,
	   const STree& tree, const Seq& rna)
{
  assert(type==EMPTY);
  typedef typename Seq::value_type rna_t;
#if 0
  const rna_t GAP = RNASymbol<rna_t>::GAP;
  CostType out = outside_(parent);
  CostType c = out*emit/total_cost_;
  ex_.add_emit_count(parent.state, EMPTY,
		     tree[parent.height], GAP, GAP, c);
#endif
}

template < class CostTable_, class Expectation_,
	   template<class> class DPTable_ >
template < class CostType, class Position, class STree, class Seq >
void
InsideOutside< CostTable_, Expectation_, DPTable_ >::
operator()(node_type_t type,
	   const CostType& emit, const CostType& state,
	   const Position& parent, const Position& child,
	   const STree& tree, const Seq& rna)
{
  typedef typename Seq::value_type rna_t;
  typedef typename STree::node_type::Property Property;
  const rna_t GAP = static_cast<rna_t>(RNASymbol<rna_t>::GAP);
  CostType out, in, c;
  if (child.state==-1) {	// L_END
    out = outside_(parent).cost();
    in = bk_.inside(child);
    c = in*state*emit*out/total_cost_;
    for (int i=child.left; i<child.right; ++i)
      ex_.add_emit_count(rna[i], c);
    const Property& pa = tree[parent.height].property(parent.state);
    ex_.add_state_count(pa.state, L_END, c);
  } else if (parent.state==-1) { // L_START
    out = bk_.outside(parent);
    in = inside_(child).cost();
    c = in*state*emit*out/total_cost_;
    for (int i=0; i<parent.left; ++i)
      ex_.add_emit_count(rna[i], c);
    for (int i=parent.right; i<(int)rna.size(); ++i)
      ex_.add_emit_count(rna[i], c);
    const Property& ch = tree[child.height].property(child.state);
    ex_.add_state_count(L_START, ch.state, c);
  } else {
    out = outside_(parent).cost();
    in = inside_(child).cost();
    c = in*state*emit*out/total_cost_;
    const Property& pa = tree[parent.height].property(parent.state);
    const Property& ch = tree[child.height].property(child.state);
    ex_.add_state_count(pa.state, ch.state, c);
  }

  if (parent.state!=-1) {
    const Property& pa = tree[parent.height].property(parent.state);
    if (pa.f_left || pa.u_left || pa.f_right || pa.u_right)
      ex_.add_emit_count(pa.f_left ? tree[parent.height].left_rna() : GAP,
			 pa.f_right ? tree[parent.height].right_rna() : GAP,
			 pa.u_left ? rna[parent.left] : GAP,
			 pa.u_right ? rna[parent.right-1] : GAP,
			 c);
  }
}
  
template < class CostTable_, class Expectation_,
	   template<class> class DPTable_ >
template < class CostType, class Position, class STree, class Seq >
void
InsideOutside< CostTable_, Expectation_, DPTable_ >::
operator()(node_type_t type,
	   const CostType& emit, const CostType& state, 
	   const Position& parent,
	   const Position& lchild, const Position& rchild,
	   const STree& tree, const Seq& rna)
{
  typedef typename Seq::value_type rna_t;
  assert(type==BRANCH);
  CostType out = outside_(parent).cost();
  CostType l_in = inside_(lchild).cost();
  CostType r_in = inside_(rchild).cost();
  CostType c = emit*state*l_in*r_in*out/total_cost_;
#if 0
  const rna_t GAP = RNASymbol<rna_t>::GAP;
  ex_.add_emit_count(parent.state, BRANCH,
		     tree[parent.height], GAP, GAP, c);
#endif
  typedef typename STree::node_type::Property Property;
  const Property& pa = tree[parent.height].property(parent.state);
  const Property& lc = tree[lchild.height].property(lchild.state);
  const Property& rc = tree[rchild.height].property(rchild.state);
  ex_.add_state_count(pa.state, lc.state, rc.state, c);
}

// template instantiation for 'cost_t'
#include <string>
#include "costtable.h"
#include "log_value.h"
#include "skeletal.h"
#include "dptable.h"

template
class InsideOutside< CostTable<cost_t>,
		     CostTable< float>, DPTable >;

#if 0
template
cost_t
InsideOutside< CostTable<cost_t>,
	       CostTable<float>,
	       DPTable >::
run(const SkeletalTree& tree, const std::string& rna, int from, int to);
#endif

template
cost_t
InsideOutside< CostTable<cost_t>,
	       CostTable<float>,
	       DPTable >::
run(const SkeletalTreeS& tree, const RNASequence& rna, int from, int to);
