/*
 * $Id: main.cpp 511 2007-09-18 06:18:08Z satoken $
 *
 * PHMMTS -- an implementation of Pair Hidden Markov Models on Tree Structures,
 *           which is based on "Pair Hidden Markov Models on Tree Structures",
 *           Yasubumi Sakakibara, ISMB 2003.
 *
 * Copyright (C) 2003-2006 Kengo Sato
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <string>
#include <iostream>
#include <sstream>
#include <fstream>
#include <iomanip>
#include <algorithm>
#include <sys/time.h>
#include "param.h"
#include "cyk.h"
#include "cyk_no_trace.h"
#include "inside.h"
#include "defaultcosttable.h"
#include "costtable.h"
#include "skeletal.h"
#include "progress.h"
#include "dptable.h"
#include "scantable.h"
#include "rna.h"
#include "fasta.h"
#include "timersub.h"

using namespace PHMMTS;

enum {
  OUTPUT_PHMMTS,
  OUTPUT_SCARNA,
  OUTPUT_STOCKHOLM
};

Param::Options opt[] = {
  { Param::BOOL, "l", "local", NULL, "\n\texecute local alignment" },
  { Param::BOOL, "L", "local-structure", NULL,
    "\n\texecute local structure search" },
  { Param::BOOL, "s", "scan", NULL,
    "\n\texecute local alignment with the scan mode" },
  { Param::BOOL, "v", "verbose", NULL, "\n\tverbose outputs" },
  { Param::STRING, "o", "output", NULL,
    "\n\tset the filename to output results" },
  { Param::STRING, "c", "cost-table", NULL,
    "\n\tfilename of cost table to be used" },
  { Param::BOOL, "a", "all-candidates", NULL,
    "\n\tcalculate the sum of the cost for all candicates" },
  { Param::INT, "w", "columns", "0", "\n\tset columns to show results" },
  { Param::BOOL, "n", "no-traceback", NULL,
    "\n\tdo not trace back but calculate cost only" },
  { Param::REAL, NULL, "min", NULL,
    "\n\tminimum cost for display in the scan mode" },
  { Param::INT, "m", "maximum-length", "0",
    "\n\tset the maximum length of alignment" },
  { Param::BOOL, "r", "redundant-output", NULL, "HIDE" },
  { Param::BOOL, NULL, "scarna", NULL, "HIDE" },
  { Param::BOOL, NULL, "stockholm", NULL, "HIDE" },
  {0,0,0,0,0}
};

//static const std::string rna_chars("AaCcGgUuRrYyMmKkSsWwHhBbVvDdNn");
static const std::string rna_chars("AaCcGgUu");
static const char* banner = "\
PHMMTS - an implementation of Pair HMM on Tree Structures\n\
Copyright (C) 2003-2006 Kengo SATO, All rights reserved.\n";

#if 0
bool
load_skeletal_tree(std::istream& is, SkeletalTree& t)
{
  std::string paren, rna;
  if (!std::getline(is, paren)) return false;
  if (!std::getline(is, rna)) return false;
  while (paren.size() < rna.size()) paren.push_back(' ');
  t.parse(rna, paren);
  return true;
}
#endif

bool
load_skeletal_tree(std::istream& is, SkeletalTreeS& t)
{
  std::string paren, rna, buf;
  uint i;
  while (std::getline(is, buf)) {
    switch (buf[0]) {
    case '>':
      i=1;
      while (buf[i]==' ') ++i;
      t.set_name(buf.substr(i));
      break;
    case ' ': case '(': case '.':
      paren=buf;
      break;
    default:
      rna=buf;
      break;
    }
    if (!rna.empty() && !paren.empty()) break;
  }
  if (rna.empty() || paren.empty()) return false;
  
  while (paren.size() < rna.size()) paren.push_back(' ');
  if (paren.size()>rna.size()) paren.resize(rna.size());

  RNASequence s(rna.size());
  std::transform(rna.begin(), rna.end(), s.begin(), char2rna);
  t.parse(s, paren);
  return true;
}

template < class CostType_ >
float
evalue(CostType_ x, float lambda, float con, unsigned int length)
{
  return con * length * exp(-lambda*log(x));
}

template < class CostType_ >
void
output_summary(std::ostream& out, bool success, bool l_opt, bool a_opt,
	       float evalue_lambda, float evalue_const,
	       const std::string& folded_name,
	       const std::string& unfolded_name,
	       const struct timeval& tm,
	       const CostType_& cost, const CostType_& local_cost,
	       const CostType_& total_cost,
	       unsigned int rna_len, unsigned int skeletal_len)
{
  char tm_buf[20];
  sprintf(tm_buf, "%ld.%03lds", tm.tv_sec, tm.tv_usec/1000);

  out << "  folded: " << folded_name << std::endl
      << "unfolded: " << unfolded_name << std::endl;
  if (success) {
    out << "Length: " << rna_len << ' '
	<< "Time: " << tm_buf << ' '
	<< "Score: " << log(cost/local_cost);
    if (a_opt) out << " " << log(total_cost);
    if (l_opt) {
      int prec = out.precision(1);
      out << " E-value: " << std::scientific
	  << evalue(cost, evalue_lambda, evalue_const, rna_len*skeletal_len)
	  << std::fixed;
      out.precision(prec);
    }
    out << std::endl;
  } else {
    out << "Length: " << rna_len << ' '
	<< "Time: " << tm_buf << ' '
	<< "Score: NA" << std::endl;
  }
}

template < class Stream >
Stream&
operator<<(Stream& out, const rna_t& s)
{
  out << static_cast<char>(toupper(PHMMTS::rna2char(s)));
  return out;
}

template < class Stream >
Stream&
operator<<(Stream& out, const RNASequence& seq)
{
  for (size_t i=0; i<seq.size(); ++i)
    out << static_cast<char>(toupper(PHMMTS::rna2char(seq[i])));
  return out;
}

template < class PCYK_, class STree, class Seq >
void
output_detail(const PCYK_& pcyk,
	      const STree& skeletal, const Seq& rna,
	      std::ostream& out, int w_opt, bool all_seq)
{
  typedef typename Seq::value_type rna_t;
  Seq unfolded, folded;
  std::string state_seq, brackets;
  Seq left_folded, left_unfolded, right_folded, right_unfolded;
  std::string left_state, left_brackets, right_state, right_brackets;
  unsigned int i, j, r;
  Position pos(pcyk.pos());
  int len = rna.size();
  const rna_t GAP = static_cast<rna_t>(RNASymbol<rna_t>::GAP);
  skeletal.to_string_outside(left_folded, left_brackets,
			     right_folded, right_brackets, pos.height);

  left_unfolded.assign(&rna[0], &rna[std::max(0, pos.left)]);
  right_unfolded.assign(&rna[std::min(pos.right, len)], &rna[len]);

  if (all_seq) {		// output for global alignment
    unsigned int len;
    pcyk.traceback(skeletal, rna, folded, unfolded, state_seq, brackets);
    while (left_folded.size()<left_unfolded.size()) {
      left_folded.push_back(GAP);
      left_brackets += ' ';
    }
    while (left_folded.size()>left_unfolded.size()) {
      left_unfolded.push_back(GAP);
    }
    left_state.append(left_brackets.size(), ' ');
    while (right_folded.size()<right_unfolded.size()) {
      right_folded.push_back(GAP);
      right_brackets += ' ';
    }
    while (right_folded.size()>right_unfolded.size()) {
      right_unfolded.push_back(GAP);
    }
    right_state.append(right_brackets.size(), ' ');

    state_seq = left_state + state_seq + right_state;
    brackets = left_brackets + brackets + right_brackets;

    Seq xunfolded, xfolded;
    {
      typename Seq::iterator x;
      for (x=left_unfolded.begin(); x!=left_unfolded.end(); ++x)
	xunfolded.push_back(*x);
      for (x=unfolded.begin(); x!=unfolded.end(); ++x)
	xunfolded.push_back(*x);
      for (x=right_unfolded.begin(); x!=right_unfolded.end(); ++x)
	xunfolded.push_back(*x);
    }
    {
      typename Seq::iterator x;
      for (x=left_folded.begin(); x!=left_folded.end(); ++x)
	xfolded.push_back(*x);
      for (x=folded.begin(); x!=folded.end(); ++x)
	xfolded.push_back(*x);
      for (x=right_folded.begin(); x!=right_folded.end(); ++x)
	xfolded.push_back(*x);
    }

    len = brackets.length();
    r = w_opt>0 ? w_opt : len;
    for (i=0; i<len; i+=r) {
      out << "           ";
      for (j=0; j<r; ++j) out << ((i+j+1)%10==0 ? '+' : '-');
      out << " (" << i+j << ")" << std::endl;
      
      out << "           ";
      for (j=0; j<r && i+j<len; ++j) out << brackets[i+j];
      out << std::endl;
	      
      out << "  folded:  ";
      for (j=0; j<r && i+j<len; ++j) out << xfolded[i+j];
      out << std::endl;
      
      out << "           ";
      for (j=0; j<r && i+j<len; ++j) out << state_seq[i+j];
      out << std::endl;
      
      out << "unfolded:  ";
      for (j=0; j<r && i+j<len; ++j) out << xunfolded[i+j];
      out << std::endl;
    }
  } else {			// output for local alignment
    int folded_start, unfolded_start;
    int folded_pos, unfolded_pos;
    unsigned int len;
    pcyk.traceback(skeletal, rna, folded, unfolded, state_seq, brackets);
    folded_start = left_folded.size();
    unfolded_start = pos.left;
    len = brackets.length();
    r = w_opt>0 ? w_opt : len;
    folded_pos = folded_start;
    unfolded_pos = unfolded_start;
    float f_w = std::log10(static_cast<float>(folded_start+len));
    float u_w = std::log10(static_cast<float>(unfolded_start+len));
    int w = static_cast<int>(std::max(f_w, u_w))+1;
    for (i=0; i<len; i+=r) {
      out << "           "; for (int k=0; k<w+2; ++k) std::cout << ' ';
      for (j=0; j<r; ++j) out << ((i+j+1)%10==0 ? '+' : '-');
      out << std::endl;
		
      out << "           "; for (int k=0; k<w+2; ++k) std::cout << ' ';
      for (j=0; j<r && i+j<len; ++j) out << brackets[i+j];
      out << std::endl;
	      
      out << "  folded: ("
	   << std::setw(w) << folded_pos << ") ";
      for (j=0; j<r && i+j<len; ++j) {
	out << folded[i+j];
	if (folded[i+j]!=GAP) folded_pos++;
      }
      out << " (" << std::setw(w) << folded_pos << ")" << std::endl;
	      
      out << "           "; for (int k=0; k<w+2; ++k) std::cout << ' ';
      for (j=0; j<r && i+j<len; ++j) out << state_seq[i+j];
      out << std::endl;
	      
      out << "unfolded: ("
	   << std::setw(w) << unfolded_pos << ") ";
      for (j=0; j<r && i+j<len; ++j) {
	out << unfolded[i+j];
	if (unfolded[i+j]!=GAP) unfolded_pos++;
      }
      out << " (" << std::setw(w) << unfolded_pos << ")" << std::endl;
    }
  }
}

template < class PCYK_, class STree, class Seq, class CostType_ >
void
output_stockholm(std::ostream& out, const PCYK_& pcyk,
		 const STree& skeletal, const Seq& rna,
		 int rna_no, const std::string& unfolded_name,
		 const CostType_& cost, const CostType_& local_cost)
{
  typedef typename Seq::value_type rna_t;
  const rna_t GAP = static_cast<rna_t>(RNASymbol<rna_t>::GAP);
  
  typename STree::Sequence skel_seq;
  skeletal.to_string(skel_seq);
  std::string state_seq, brackets;

  Seq unfolded, folded;
  pcyk.traceback(skeletal, rna, folded, unfolded, state_seq, brackets);
  for (uint i=0; i<brackets.size(); ++i)
    if (brackets[i]==' ') brackets[i]='.';

  Seq left_folded, left_unfolded, right_folded, right_unfolded;
  std::string left_brackets, right_brackets;
  Position pos(pcyk.pos());
  int len = rna.size();
  skeletal.to_string_outside(left_folded, left_brackets,
			     right_folded, right_brackets, pos.height);
  left_unfolded.assign(&rna[0], &rna[std::max(0, pos.left)]);
  right_unfolded.assign(&rna[std::min(pos.right, len)], &rna[len]);

  uint folded_start, folded_end;
  folded_start = left_folded.size();
  folded_end=folded_start;
  for (uint i=0; i<folded.size(); ++i)
    if (folded[i]!=GAP) folded_end++;
  
  uint unfolded_start, unfolded_end;
  unfolded_start = pos.left;
  unfolded_end=unfolded_start;
  for (uint i=0; i<unfolded.size(); ++i)
    if (unfolded[i]!=GAP) unfolded_end++;

  std::stringstream fold_ss, unfold_ss, cons_ss, sc_ss;
  fold_ss << skeletal.name() << "/" << folded_start << "-" << folded_end;
  unfold_ss << unfolded_name << "/" << unfolded_start << "-" << unfolded_end;
  cons_ss << "#=GC SS_cons";
  sc_ss << "#=GC SC";
  while (fold_ss.str().size()<unfold_ss.str().size()) fold_ss << " ";
  while (fold_ss.str().size()>unfold_ss.str().size()) unfold_ss << " ";
  while (fold_ss.str().size()>cons_ss.str().size()) cons_ss << " ";
  while (fold_ss.str().size()>sc_ss.str().size()) sc_ss << " ";

  out << "# STOCKHOLM 1.0" << std::endl
      << fold_ss.str() << "\t" << folded << std::endl
      << unfold_ss.str() << "\t" << unfolded << std::endl
      << cons_ss.str() << "\t" << brackets << std::endl
      << sc_ss.str() << "\t" << log(cost/local_cost) << std::endl
      << "//" << std::endl;
}

template < class PCYK_, class STree, class Seq, class CostType_ >
void
output_scarna_xml(std::ostream& out, const PCYK_& pcyk,
		  const STree& skeletal, const Seq& rna,
		  int rna_no, const std::string& unfolded_name,
		  const CostType_& cost, const CostType_& local_cost)
{
  typedef typename Seq::value_type rna_t;
  const rna_t GAP = static_cast<rna_t>(RNASymbol<rna_t>::GAP);

  typename STree::Sequence skel_seq;
  skeletal.to_string(skel_seq);
  std::string state_seq, brackets;

  Seq unfolded, folded;
  pcyk.traceback(skeletal, rna, folded, unfolded, state_seq, brackets);
  for (uint i=0; i<brackets.size(); ++i)
    if (brackets[i]==' ') brackets[i]='.';

  Seq left_folded, left_unfolded, right_folded, right_unfolded;
  std::string left_brackets, right_brackets;
  Position pos(pcyk.pos());
  int len = rna.size();
  skeletal.to_string_outside(left_folded, left_brackets,
			     right_folded, right_brackets, pos.height);
  left_unfolded.assign(&rna[0], &rna[std::max(0, pos.left)]);
  right_unfolded.assign(&rna[std::min(pos.right, len)], &rna[len]);

  uint folded_start, folded_end;
  folded_start = left_folded.size();
  folded_end=folded_start;
  for (uint i=0; i<folded.size(); ++i)
    if (folded[i]!=GAP) folded_end++;
  
  uint unfolded_start, unfolded_end;
  unfolded_start = pos.left;
  unfolded_end=unfolded_start;
  for (uint i=0; i<unfolded.size(); ++i)
    if (unfolded[i]!=GAP) unfolded_end++;
  
  out << "<hit value=\"" << rna_no << "\">" << std::endl
      << "<db_name>" << std::endl
      << unfolded_name << std::endl
      << "</db_name>" << std::endl
      << "<query_name>" << std::endl
      << skeletal.name() << std::endl
      << "</query_name>" << std::endl
#if 0
      << "<scs_score>" << std::endl
      << 0.0 << std::end
      << "</scs_score>" << std::endl
#endif
      << "<rna_score>" << std::endl
      << log(cost/local_cost) << std::endl
      << "</rna_score>" << std::endl
      << "<db_seq>" << std::endl
      << rna << std::endl
      << "</db_seq>" << std::endl
      << "<query_seq>" << std::endl
      << skel_seq << std::endl
      << "</query_seq>" << std::endl
      << "<db_len>" << std::endl
      << rna.size() << std::endl
      << "</db_len>" << std::endl
      << "<query_len>" << std::endl
      << skel_seq.size() << std::endl
      << "</query_len>" << std::endl
      << "<db_match_range>" << std::endl
      << "<number1 value=\"" << unfolded_start << "\" />" << std::endl
      << "<number2 value=\"" << unfolded_end << "\" />" << std::endl
      << "</db_match_range>" << std::endl
      << "<query_match_range>" << std::endl
      << "<number3 value=\"" << folded_start << "\" />" << std::endl
      << "<number4 value=\"" << folded_end << "\" />" << std::endl
      << "</query_match_range>" << std::endl
      << "<db_match_sequence>" << std::endl
      << unfolded << std::endl
      << "</db_match_sequence>" << std::endl
      << "<query_match_sequence>" << std::endl
      << folded << std::endl
      << "</query_match_sequence>" << std::endl
      << "<stockholm>" << std::endl
      << brackets << std::endl
      << "</stockholm>" << std::endl
      << "</hit>" << std::endl;
}

template < class CostTable_, class STree, class Seq >
void
do_phmmts(std::ostream& out,
	  const STree& skeletal, const Seq& rna,
	  const CostTable_& cost_table, ProgressBar* progress,
	  int rna_no,  const std::string& name, search_type_t search_type,
	  bool n_opt, bool r_opt, int w_opt, int m_opt, bool a_opt,
	  uint output_format)
{
  typedef typename CostTable_::cost_type cost_type;
  bool l_opt=search_type!=GLOBAL_ALIGNMENT;
  if (!n_opt) {
    cost_type cost, total_cost, local_cost;
    CYK<CostTable_> pcyk(cost_table, search_type, progress, m_opt);
    struct timeval start_tm, finish_tm, calc_tm;

    gettimeofday(&start_tm, NULL);
    if (a_opt) {
      cost=pcyk.run(skeletal, rna);
      Inside<CostTable_> inside(cost_table, search_type, m_opt);
      total_cost=inside.run(skeletal, rna);
    } else {
      cost=pcyk.run(skeletal, rna);
    }
    if (l_opt) local_cost=pcyk.local_cost(skeletal);

    gettimeofday(&finish_tm, NULL);
    timersub(&finish_tm, &start_tm, &calc_tm);
    switch (output_format) {
    default:
    case OUTPUT_PHMMTS:
      output_summary(out, cost>zero<cost_type>(),
		     l_opt && cost_table.has_evalue_parameters(), a_opt,
		     cost_table.evalue_lambda(), cost_table.evalue_const(),
		     skeletal.name(), name,
		     calc_tm, cost, local_cost, total_cost,
		     rna.size(), skeletal.to_string().size());
      if (cost>zero<cost_type>())
	output_detail(pcyk, skeletal, rna, out, w_opt, !l_opt || r_opt);
      out << "///" << std::endl;
      break;
    case OUTPUT_SCARNA:
      output_scarna_xml(out, pcyk, skeletal, rna, rna_no, name,
			cost, local_cost);
      break;
    case OUTPUT_STOCKHOLM:
      output_stockholm(out, pcyk, skeletal, rna, rna_no, name,
		       cost, local_cost);
      break;
    }
  } else {
    cost_type cost;
    CYKnoTrace<CostTable_> pcyk(cost_table, search_type, progress, m_opt);
    struct timeval start_tm, finish_tm, calc_tm;

    gettimeofday(&start_tm, NULL);
    cost=pcyk.run(skeletal, rna);
    gettimeofday(&finish_tm, NULL);
    timersub(&finish_tm, &start_tm, &calc_tm);

    output_summary(out, cost>zero<cost_type>(),
		   l_opt && cost_table.has_evalue_parameters(), false,
		   cost_table.evalue_lambda(), cost_table.evalue_const(),
		   skeletal.name(), name, calc_tm, cost,
		   static_cast<cost_type>(0.0), cost,
		   rna.size(), skeletal.to_string().size());
  }
}

template < class CostTable_, class STree, class Seq >
void
do_scan(std::ostream& out, const STree& skeletal, const Seq& rna,
	const CostTable_& cost_table, search_type_t search_type,
	bool n_opt, int w_opt, int m_opt, double min)
{
  typedef typename CostTable_::cost_type cost_type;
  cost_type cost;
  Position pos;
  Position max_pos;
  if (m_opt==0) {
    m_opt = std::min(static_cast<int>(skeletal.to_string().size()*1.2),
		     static_cast<int>(rna.size()));
  }

#ifndef REAL_SCAN
  CYK<CostTable_,ScanTable> pcyk(cost_table, search_type, NULL, m_opt);
#else
  CYK<CostTable_,DPTable> pcyk(cost_table, search_type, NULL, m_opt);
#endif
  Seq unfolded, folded;
  std::string state_seq, brackets;
  
  size_t from=0;
  size_t to=from+m_opt;
#ifndef REAL_SCAN
  cost=pcyk.scan_first(skeletal, rna, from, to);
#else
  cost=pcyk.run(skeletal, rna, from, to);
#endif
  pos=pcyk.pos();
  if (pos.valid() && 
      (pos.right!=static_cast<int>(to) ||
       pos.left==static_cast<int>(from))) {
    max_pos = pos;
    if (log(cost/pcyk.local_cost(skeletal))>min) {
      out << "pos=(" << pos.left << "," << pos.right << ") "
	  << "cost=" << log(cost/pcyk.local_cost(skeletal))
	  << std::endl;
      if (!n_opt) {
	output_detail(pcyk, skeletal, rna, out, w_opt, false);
	std::cout << "///" << std::endl;
      }
    }
  }

  from++; to++;
  while (to<rna.size()) {

    if (!pos.valid()) {
      from=std::min(rna.size(), to);
      to=std::min(rna.size(), from+m_opt);
#ifndef REAL_SCAN
      cost=pcyk.scan_first(skeletal, rna, from, to);
#else
      cost=pcyk.run(skeletal, rna, from, to);
#endif
    } else if (pos.left<static_cast<int>(from)) {
      from=pos.left+1;
      to=from+m_opt;
#ifndef REAL_SCAN
      cost=pcyk.scan_next(skeletal, rna, from, to, pos.right);
#else
      cost=pcyk.run(skeletal, rna, from, to);
#endif
    } else {
#ifndef REAL_SCAN
      cost=pcyk.scan_next(skeletal, rna, from, to, to);
#else
      cost=pcyk.run(skeletal, rna, from, to);
#endif
      cost=pcyk.scan_next(skeletal, rna, from, to, to);
    }
    Position ppos=pcyk.pos();
    if (ppos.valid() && ppos!=max_pos &&
	(ppos.right!=static_cast<int>(to) ||
	 ppos.left==static_cast<int>(from))) {
      max_pos=ppos;
      if (log(cost/pcyk.local_cost(skeletal))>min) {
	out << "pos=(" << ppos.left << "," << ppos.right << ") "
	    << "cost=" << log(cost/pcyk.local_cost(skeletal))
	    << std::endl;
	if (!n_opt) {
	  output_detail(pcyk, skeletal, rna, out, w_opt, false);
	  std::cout << "///" << std::endl;
	}
      }
    }
    pos=ppos;
    from++; to++;
  }
}

class InputSequence {
  std::istream& is_;
  Fasta* fasta_;
  RNASequence seq_;
  std::string name_;
public:
  InputSequence(std::istream& is)
    : is_(is), fasta_(NULL), seq_(), name_("unknown")
  {
    char c;
    is_.get(c);
    is_.unget();
    if (c=='>') fasta_ = new Fasta(is_);
  }

  ~InputSequence()
  {
    if (fasta_) delete fasta_;
  }

  const RNASequence& seq() const { return seq_; }

  const std::string& name() const { return name_; }

  bool is_eof() const
  {
    if (fasta_) return fasta_->IsEOF();
    else return is_.eof();
  }

  bool get_next()
  {
    if (fasta_) {
      if (!fasta_->IsEOF() && fasta_->GetNextSeq()) {
	fasta_->ToUpper();
	const std::string& s(fasta_->Seq());
	seq_.clear();
	seq_.resize(s.size());
	std::transform(s.begin(), s.end(), seq_.begin(), char2rna);
	name_=fasta_->Name();
	return true;
      } else {
	return false;
      }
    } else {
      std::string buf, rna;
      seq_.clear();
      while (std::getline(is_, buf)) {
	for (uint i=0; i<buf.length(); ++i) {
	  if (static_cast<int>(rna_chars.find(buf[i]))>=0) {
	    rna.push_back(toupper(buf[i]));
	  }
	}
	if (!rna.empty()) {
	  seq_.resize(rna.size());
	  std::transform(rna.begin(), rna.end(), seq_.begin(), char2rna);
	  return true;
	}
      }
      return false;
    }
  }
};

int
main(int argc, const char** argv)
{
  // parse options
  Param param(opt, "skeletal-tree RNA-sequence", banner);
  if (!param.parse(argc, argv)) return 1;
  bool n_opt = param.get_bool("no-traceback");
  bool l_opt = param.get_bool("local");
  bool v_opt = param.get_bool("verbose");
  int m_opt = param.get_int("maximum-length");
  int w_opt = param.get_int("columns");
  bool r_opt = param.get_bool("redundant-output");
  std::string o_opt = param.get_string("output");
  std::string c_opt = param.get_string("cost-table");
  bool a_opt = param.get_bool("all-candidates");
  bool s_opt = param.get_bool("scan");
  bool output_scarna_format = param.get_bool("scarna");
  bool output_stockholm_format = param.get_bool("stockholm");
  uint output_format = OUTPUT_PHMMTS;
  if (output_scarna_format) output_format = OUTPUT_SCARNA;
  if (output_stockholm_format) output_format = OUTPUT_STOCKHOLM;
  bool unfolded_stdin = false;
  bool all_stdin = false;
  std::istream* is = NULL;
  std::ostream* os = NULL;
  ProgressBar* progress = NULL;
  float min_opt = std::log(0.0);
  if (param.find("min")) min_opt = param.get_real("min");

  // search mode
  search_type_t search_type = GLOBAL_ALIGNMENT;
  if (l_opt || s_opt) search_type = LOCAL_ALIGNMENT;
  if (param.get_bool("local-structure")) search_type = LOCAL_STRUCTURE;

  if (argc==0) { all_stdin=true; unfolded_stdin=true; }
  else if (argc==1) { unfolded_stdin=true; }

  // initialize the cost table
  DefaultCostTable* default_cost_table = NULL;
  CostTable<cost_t>* cost_table = NULL;
  if (!c_opt.empty()) {
    std::ifstream is(c_opt.c_str());
    if (!is.is_open()) {
      perror(c_opt.c_str());
      return 1;
    }
    std::string header;
    if (std::getline(is, header)) {
      if (header=="# CostTable for PHMMTS (short)") {
	default_cost_table = new DefaultCostTable();
	default_cost_table->restore(is);
      } else if (header=="# CostTable for PHMMTS") {
	cost_table = new CostTable<cost_t>;
	cost_table->restore(is);
      } else {
	std::cout << "This format is not supported: "
		  << c_opt << std::endl;
	return 1;
      }
    }
  } else {
    default_cost_table = new DefaultCostTable();
    set_default_cost_table(*default_cost_table);
  }

  // load folded sequence
  SkeletalTreeS skeletal;
  if (!all_stdin) {
    std::ifstream is(argv[0]);
    if (!is.is_open()) {
      perror(argv[0]);
      return 1;
    }
    load_skeletal_tree(is, skeletal);
  }

  // open the output stream
  if (!o_opt.empty()) {
    os = new std::ofstream(o_opt.c_str());
    if (!((std::ofstream*)os)->is_open()) {
      perror(o_opt.c_str());
      return 1;
    }
  } else {
    os = &std::cout;
  }

  // open the input stream for unfolded sequences
  if (unfolded_stdin) {
    is = &std::cin;
  } else {
    is = new std::ifstream(argv[1]);
    if (!((std::ifstream*)is)->is_open()) {
      perror(argv[1]);
      return 1;
    }
  }

  // prepare prgress bar
  if (v_opt) progress = new ProgressBar(w_opt>0 ? w_opt : 80);

  // read an RNA sequence
  int rna_no = 0;
  InputSequence input(*is);
  if (output_scarna_format) *os << "<results>" << std::endl;
  while (true) {
    if (all_stdin) {
      skeletal.clear();
      load_skeletal_tree(*is, skeletal);
    }
    
    if (!input.get_next()) break;
    const RNASequence& rna(input.seq());
    if (rna.empty()) continue;
    rna_no++;

    // execute PHMMTS
    if (default_cost_table) {
      if (!s_opt)
	do_phmmts(*os, skeletal, rna, *default_cost_table, progress,
		  rna_no, input.name(), search_type, n_opt, r_opt, w_opt,
		  m_opt, a_opt, output_format);
      else
	do_scan(*os, skeletal, rna, *default_cost_table, search_type, 
		n_opt, w_opt, m_opt, min_opt);
    } else if (cost_table) {
      if (!s_opt)
	do_phmmts(*os, skeletal, rna, *cost_table, progress,
		  rna_no, input.name(), search_type, n_opt, r_opt, w_opt,
		  m_opt, a_opt, output_format);
      else
	do_scan(*os, skeletal, rna, *cost_table, search_type, 
		n_opt, w_opt, m_opt, min_opt);
    } else {
      std::cerr << "No CostTable is available!" << std::endl;
      return 1;
    }
  }
  if (output_scarna_format) *os << "</results>" << std::endl;

  // close the input stream
  if (is!=NULL && is!=&std::cin) {
    ((std::ifstream*)is)->close();
    delete is;
  }

  // close the output stream
  if (os!=NULL && os!=&std::cout) {
    ((std::ofstream*)os)->close();
    delete os;
  }

  if (progress) delete progress;
  if (default_cost_table) delete default_cost_table;
  if (cost_table) delete cost_table;

  return 0;
}
