/*
 * $Id: outside.cpp 482 2006-04-12 09:00:23Z satoken $
 *
 * PHMMTS -- an implementation of Pair Hidden Markov Models on Tree Structures,
 *           which is based on "Pair Hidden Markov Models on Tree Structures",
 *           Yasubumi Sakakibara, ISMB 2003.
 *
 * Copyright (C) 2003-2006 Kengo Sato
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "outside.h"
#include "inside.h"
#include "algo_impl.h"
#include <vector>

using namespace PHMMTS;

template <class CostTable_, template <class> class DPTable_>
Outside<CostTable_,DPTable_>::
Outside(const CostTable_& cost_table,
	const Inside<CostTable_,DPTable_>& in,
	search_type_t search_type, int max_length /*=0*/)
  : in_(in), inside_(in.table()), outside_(), bk_(in.background()), seq_out_(),
    impl_(NULL), search_type_(search_type)
{
  int l=0;
  switch (search_type_) {
  case GLOBAL_ALIGNMENT:
    break;
  case LOCAL_ALIGNMENT:
    l = Impl::UNFOLDED_LSTART;
    break;
  case LOCAL_STRUCTURE:
    l = (Impl::FOLDED_LSTART | Impl::FOLDED_LEND |
	 Impl::UNFOLDED_LSTART | Impl::UNFOLDED_LEND );
    break;
  default:
    assert(!"no such search type.");
    break;
  }
  impl_ = new Impl(cost_table, l, max_length);
}

template <class CostTable_, template <class> class DPTable_>
Outside<CostTable_,DPTable_>::
~Outside()
{
  delete impl_;
}

template <class CostTable_, template <class> class DPTable_>
template <class STree, class Sequence>
typename Outside<CostTable_,DPTable_>::cost_type
Outside<CostTable_,DPTable_>::
run(const STree& tree, const Sequence& rna, int from, int to)
{
  // initialize CYK tables
  outside_.set_size(tree.size(), from, to);
  seq_out_.resize(to-from);
  std::fill(seq_out_.begin(), seq_out_.end(), static_cast<cost_type>(0.0));

  // run CYK algorithm for each table
  for (uint h=tree.size()-1; ; --h) {
    outside_.create(h, tree[h].size(), static_cast<cost_type>(0.0));
    if (h==tree.size()-1 && search_type_==GLOBAL_ALIGNMENT) {
      for (uint s=0; s<tree[h].size(); ++s) {
	outside_(h,s,from,to)=static_cast<cost_type>(1.0);
      }
    }
    if (h==0) break;
  }

  for (uint h=tree.size()-1; ; --h) {
    switch(tree[h].type()) {
    case EMPTY:
      impl_->outside_empty(tree, h, rna, from, to, *this);
      break;
    case LSINGLE:
    case RSINGLE:
    case STEM:
    case START:
      impl_->outside_normal(tree, h, rna, from, to, *this);
      break;
    case BRANCH:
      impl_->outside_branch(tree, h, rna, from, to, *this);
      break;
    default:
      assert(!"no such node type");
      break;
    }
    if (h==0) break;
  }

#if 0
#ifndef NDEBUG
  for (unsigned int i=0; i<seq_out_.size(); ++i) {
    assert(fabs((in_.cost()-seq_out_[i])/seq_out_[i])<1e-05);
  }
#endif
#endif
  return in_.cost();
}

template <class CostTable_, template <class> class DPTable_>
template < class CostType, class Position, class STree, class Seq >
void
Outside<CostTable_,DPTable_>::
operator()(node_type_t type, const CostType& emit, const Position& parent,
	   const STree& tree, const Seq& rna)
{
  assert(type==EMPTY);
}

template <class CostTable_, template <class> class DPTable_>
template < class CostType, class Position, class STree, class Seq >
void
Outside<CostTable_,DPTable_>::
operator()(node_type_t type,
	   const CostType& emit, const CostType& state,
	   const Position& parent, const Position& child,
	   const STree& tree, const Seq& rna)
{
  if (parent.state==-1) {	// L_START
    CostType cost = bk_.outside(parent) * state;
    outside_(child).update(cost);
    for (int i=0; i<parent.left; ++i)
      seq_out_[i] += cost;
    for (int i=parent.right; i<outside_.get_length()-1; ++i)
      seq_out_[i] += cost;
  } else {
    if (child.state==-1) {	// L_END
      if (outside_(parent).use()) {
	CostType cost =
	  outside_(parent).cost() * state * emit * bk_.inside(child);
	for (int i=child.left; i<child.right; ++i)
	  seq_out_[i] += cost;
      }
    } else {
      if (outside_(parent).use()) {
	CostType cost = outside_(parent).cost()*emit*state;
	outside_(child).update(cost);
      }
    }

    if (parent.state==MATCH || parent.state==INSERT) {
      if (child.state!=-1) {
	if (inside_(child).use() && outside_(parent).use()) {
	  CostType in = inside_(child).cost();
	  CostType c = emit*state*outside_(parent).cost()*in;
	  switch (type) {
	  case LSINGLE:
	    seq_out_[parent.left] += c;
	    break;
	  case RSINGLE:
	    seq_out_[parent.right-1] += c;
	    break;
	  case STEM:
	    seq_out_[parent.left] += c;
	    seq_out_[parent.right-1] += c;
	    break;
	  default:
	    break;
	  }
	}
      } else {
	if (outside_(parent).use()) {
	  CostType in = bk_.inside(child);
	  CostType c = emit*state*outside_(parent).cost()*in;
	  switch (type) {
	  case LSINGLE:
	    seq_out_[parent.left] += c;
	    break;
	  case RSINGLE:
	    seq_out_[parent.right-1] += c;
	    break;
	  case STEM:
	    seq_out_[parent.left] += c;
	    seq_out_[parent.right-1] += c;
	    break;
	  default:
	    break;
	  }
	}
      }
    }
  }
}

template <class CostTable_, template <class> class DPTable_>
template < class CostType, class Position, class STree, class Seq >
void
Outside<CostTable_,DPTable_>::
operator()(node_type_t type,
	   const CostType& emit, const CostType& state, 
	   const Position& parent,
	   const Position& lchild, const Position& rchild,
	   const STree& tree, const Seq& rna)
{
  assert(type==BRANCH);
  if (outside_(parent).use() &&
      inside_(lchild).use() && inside_(rchild).use()) {
    CostType cost = outside_(parent).cost()*state*emit;
    CostType l_in = inside_(lchild).cost();
    CostType r_in = inside_(rchild).cost();
    outside_(lchild).update(cost*r_in);
    outside_(rchild).update(cost*l_in);
  }
}

// template instantiation for 'cost_t'
#include <string>
#include "costtable.h"
#include "log_value.h"
#include "skeletal.h"
#include "dptable.h"

template
class Outside< CostTable<cost_t>, DPTable >;

#if 0
template
cost_t
Outside< CostTable<cost_t>, DPTable >::
run(const SkeletalTree& tree, const std::string& rna, int from, int to);
#endif

template
cost_t
Outside< CostTable<cost_t>, DPTable >::
run(const SkeletalTreeS& tree, const RNASequence& rna, int from, int to);

