/*
 * $Id: param.cpp 482 2006-04-12 09:00:23Z satoken $
 *
 * PHMMTS -- an implementation of Pair Hidden Markov Models on Tree Structures,
 *           which is based on "Pair Hidden Markov Models on Tree Structures",
 *           Yasubumi Sakakibara, ISMB 2003.
 *
 * Copyright (C) 2003-2006 Kengo Sato
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <cstdio>
#include <cstdlib>
#include <cstring>
#include "param.h"

Param::Param(const Options* opt,
	     const char* more_opt /*=NULL*/,
	     const char* banner /*=NULL*/)
  : app_(NULL), more_opt_(more_opt), banner_(banner),
    opt_(opt), param_(), cmd_line_()
    
{
  const Options* x;
  for (x=opt_; x->type!=NONE; ++x) {
    if (x->default_value!=NULL) {
      if (x->type!=BOOL)
	param_[x->long_name] = std::string(x->default_value);
      else
	param_[x->long_name] = std::string();
    }
  }
}

bool
Param::parse(int& argc, const char**& argv, std::ostream& os /*= std::cout*/)
{
  for (int i=0; i<argc-1; ++i) {
    cmd_line_ += argv[i];
    cmd_line_ += ' ';
  }
  cmd_line_ += argv[argc-1];

  app_ = argv[0];
  argc--; argv++;

  int c=0;
  while (argc>0) {
    if (argv[0][0]=='-') {
      if (argv[0][1]==0) {
	break;
      } else if (argv[0][1]=='-') {	// long name
	const Options* x;
	for (x=opt_; x->type!=NONE; ++x) {
	  if (strcmp(x->long_name, &argv[0][2])==0) {
	    if (x->type==BOOL) {
	      param_[x->long_name] = std::string();
	    } else if (argc>1) {
	      param_[x->long_name] = std::string(argv[1]);
	      argc--; argv++;
	    }
	    c++;
	    break;
	  }
	}
	if (x->type==NONE) {
	  usage(os);
	  return false;
	}
      } else {			// short name
	const Options* x;
	for (x=opt_; x->type!=NONE; ++x) {
	  if (x->short_name && strcmp(x->short_name, &argv[0][1])==0) {
	    if (x->type==BOOL) {
	      param_[x->long_name] = std::string();
	    } else if (argc>1) {
	      param_[x->long_name] = std::string(argv[1]);
	      argc--; argv++;
	    }
	    c++;
	    break;
	  }
	}
	if (x->type==NONE) {
	  usage(os);
	  return false;
	}
      }
    } else {
      break;
    }
    argc--; argv++;
  }
  return true;
}

bool
Param::get_bool(const char* long_name) const
{
  return param_.find(long_name)!=param_.end();
}

int
Param::get_int(const char* long_name) const
{
  return atoi(get_string(long_name).c_str());
}

double
Param::get_real(const char* long_name) const
{
  return atof(get_string(long_name).c_str());
}

std::string
Param::get_string(const char* long_name) const
{
  std::map< const char*, std::string >::const_iterator x;
  x = param_.find(long_name);
  if (x!=param_.end())
    return (*x).second;
  else
    return std::string();
}

void
Param::set_bool(const char* long_name, bool val)
{
  if (val) {
    param_[long_name] = std::string();
  } else {
    param_.erase(long_name);
  }
}

void
Param::set_int(const char* long_name, int val)
{
  char buf[100];
  sprintf(buf, "%d", val);
  param_[long_name] = std::string(buf);
}

void
Param::set_real(const char* long_name, double val)
{
  char buf[100];
  sprintf(buf, "%lf", val);
  param_[long_name] = std::string(buf);
}

void
Param::set_string(const char* long_name, const char* val)
{
  param_[long_name] = std::string(val);
}

void
Param::usage(std::ostream& os /*=std::cout*/) const
{
  if (banner_) os << banner_;
  os << "Usage: ";
  if (app_) {
    os << app_ << " [options]";
    if (more_opt_) os << " " << more_opt_;
    os << std::endl;
  }
  os << "Options:" << std::endl;
  const Options* x;
  for (x=opt_; x->type!=NONE; ++x) {
    if (strcmp("HIDE", x->description)==0) continue;
    os << "   ";
    if (x->short_name)
      os << "-" << x->short_name << ", ";
    os << "--" << x->long_name << " ";
    switch (x->type) {
    case NONE:
    default:
    case BOOL:
      break;
    case INT:
      os << "(integer)";
      break;
    case REAL:
      os << "(real)";
      break;
    case STRING:
      os << "(string)";
      break;
    }
    if (x->description) os << "\t" << x->description;
    if (x->default_value) os << " (default: " << x->default_value << ")";
    os << std::endl;
  }
}

#ifdef TEST
using namespace std;
int
main(int argc, const char* argv[])
{
  Param::Options opt[] = {
    { Param::BOOL, "b", "bool", NULL, NULL},
    { Param::INT, "i", "int", "0", NULL},
    { Param::REAL, "r", "real", "0.0", NULL},
    { Param::STRING, "s", "string", "string", NULL},
    { Param::NONE, NULL, NULL, NULL, NULL}
  };

  cout << argc << " " << argv << endl;
  Param param(opt, argv[0]);
  argc--; argv++;
  param.parse(argc, argv);
  cout << argc << " " << argv << endl;

  cout << "bool " << param.get_bool("bool") << endl
       << "int " << param.get_int("int") << endl
       << "real " << param.get_real("real") << endl
       << "string " << param.get_string("string") << endl;
  
  
  return 0;
}
#endif
