/*
 * $Id: rnatree.cpp 498 2007-03-06 08:44:28Z satoken $
 *
 * PHMMTS -- an implementation of Pair Hidden Markov Models on Tree Structures,
 *           which is based on "Pair Hidden Markov Models on Tree Structures",
 *           Yasubumi Sakakibara, ISMB 2003.
 *
 * Copyright (C) 2003-2006 Kengo Sato
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifndef __TMPL_RNATREE_CPP__
#define __TMPL_RNATREE_CPP__

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <iostream>
#include <string>
#include "rnatree.h"

const std::string left_bracket("(<");
const std::string right_bracket(">)");

using namespace PHMMTS;

template < class Sequence >
void
RNATree<Sequence>::destroy()
{
  if (left()) left()->destroy();
  if (right()) right()->destroy();
  delete this;
}

template < class Sequence >
// static 
RNATree<Sequence>*
RNATree<Sequence>::
normalize(RNATree<Sequence>* tree)
{
  if (tree==NULL) return NULL;
  if (!tree->is_branch()) {
    tree->set_right(normalize(tree->right()));
    return tree;
  }
  
  RNATree* r = right_most_left(tree->right());
  if (r==NULL) return tree;

  assert(r->parent()->is_branch());
  RNATree* x = NULL;
  if (tree==r->parent()) { // right most are direct children
    x = tree->left();
    delete tree;
  } else {		// right most are indirect children
    x = tree;
    RNATree* d = r->parent();
    d->parent()->set_right(d->left());
    delete d;
  }
  RNATree* p = r;
  while (p!=NULL) {
    RNATree* prev = x;
    x = new RNATree(RSINGLE, p->left_rna());
    x->set_right(prev);
    p = p->right();
  }
  r->destroy();
  return x;
}

template < class Sequence >
// static
RNATree<Sequence>*
RNATree<Sequence>::right_most_left(RNATree* t)
{
  if (t==NULL) return NULL;
  if (t->is_branch()) return right_most_left(t->right());
  if (t->is_left_single()) {
    RNATree* ch = right_most_left(t->right());
    if (ch==t->right()) return t;
    else return ch;
  }
  return NULL;
}

#if 0
// a general case.
template < class Sequence >
// static
RNATree<Sequence>*
RNATree<Sequence>::parse(const Sequence& str,
			 const std::string& paren)
{
  struct local {
    static bool is_gap(const RNA& x) { return false; }
  };
  return parse(str, paren, local::is_gap);
}

// a special case in which Sequence == std::string
// static
RNATree<std::string>*
RNATree<std::string>::parse(const std::string& str,
			    const std::string& paren)
{
  struct local {
    static bool is_gap(const RNA& x) { return x=='-'; }
    
  };
  return parse(str, paren, local::is_gap);
}
#endif

template < class Sequence >
template < class Predicate >
//static
RNATree<Sequence>*
RNATree<Sequence>::parse(const Sequence& str,
			 const std::string& paren,
			 Predicate is_gap,
			 int start, int& end)
{
  /*
    production rules for RNA trees are:
      T -> empty
      T -> ( T ) T
      T -> . T
  */

  assert(str.size()==paren.size());
  if (start >= static_cast<int>(paren.size()) ||
      static_cast<int>(right_bracket.find(paren[start]))>=0) {
    end = start;
    return NULL;
  } else if (static_cast<int>(left_bracket.find(paren[start]))>=0) {
    // if an open bracket is appeared
    RNATree* ch = parse(str, paren, is_gap, start+1, end);
    if (static_cast<int>(right_bracket.find(paren[end]))>=0) {
      bool lgap = is_gap(str[start]);
      bool rgap = is_gap(str[end]);
      if (ch!=NULL) {
	RNATree* p = NULL;
	if (!lgap) {
	  if (!rgap) {
	    p = new RNATree(STEM, str[start], str[end], normalize(ch));
	  } else /* if (rgap) */ {
	    p = new RNATree(LSINGLE, str[start]);
	    p->set_right(normalize(ch));
	  }
	} else /* if (lgap) */ {
	  if (!rgap) {
	    p = new RNATree(RSINGLE, str[end]);
	    p->set_right(normalize(ch));
	  } else /* if (rgap) */ {
	    p = normalize(ch);
	  }
	}
	end++;
	start=end;
	ch = parse(str, paren, is_gap, start, end);
	return ch==NULL ? p : p!=NULL ? new RNATree(BRANCH, p, ch) : ch;
      } else {	     		// it seem to be a stem, but not a stem
	RNATree* p = NULL;
	RNATree* q = NULL;
	if (!lgap) p = new RNATree(LSINGLE, str[start]);
	if (!rgap) q = new RNATree(LSINGLE, str[end]);
	ch = parse(str, paren, is_gap, end+1, end);
	if (q!=NULL) {
	  q->set_right(ch);
	  if (p==NULL) {
	    return q;
	  } else {
	    p->set_right(q);
	    return p;
	  }
	} else if (p==NULL /*&& q==NULL*/) {
	  return ch;
	} else /*if (p!=NULL && q==NULL)*/ {
	  p->set_right(ch);
	  return p;
	}
      }
    } else {
      // parse error
      throw typename RNATree::ParseError(end);
    }
  } else {
    // this is a single node
    if (is_gap(str[start])) {	// A gap is skipped.
      return parse(str, paren, is_gap, start+1, end);
    } else {
      RNATree* ch = parse(str, paren, is_gap, start+1, end);
      RNATree* p = new RNATree(LSINGLE, str[start]);
      p->set_right(ch);
      return p;
    }
  }
}
  
template < class Sequence > 
std::string&
RNATree<Sequence>::to_string(std::string& str,
			     bool include_parenthesis /*=false*/) const
{
  if (is_stem()) {
    if (include_parenthesis) str+='(';
    str+=left_rna();
    if (left()) left()->to_string(str, include_parenthesis);
    str+=right_rna();
    if (include_parenthesis) str+=')';
  } else if (is_branch()) {
    if (left()) left()->to_string(str, include_parenthesis);
    if (right()) right()->to_string(str, include_parenthesis);
  } else if (is_left_single()) {
    str+=left_rna();
    if (right()) right()->to_string(str, include_parenthesis);
  } else /*if (is_right_single())*/ {
    if (right()) right()->to_string(str, include_parenthesis);
    str+=right_rna();
  }
  return str;
}

template < class Sequence >
std::ostream&
PHMMTS::operator<<(std::ostream& os, const RNATree<Sequence>* tree)
{
  std::string str;
  if (tree) os << tree->to_string(str, true);
  return os;
}

#if 0
int
main()
{
  std::string str =   "ACAGCGAUGCUAACGUGU";
  std::string paren = "(( ) (  )  )(  )";
  RNATree<std::string>* rna1 = RNATree<std::string>::parse(str, paren);
  std::cout << rna1 << std::endl;
  return 0;
}
#endif

#endif //  __TMPL_RNATREE_CPP__
