/*
 * $Id: rnatree.h 482 2006-04-12 09:00:23Z satoken $
 *
 * PHMMTS -- an implementation of Pair Hidden Markov Models on Tree Structures,
 *           which is based on "Pair Hidden Markov Models on Tree Structures",
 *           Yasubumi Sakakibara, ISMB 2003.
 *
 * Copyright (C) 2003-2006 Kengo Sato
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */


#ifndef __INC_RNATREE_H__
#define __INC_RNATREE_H__

#include <exception>
#include <string>
#include <cstdio>
#include <cctype>
#include <cassert>
#include "bintree.h"
#include "defines.h"

namespace PHMMTS {

  /**
   * @brief  A class of nodes of RNA tree which is used in RNATree.
   * @see RNATree
   */
  template <class RNA>
  class RNANode
  {
  public:
    RNA left_rna; ///< An RNA on the left side
    RNA right_rna; ///< An RNA on the right side
    node_type_t type;

    RNANode(node_type_t type) : type(type)
    {
      assert(type==BRANCH);
    }
    
    RNANode(node_type_t type, const RNA& rna) : type(type)
    {
      if (type==LSINGLE)
	left_rna = rna;
      else if (type==RSINGLE)
	right_rna = rna;
      else
	assert(false);
    }

    RNANode(node_type_t type, const RNA& left_rna, const RNA& right_rna)
      : left_rna(left_rna), right_rna(right_rna), type(type)
    {
      assert(type==STEM);
    }

    /**
     * @return Ture if the node is a stem node.
     */
    bool is_stem() const { return type==STEM; }

    /**
     * @return True if the node is a branch node.
     */
    bool is_branch() const { return type==BRANCH; }

    /**
     * @return True if the node is a single node on the left side.
     */
    bool is_left_single() const { return type==LSINGLE; }

    /**
     * @return True if the node is a single node on the right side.
     */
    bool is_right_single() const { return type==RSINGLE; }
  };

  /**
   * @brief  A class of RNA trees
   */
  template < class Sequence >
  class RNATree : public BinTree<RNANode<typename Sequence::value_type>*>
  {
    typedef typename Sequence::value_type RNA;
    typedef BinTree<RNANode<RNA>*> base_type;
    typedef RNANode<RNA> node_type;

  private:
    int size_; ///< The number of RNAs included in the tree.

  public:
    RNATree(node_type_t type, RNATree* left=NULL, RNATree* right=NULL)
      : base_type(new node_type(type), left, right), size_(0)
    {
      if (left) size_+=left->size_;
      if (right) size_+=right->size_;
    }

    RNATree(node_type_t type, const RNA& rna, RNATree* left=NULL)
      : base_type(new node_type(type, rna), left, NULL), size_(1)
    {
      if (left) size_+=left->size_;
    }

    RNATree(node_type_t type, const RNA& left_rna, const RNA& right_rna,
	    RNATree* left=NULL)
      : base_type(new node_type(type, left_rna, right_rna), left, NULL),
	size_(2)
    {
      if (left) size_+=left->size_;
    }


    RNATree(RNATree* parent=NULL)
      : base_type(NULL, NULL, NULL), size_(0)
    {
      set_parent(parent);
    }

    /**
     * @brief A destructor.
     */
    ~RNATree() { delete data(); }

    using base_type::is_leaf;
    using base_type::data;

    /**
     * @brief Destroy all nodes in the tree
     */
    void destroy();

    /**
     * @return  A pointer to the parent node
     */
    RNATree* parent() const { return (RNATree*)base_type::parent(); }

    /**
     * @return  A pointer to the left node
     */
    RNATree* left() const { return (RNATree*)base_type::left(); }

    /**
     * @return  A pointer to the right node
     */
    RNATree* right() const { return (RNATree*)base_type::right(); }

    /**
     * @return  True if the node is a stem node
     */
    bool is_stem() const { return data()->is_stem() && !is_leaf(); }

    /**
     * @return  True if the node is a branch node
     */
    bool is_branch() const { return data()->is_branch() && !is_leaf(); }

    /**
     * @return  True if the node is a single node
     * (backword compatibility)
     */
    //bool is_single() const { return data()->is_single(); }

    /**
     * @return  True if the node is a single node on the left side
     */
    bool is_left_single() const { return data()->is_left_single(); }

    /**
     * @return  True if the node is a single node on the right side
     */
    bool is_right_single() const { return data()->is_right_single(); }

    /**
     * @return  An RNA on the left side
     */
    const RNA& left_rna() const { return data()->left_rna; }

    /**
     * @return  An RNA on the right side.
     */
    const RNA& right_rna() const { return data()->right_rna; }

    /**
     * @return  The number of RNAs in the tree.
     */
    int size() const { return size_; }

    /**
     * @brief  Convert the tree to human-readable string.
     *
     * @param str                 A buffer area where the converted string is stored.
     * @param include_parenthesis If true, returned string includes parenthesis (brackets).
     *
     * @return the buffer area which is identical to 'str'
     */
    std::string& to_string(std::string& str,
			   bool include_parenthesis=false) const;
    
  public:
    template < class Predicate >
    static RNATree* parse(const Sequence& str,
			  const std::string& paren,
			  Predicate is_gap)
    {
      int end;
      return normalize(parse(str, paren, is_gap, 0, end));
    }

  private:
    template < class Predicate >
    static RNATree* parse(const Sequence& str,
			  const std::string& paren,
			  Predicate is_gap,
			  int start, int& end);

  private:
    static RNATree* right_most_left(RNATree* t);
    static RNATree* normalize(RNATree* t);
    
  public:
    /**
     * @brief  A class of the exception where parse error is happened.
     */
    class ParseError : public std::exception
    {
    private:
      int _pos;
      std::string _what;
    public:
      ParseError(int pos) : _pos(pos) {
	char buf[100];
	sprintf(buf, "parse error at column %d", pos);
	_what=buf;
      }
      virtual ~ParseError() throw() {};
      virtual const char* what() const throw () {
	return _what.c_str();
      }
      int pos() const { return _pos; }
    };
  };

  template < class RNA >
  std::ostream& operator<<(std::ostream& os, const RNATree<RNA>* tree);
}

#include "rnatree.cpp"

#endif // __INC_RNATREE_H__


// Local Variables:
// mode:C++
// End:
