/*
 * $Id:$
 *
 * PHMMTS -- an implementation of Pair Hidden Markov Models on Tree Structures,
 *           which is based on "Pair Hidden Markov Models on Tree Structures",
 *           Yasubumi Sakakibara, ISMB 2003.
 *
 * Copyright (C) 2003-2006 Kengo Sato
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifndef __INC_SCANTABLE_H__
#define __INC_SCANTABLE_H__

#include "defines.h"
#include <vector>

namespace PHMMTS
{
  template <class T>
  class ScanTable
  {
  public:
    typedef T value_type;
    typedef value_type node_type;

  private:
    class Table
    {
      T* table_;
      T** ptr_;
      size_t size_;

    public:
      Table(size_t size) : table_(NULL), ptr_(NULL), size_(size)
      {
	table_ = new T[size_*size_];
	ptr_ = new T*[size_];
	for (size_t i=0; i<size_; ++i)
	  ptr_[i] = &table_[size_*i];
      }

      ~Table() { delete table_; delete ptr_; }

      size_t size() const { return size_; }

      void fill(const T& val)
      {
	for (size_t i=0; i<size_*size_; ++i)
	  table_[i] = val;
      }

      void put(int i, int j, const T& val)
      {
	assert(i<=j);
	assert(static_cast<size_t>(j-i)<size_);
	ptr_[i%size_][j-i]=val;
      }

      const T& get(int i, int j) const
      {
	assert(i<=j);
	assert(static_cast<size_t>(j-i)<size_);
	return ptr_[i%size_][j-i];
      }

      T& get(int i, int j)
      {
	assert(i<=j);
	assert(static_cast<size_t>(j-i)<size_);
	return ptr_[i%size_][j-i];
      }

      const T& operator()(int i, int j) const
      {
	return get(i,j);
      }
  
      T& operator()(int i, int j)
      {
	return get(i,j);
      }
    };

  private:
    int height_;
    int length_;
    std::vector<Table*>* table_;

  public:
    ScanTable() : height_(0), length_(0), table_(NULL)
    { }
    
    ScanTable(int height, int length)
      : height_(height), length_(length), table_(NULL)
    {
      create();
    }

    ~ScanTable() { destroy_all(); }

    int get_length() const { return length_; }
    int get_height() const { return height_; }

    void set_size(int height, int length)
    {
      if (height_!=height || length_!=length) {
	destroy_all();
	height_ = height;
	length_ = length;
	create();
      }
    }

    void create()
    {
      destroy_all();
      table_ = new std::vector<Table*>[height_];
    }

    void create(int h, size_t states)
    {
      if (table_==NULL) create();
      if (table_[h].size()!=states) {
	destroy(h);
	table_[h].resize(states);
	for (uint s=0; s<states; ++s)
	  table_[h][s] = new Table(length_);
      }
    }

    void create(int h, int states, const T& val)
    {
      create(h, states);
      fill(h, val);
    }
    
    void destroy(int h)
    {
      for (uint s=0; s<table_[h].size(); ++s)
	delete table_[h][s];
      table_[h].clear();
    }

    void destroy_all()
    {
      if (table_!=NULL) {
	for (int h=0; h<height_; ++h)
	  destroy(h);
	delete[] table_;
	table_=NULL;
      }
    }

    void fill(int h, const T& val)
    {
      for (uint s=0; s<table_[h].size(); ++s)
	table_[h][s]->fill(val);
    }

    void fill(const T& val)
    {
      for (int h=0; h<height_; ++h)
	fill(h, val);
    }

  private:
    void put(int h, int s, int i, int j, const T& val)
    {
      table_[h][s]->put(i,j,val);
    }

    void put(const Position& pos, const T& val)
    {
      table_[pos.height][pos.state]->put(pos.left, pos.right, val);
    }

    const T& get(int h, int s, int i, int j) const
    {
      return table_[h][s]->get(i,j);
    }

    const T& get(const Position& pos) const
    {
      return table_[pos.height][pos.state]->get(pos.left, pos.right);
    }

    T& get(int h, int s, int i, int j)
    {
      return table_[h][s]->get(i,j);
    }

    T& get(const Position& pos)
    {
      return table_[pos.height][pos.state]->get(pos.left, pos.right);
    }

  public:
    const T& operator()(int h, int s, int i, int j) const
    {
      return get(h,s,i,j);
    }

    const T& operator()(const Position& pos) const
    {
      return get(pos);
    }

    T& operator()(int h, int s, int i, int j)
    {
      return get(h,s,i,j);
    }

    T& operator()(const Position& pos)
    {
      return get(pos);
    }
  };
}

#endif // __INC_SCANTABLE_H__

// Local Variables:
// mode: C++
// End:
