/*
 * $Id: skeletal_base.cpp 498 2007-03-06 08:44:28Z satoken $
 *
 * PHMMTS -- an implementation of Pair Hidden Markov Models on Tree Structures,
 *           which is based on "Pair Hidden Markov Models on Tree Structures",
 *           Yasubumi Sakakibara, ISMB 2003.
 *
 * Copyright (C) 2003-2006 Kengo Sato
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <iostream>
#include <cassert>
#include <string>
#include <vector>
#include "skeletal_base.h"
#include "rnatree.h"

using namespace PHMMTS;

template < class RNA >
SkeletalTreeNodeBase<RNA>::
SkeletalTreeNodeBase(const SkeletalTreeNodeBase& n)
  : parent_(n.parent_), left_(n.left_), right_(n.right_),
    type_(n.type_), left_rna_(n.left_rna_), right_rna_(n.right_rna_),
    prop_(n.prop_)
{

}

template < class RNA >
SkeletalTreeNodeBase<RNA>&
SkeletalTreeNodeBase<RNA>::
operator=(const SkeletalTreeNodeBase& n)
{
  if (this!=&n) {
    parent_=n.parent_;
    left_=n.left_;
    right_=n.right_;
    type_=n.type_;
    left_rna_=n.left_rna_;
    right_rna_=n.right_rna_;
    prop_=n.prop_;
  }
  return *this;
}

template < class RNA >
SkeletalTreeNodeBase<RNA>::
SkeletalTreeNodeBase(node_type_t type)
  : parent_(-1), left_(-1), right_(-1), type_(type), prop_()
{
  assert(type==EMPTY);
}

template < class RNA >
SkeletalTreeNodeBase<RNA>::
SkeletalTreeNodeBase(node_type_t type, int left, RNA rna)
  : parent_(-1), left_(left), right_(-1), type_(type), prop_()
{
  switch (type) {
  case LSINGLE:
    left_rna_ = rna;
    break;

  case RSINGLE:
    right_rna_ = rna;
    break;

  default:
    assert(!"unreachable condition");
    break;
  }
}

template < class RNA >
SkeletalTreeNodeBase<RNA>::
SkeletalTreeNodeBase(node_type_t type,
		     int left, RNA left_rna, RNA right_rna)
  : parent_(-1), left_(left), right_(-1), type_(type),
    left_rna_(left_rna), right_rna_(right_rna), prop_()
{
  assert(type==STEM);
}

template < class RNA >
SkeletalTreeNodeBase<RNA>::
SkeletalTreeNodeBase(node_type_t type, int left)
  : parent_(-1), left_(left), right_(-1), type_(type), prop_()
{
  assert(type==START);
}

template < class RNA >
SkeletalTreeNodeBase<RNA>::
SkeletalTreeNodeBase(node_type_t type, int left, int right)
  : parent_(-1), left_(left), right_(right), type_(type), prop_()
{
  assert(type==BRANCH);
}

template < class RNA >
void
SkeletalTreeNodeBase<RNA>::
show_children()
{
  for (uint i=0; i<prop_.size(); ++i) {
    std::cout << "  " << i << "  ";
    typename Property::trans_list_t::const_iterator x;
    for (x=prop_[i].children.begin(); x!=prop_[i].children.end(); ++x) {
      std::cout << "(" << x->first << "," << x->second << "), ";
    }
    std::cout << std::endl;
    if (!prop_[i].rchildren.empty()) {
      assert(type_==BRANCH);
      std::cout << "  " << i << "r ";
      typename Property::trans_list_t::const_iterator x;
      for (x=prop_[i].rchildren.begin(); x!=prop_[i].rchildren.end(); ++x) {
	std::cout << "(" << x->first << "," << x->second << "), ";
      }
      std::cout << std::endl;
    }
  }
}

// public
template < class Sequence, class Node >
void
SkeletalTreeBase<Sequence,Node>::
to_string(Sequence& str, std::string& paren, int pos) const
{
  switch(nodes_[pos].type()) {
  case BRANCH:
  case START:
    if (nodes_[pos].left()>=0)
      to_string(str, paren, nodes_[pos].left());
    if (nodes_[pos].right()>=0)
      to_string(str, paren, nodes_[pos].right());
    break;
  case STEM:
    paren += '(';
    str.push_back(nodes_[pos].left_rna());
    if (nodes_[pos].left()>=0)
      to_string(str, paren, nodes_[pos].left());
    str.push_back(nodes_[pos].right_rna());
    paren += ')';
    break;
  case LSINGLE:
    str.push_back(nodes_[pos].left_rna());
    paren += ' ';
    if (nodes_[pos].left()>=0)
      to_string(str, paren, nodes_[pos].left());
    break;
  case RSINGLE:
    if (nodes_[pos].left()>=0)
      to_string(str, paren, nodes_[pos].left());
    str.push_back(nodes_[pos].right_rna());
    paren += ' ';
    break;
  case EMPTY:
  default:
    break;
  }
}

// protected
template < class Sequence, class Node >
void
SkeletalTreeBase<Sequence,Node>::
to_string_outside(Sequence& left_str, std::string& left_paren,
		  Sequence& right_str, std::string& right_paren,
		  int from, int pos, bool& right_p) const
{
  if (from==pos) {
    right_p = true;
    return;
  }
  
  switch(nodes_[pos].type()) {
  case BRANCH:
  case START:
    if (nodes_[pos].left()>=0)
      to_string_outside(left_str, left_paren, right_str, right_paren,
			from, nodes_[pos].left(), right_p);
    if (nodes_[pos].right()>=0)
      to_string_outside(left_str, left_paren, right_str, right_paren,
			from, nodes_[pos].right(), right_p);
    break;
  case STEM:
    if (!right_p) {
      left_paren += '(';
      left_str.push_back(nodes_[pos].left_rna());
    } else {
      right_paren += '(';
      right_str.push_back(nodes_[pos].left_rna());
    }
    if (nodes_[pos].left()>=0)
      to_string_outside(left_str, left_paren, right_str, right_paren,
			from, nodes_[pos].left(), right_p);
    if (!right_p) {
      left_str.push_back(nodes_[pos].right_rna());
      left_paren += ')';
    } else {
      right_str.push_back(nodes_[pos].right_rna());
      right_paren += ')';
    }
    break;
  case LSINGLE:
    if (!right_p) {
      left_str.push_back(nodes_[pos].left_rna());
      left_paren += ' ';
    } else {
      right_str.push_back(nodes_[pos].left_rna());
      right_paren += ' ';
    }
    if (nodes_[pos].left()>=0)
      to_string_outside(left_str, left_paren, right_str, right_paren,
			from, nodes_[pos].left(), right_p);
    break;
  case RSINGLE:
    if (nodes_[pos].left()>=0)
      to_string_outside(left_str, left_paren, right_str, right_paren,
			from, nodes_[pos].left(), right_p);
    if (!right_p) {
      left_str.push_back(nodes_[pos].right_rna());
      left_paren += ' ';
    } else {
      right_str.push_back(nodes_[pos].right_rna());
      right_paren += ' ';
    }
    break;
  case EMPTY:
  default:
    break;
  }
}

// public
template < class Sequence, class Node >
void
SkeletalTreeBase<Sequence,Node>::
to_string_outside(Sequence& left_str, std::string& left_paren,
		  Sequence& right_str, std::string& right_paren,
		  int from) const
{
  bool right_p=false;
  to_string_outside(left_str, left_paren, right_str, right_paren,
		    from, nodes_.size()-1, right_p);
}

template < class Sequence, class Node >
void
SimpleSkeletalTree<Sequence,Node>::
parse(const Sequence& str, const std::string& paren)
{
  struct local {
    static bool is_gap(const typename Sequence::value_type& val)
    {
      return false;
    }
  };
  RNATree<Sequence>* t =
    RNATree<Sequence>::parse(str, paren, local::is_gap);
  parse(t);
  t->destroy();
}

template < class Sequence, class Node >
template < class Predicate >
void
SimpleSkeletalTree<Sequence,Node>::
parse(const Sequence& str, const std::string& paren, Predicate is_gap)
{
  RNATree<Sequence>* t = RNATree<Sequence>::parse(str, paren, is_gap);
  parse(t);
  t->destroy();
}

template < class Node >
static
void
make_properties_helper(typename Node::Property p[], size_t sz,
		       const Node& self, const Node& ch)
{
  assert(self.type()==LSINGLE || self.type()==RSINGLE ||
	 self.type()==STEM || self.type()==START);

  typedef typename Node::Property Property;
  typedef typename Property::trans_list_t trans_list_t;
  typedef typename Property::trans_t trans_t;

  trans_list_t ch_trans;
  for (uint i=0; i<ch.size(); ++i)
    ch_trans.push_back(trans_t(self.left(), i));

  trans_list_t self_trans;
  for (uint i=0; i<sz; ++i)
    self_trans.push_back(trans_t(ch.parent(), i));

  bool local_end = ch.type()!=BRANCH && ch.type()!=START;

  for (uint i=0; i<sz; ++i) {
    switch (p[i].state) {
    case INSERT:
      p[i].children = self_trans;
      break;
    default:
      p[i].children = ch_trans;
      break;
    }
    p[i].local_end = p[i].local_end && local_end;
  }
}

template < class Node >
static
void
make_properties_helper(typename Node::Property p[], size_t sz,
		       const Node& self,
		       const Node& ch_left,
		       const Node& ch_right)
{
  assert(self.type()==BRANCH);
  assert(ch_left.parent()==ch_right.parent());

  typedef typename Node::Property Property;
  typedef typename Property::trans_list_t trans_list_t;
  typedef typename Property::trans_t trans_t;

  trans_list_t ch_trans;
  for (uint i=0; i<ch_left.size(); ++i)
    ch_trans.push_back(trans_t(self.left(), i));
  trans_list_t rch_trans;
  for (uint i=0; i<ch_right.size(); ++i)
    rch_trans.push_back(trans_t(self.right(), i));
  trans_list_t self_trans;
  for (uint i=0; i<sz; ++i)
    self_trans.push_back(trans_t(ch_left.parent(), i));

  for (uint i=0; i<sz; ++i) {
    switch (p[i].state) {
    case INSERT:
      p[i].children = self_trans;
      break;
    default:
      p[i].children = ch_trans;
      p[i].rchildren = rch_trans;
      break;
    }
  }
}

template < class STree >
static
void
make_properties(STree& tree)
{
  typedef typename STree::node_type Node;
  typedef typename Node::Property Property;

  for (uint i=0; i<tree.size(); ++i) {
    switch (tree[i].type()) {
    case EMPTY:
      {
	Property p[] = {
	  Property(MATCH, 0, 0, 0, 0, ' ', ' ', 0, false, false),
	};
	tree[i].set_properties(p, p+1);
      }
      break;
    case LSINGLE:
      {
	Property p[] = {
	  Property(MATCH,  1, 0, 1, 0, '|', '*', 0, true, true),
	  Property(DELETE, 1, 0, 0, 0, '-', '-', 0, false, false),
	  Property(INSERT, 0, 0, 1, 0, '-', '-', 0, false, false),
	};
	make_properties_helper(p, 3, tree[i], tree[tree[i].left()]);
	tree[i].set_properties(p, p+3);
      }
      break;
    case RSINGLE:
      {
	Property p[] = {
	  Property(MATCH,  0, 1, 0, 1, '|', '*', 0, true, true),
	  Property(DELETE, 0, 1, 0, 0, '-', '-', 0, false, false),
	  Property(INSERT, 0, 0, 0, 1, '-', '-', 0, false, false),
	};
	make_properties_helper(p, 3, tree[i], tree[tree[i].left()]);
	tree[i].set_properties(p, p+3);
      }
      break;
    case STEM:
      {
	Property p[] = {
	  Property(MATCH,  1, 1, 1, 1, '|', '*', 1, true, true),
	  Property(DELETE, 1, 1, 0, 0, '-', '-', 0, false, false),
	  Property(INSERT, 0, 0, 1, 1, '-', '-', 1, false, false),
	  Property(INSERT, 0, 0, 1, 0, '-', '-', 0, false, false),
	  Property(INSERT, 0, 0, 0, 1, '-', '-', 0, false, false),
	};
	make_properties_helper(p, 5, tree[i], tree[tree[i].left()]);
	tree[i].set_properties(p, p+5);
      }
      break;
    case START:
      {
	Property p[] = {
	  Property(MATCH,  0, 0, 0, 0, ' ', ' ', 0, true, true),
	};
	make_properties_helper(p, 1, tree[i], tree[tree[i].left()]);
	tree[i].set_properties(p, p+1);
      }
      break;
    case BRANCH:
      {
	Property p[] = {
	  Property(MATCH,  0, 0, 0, 0, ' ', ' ', 0, true, false),
	  Property(INSERT, 0, 0, 1, 1, '-', '-', 1, false, false),
	  Property(INSERT, 0, 0, 1, 0, '-', '-', 0, false, false),
	  Property(INSERT, 0, 0, 0, 1, '-', '-', 0, false, false),
	};
	make_properties_helper(p, 4, tree[i],
			       tree[tree[i].left()],
			       tree[tree[i].right()]);
	tree[i].set_properties(p, p+4);
      }
      break;
    default:
      assert(!"unreachable");
      break;
    }
  }
}

// public
template < class Sequence, class Node >
void
SimpleSkeletalTree<Sequence,Node>::
parse(const RNATree<Sequence>* tree)
{
  int l,c;
  this->clear();
  parse_helper(tree);
  l = this->size()-1;
  c = l+1;
  this->back().parent() = c;
  this->push_back(node_type(START, l));
  make_properties(*this);

#if 0
  for (uint i=0; i<this->size(); ++i) {
    std::cout << "node " << i << ": " << (*this)[i].type() << std::endl;
    (*this)[i].show_children();
  }
#endif
}

// private
template < class Sequence, class Node >
void
SimpleSkeletalTree<Sequence,Node>::
parse_helper(const RNATree<Sequence>* tree)
{
  // make a empty node
  if (tree->is_leaf()) {
    assert(!tree->is_stem());
    assert(!tree->is_branch());
    this->push_back(node_type(EMPTY));
  }
  if (tree->is_left_single()) {
    // make a left-single node
    int l, c;
    if (tree->right()) parse_helper(tree->right());
    l = this->size()-1;
    c = l+1;
    this->back().parent() = c;
    this->push_back(node_type(LSINGLE, l, tree->left_rna()));
  } else if (tree->is_right_single()) {
    // make a right-single node
    int l, c;
    if (tree->right()) parse_helper(tree->right());
    l = this->size()-1;
    c = l+1;
    this->back().parent() = c;
    this->push_back(node_type(RSINGLE, l, tree->right_rna()));
  } else if (tree->is_stem()) {
    // make a stem node
    int l, c;
    assert(tree->left()!=NULL);
    parse_helper(tree->left());
    l = this->size()-1;
    c = l+1;
    this->back().parent() = c;
    this->push_back(node_type(STEM, l, tree->left_rna(), tree->right_rna()));
  } else /* if (tree->is_branch()) */ {
    int l, r, c;
    assert(tree->left()!=NULL);
    assert(tree->right()!=NULL);
    // parse left nodes
    parse_helper(tree->left());
    l = this->size()-1;
    c = l+1;
    this->back().parent() = c;
    this->push_back(node_type(START, l));
    l = this->size()-1;
    // parse right nodes
    parse_helper(tree->right());
    r = this->size()-1;
    c = r+1;
    this->back().parent() = c;
    this->push_back(node_type(START, r));
    r = this->size()-1;
    // make a branch node
    c = r+1;
    (*this)[l].parent()=c;
    (*this)[r].parent()=c;
    this->push_back(node_type(BRANCH, l, r));
  }
}

