/*
 * $Id: skeletal_base.h 482 2006-04-12 09:00:23Z satoken $
 *
 * PHMMTS -- an implementation of Pair Hidden Markov Models on Tree Structures,
 *           which is based on "Pair Hidden Markov Models on Tree Structures",
 *           Yasubumi Sakakibara, ISMB 2003.
 *
 * Copyright (C) 2003-2006 Kengo Sato
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifndef __INC_SKELETAL_BASE_H__
#define __INC_SKELETAL_BASE_H__

#include <vector>
#include <string>
#include <iterator>
#include "rna.h"
#include "defines.h"

namespace PHMMTS {

  template < class RNA >
  class SkeletalTreeNodeBase {
  public:
    /**
     * An inner structure for controlling behavior on the algorithm.
     */
    struct Property {
      typedef std::pair<int,int> trans_t;
      typedef std::vector<trans_t> trans_list_t;

      match_state_t state;	/**< state type for cost table */
      int8_t f_left;  /**< # of RNAs on the left for foled sequence */
      int8_t f_right; /**< # of RNAs on the right for foled sequence */
      int8_t u_left;  /**< # of RNAs on the left for unfoled sequence */
      int8_t u_right; /**< # of RNAs on the right for unfoled sequence */
      char match;		/**< charactor for match state */
      char mismatch;		/**< charactor for missmatch state */
      int loop;	    /**< minimum # of RNAs the child nodes must have */
      bool local_start;	/**< this node can be used for local start or not */
      bool local_end; /**< this node can be used for local end or not */
      trans_list_t children;
      trans_list_t rchildren;

      Property() : children(), rchildren() { }
      
      Property(match_state_t state, int8_t f_left, int8_t f_right,
	       int8_t u_left, int8_t u_right,
	       char match, char mismatch, int loop,
	       bool local_start, bool local_end)
	: state(state), f_left(f_left), f_right(f_right),
	  u_left(u_left), u_right(u_right),
	  match(match), mismatch(mismatch), loop(loop),
	  local_start(local_start), local_end(local_end),
	  children(), rchildren()
      { }

      Property(const Property& p)
	: state(p.state), f_left(p.f_left), f_right(p.f_right),
	  u_left(p.u_left), u_right(p.u_right),
	  match(p.match), mismatch(p.mismatch), loop(p.loop),
	  local_start(p.local_start), local_end(p.local_end),
	  children(p.children),
	  rchildren(p.rchildren)
      { }

      Property& operator=(const Property& p)
      {
	if (this!=&p) {
	  state=p.state;
	  f_left=p.f_left; f_right=p.f_right;
	  u_left=p.u_left; u_right=p.u_right;
	  match=p.match; mismatch=p.mismatch; loop=p.loop;
	  local_start=p.local_start; local_end=p.local_end;
	  children=p.children; rchildren=p.rchildren;
	}
	return *this;
      }
    };
    
  private:
    int parent_;		/**< The index of the parent node */
    int left_;			/**< The index of the left child node */
    int right_;			/**< The index of the right child node */
    node_type_t type_;		/**< The type of a node */
    RNA left_rna_;		/**< An RNA on the left side */
    RNA right_rna_;		/**< An RNA on the right side */
    std::vector<Property> prop_; /**< properties of this node */

  public:
    // copy constructor
    SkeletalTreeNodeBase(const SkeletalTreeNodeBase& n);
    // constructor for EMPTY
    SkeletalTreeNodeBase(node_type_t type);
    // constructor for LSINGLE and RSINGLE
    SkeletalTreeNodeBase(node_type_t type, int left, RNA rna);
    // constructor for STEM
    SkeletalTreeNodeBase(node_type_t type, 
			 int left, RNA left_rna, RNA right_rna);
    // constructor for START
    SkeletalTreeNodeBase(node_type_t type, int left);
    // constructor for BRANCH
    SkeletalTreeNodeBase(node_type_t type, int left, int right);

    SkeletalTreeNodeBase& operator=(const SkeletalTreeNodeBase& n);

    template <class iterator>
    void set_properties(iterator s, iterator e)
    {
      prop_.resize(std::distance(s, e));
      std::copy(s, e, prop_.begin());
    }
    
    void show_children();
    
    int parent() const { return parent_; }
    int& parent() { return parent_; }
    int left() const { return left_; }
    int right() const { return right_; }
    node_type_t type() const { return type_; }
    const RNA& left_rna() const { return left_rna_; }
    const RNA& right_rna() const { return right_rna_; }
    size_t size() const { return prop_.size(); }
    const Property& property(int i) const { return prop_[i]; }
  };

  template < class Sequence_, class Node_ >
  class SkeletalTreeBase
  {
    typedef typename Sequence_::value_type RNA;
    std::vector< Node_ > nodes_;
    std::string name_;
    
  public:
    typedef Sequence_ Sequence;
    typedef Node_ node_type;

  public:
    // default constructor
    SkeletalTreeBase() : nodes_(), name_() {}
    // copy constructor
    SkeletalTreeBase(const SkeletalTreeBase& t)
      : nodes_(t.nodes_), name_(t.name_) { }

    SkeletalTreeBase& operator=(const SkeletalTreeBase& t)
    {
      nodes_ = t.nodes_;
      name_ = t.name_;
      return *this;
    }

    Sequence& to_string(Sequence& str) const
    {
      std::string paren;
      to_string(str, paren, size()-1);
      return str;
    }
    
    Sequence to_string() const
    {
      Sequence ret;
      return to_string(ret);
    }

    void to_string(Sequence& str, std::string& paren, int pos) const;
    
    void to_string(Sequence& str, std::string& paren) const
    {
      to_string(str, paren, size()-1);
    }

    void to_string_outside(Sequence& left_str, std::string& left_paren,
			   Sequence& right_str, std::string& right_paren,
			   int from) const;

    size_t size() const { return nodes_.size(); }
    void clear() { nodes_.clear(); }
    const node_type& operator[](size_t n) const { return nodes_[n]; }
    node_type& operator[](size_t n) { return nodes_[n]; }
    void push_back(const node_type& n) { nodes_.push_back(n); }
    const node_type& root() const { return nodes_.back(); }
    node_type& back() { return nodes_.back(); }
    const node_type& back() const { return nodes_.back(); }
    node_type& root() { return nodes_.back(); }
    const std::string& name() const { return name_; }
    void set_name(const std::string& n) { name_ = n; }

  private:
    void to_string_outside(Sequence& left_str, std::string& left_paren,
			   Sequence& right_str, std::string& right_paren,
			   int from, int pos, bool& right_p) const;

  public:
    template <class ST, class Func>
    static void traverse(ST& st, Func& func)
    {
      traverse(st, func, st.size()-1);
    }

  private:
    template <class ST, class Func>
    static void traverse(ST& st, Func& func, size_t n)
    {
      switch (st[n].type()) {
      case BRANCH:
	if (st[n].left()>=0)
	  traverse(st, func, st[n].left());
	if (st[n].right()>=0)
	  traverse(st, func, st[n].right());
	break;

      case START:
      case STEM:
      case LSINGLE:
      case RSINGLE:
	assert(st[n].left()>=0);
	traverse(st, func, st[n].left());
	break;

      case EMPTY:
	break;

      default:
      case NOTUSE:
	assert(!"NOTUSE doesn't use.");
	break;
      }
      func(st, n);
    }
  };

  template <class> class RNATree;

  template < class Sequence_, class Node_ >
  class SimpleSkeletalTree : public SkeletalTreeBase<Sequence_,Node_>
  {
    typedef typename Sequence_::value_type RNA;
    typedef SkeletalTreeBase<Sequence_,Node_> base_;
    typedef typename Node_::Property Property;

  public:
    typedef Sequence_ Sequence;
    typedef Node_ node_type;

  public:
    // default constructor
    SimpleSkeletalTree() : base_() {}
    // copy constructor
    SimpleSkeletalTree(const SimpleSkeletalTree& t) : base_(t) { }
    // construct from RNATree
    SimpleSkeletalTree(const RNATree<Sequence>* tree) : base_()
    {
      parse(tree);
    }
    // construct from strings
    SimpleSkeletalTree(const Sequence& str, const std::string& paren)
      : base_()
    {
      parse(str, paren);
    }
    // construct from strings
    template < class Predicate >
    SimpleSkeletalTree(const Sequence& str, const std::string& paren,
		       Predicate is_gap) : base_()
    {
      parse(str, paren, is_gap);
    }

    void parse(const RNATree<Sequence>* tree);

    void parse(const Sequence& str, const std::string& paren);

    template < class Predicate >
    void parse(const Sequence& str, const std::string& paren,
	       Predicate is_gap);

  private:
    void parse_helper(const RNATree<Sequence>* tree);
  };
}

#endif // __INC_SKELETAL_BASE_H__

// Local Variables:
// mode: C++
// End:
