/*
 * $Id: train.cpp 511 2007-09-18 06:18:08Z satoken $
 *
 * PHMMTS -- an implementation of Pair Hidden Markov Models on Tree Structures,
 *           which is based on "Pair Hidden Markov Models on Tree Structures",
 *           Yasubumi Sakakibara, ISMB 2003.
 *
 * Copyright (C) 2003-2006 Kengo Sato
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <iostream>
#include <fstream>
#include <string>
#include <list>
#include <vector>
#include <sys/time.h>
#ifdef HAVE_MPI
#include <mpi.h>
#endif
#include "example.h"
#include "log_value.h"
#include "costtable.h"
#include "param.h"
#include "timersub.h"

using namespace PHMMTS;

typedef cost_t LCostType;
typedef CostTable<LCostType> LCostTable;
typedef CostTable<float> Expectation;

template < class Expl >
LCostTable
em_training(const std::list<Expl>& train, const LCostTable& init,
	    size_t max_it, float eps, int verbose,
	    size_t& it, float& f, search_type_t search_type=GLOBAL_ALIGNMENT);

template < class Expl >
void
crf_training(const std::list<Expl>& train,
	     std::vector<double>& x,
	     const Expectation& cf,
	     size_t max_it, float eps, float sigma, int verbose,
	     size_t& it, float& ll, search_type_t search_type=GLOBAL_ALIGNMENT);


Param::Options opts[] = {
  { Param::BOOL,   "v", "verbose", NULL,
    "\n\tenable verbose outputs"},
  { Param::INT,    "m", "max-iteration", "100",
    "\n\tthe maximum iteration of the algorithm"},
  { Param::STRING, "a", "algorithm", "EM",
    "\n\tthe algorithm to be used for paramter estimation"},
  { Param::REAL,   "c", "conv-rate", "1e-5",
    "\n\tthe threshold of convergence rate"},
  { Param::BOOL, "l", "local", NULL,
    "\n\testmate paremeters for local alignment" },
  { Param::BOOL, "l", "local", NULL,
    "\n\testmate paremeters for local structure search" },
  { Param::REAL,   "s", "sigma", "1.0",
    "\n\tthe meta parameter of Gaussian prior"},
  { Param::STRING, "i", "initial-model", NULL,
    "\n\tthe file name of an initial model"},
  {0,0,0,0,0}
};

template <class CostTable>
void
set_default_parameters(CostTable& cost_table)
{
  typedef typename CostTable::cost_type cost_type;
  // emit
  char bases[] = "AUGC";
  char base_pairs[][3] = {
    "AU", "CG", "GC", "GU", "UA", "UG", ""
  };
  cost_type mat(0.0);
  cost_type ins(0.0);
  cost_type del(0.0);

  // stem
  for (int i=0; bases[i]!=0; ++i) {
    for (int j=0; bases[j]!=0; ++j) {
      for (int k=0; bases[k]!=0; ++k) {
	for (int l=0; bases[l]!=0; ++l) {
	  cost_table.add_emit_count(bases[i], bases[j], bases[k], bases[l],
				    static_cast<cost_type>(0.1));
	  mat+=static_cast<cost_type>(0.1);
	}
      }

      cost_table.add_emit_count('-', '-', bases[i], bases[j],
				static_cast<cost_type>(0.1));
      ins+=static_cast<cost_type>(0.1);
      cost_table.add_emit_count(bases[i], bases[j], '-', '-',
				static_cast<cost_type>(0.1));
      del+=static_cast<cost_type>(0.1);
    }
    cost_table.add_emit_count(bases[i], static_cast<cost_type>(0.1));
  }

  for (int i=0; base_pairs[i][0]!=0; ++i) {
    // for match
    for (int j=0; base_pairs[j][0]!=0; ++j) {
      cost_type a(
	base_pairs[i][0]==base_pairs[j][0] &&
	base_pairs[i][1]==base_pairs[j][1] ? 50.0 : 1.0);
      cost_table.add_emit_count(base_pairs[i][0], base_pairs[i][1],
				base_pairs[j][0], base_pairs[j][1],
				static_cast<cost_type>(a));
      mat+=a;
    }
    // for insert
    cost_table.add_emit_count('-', '-', base_pairs[i][0], base_pairs[i][1],
			      static_cast<cost_type>(1.0));
    ins+=static_cast<cost_type>(1.0);
    // for delete
    cost_table.add_emit_count(base_pairs[i][0], base_pairs[i][1], '-', '-',
			      static_cast<cost_type>(1.0));
    del+=static_cast<cost_type>(1.0);
  }

  // single
  for (int i=0; bases[i]!=0; ++i) {
    // for match
    for (int j=0; bases[j]!=0; ++j) {
      cost_type a(bases[i]==bases[j] ? 100.0 : 1.0);
      cost_table.add_emit_count(bases[i], '-', bases[j], '-', a);
      mat+=a;
    }
    // for insert
    cost_table.add_emit_count('-', '-', bases[i], '-',
			      static_cast<cost_type>(1.0));
    ins+=static_cast<cost_type>(1.0);
    // for delete
    cost_table.add_emit_count(bases[i], '-', '-', '-',
			      static_cast<cost_type>(1.0));
    del+=static_cast<cost_type>(1.0);

    // for local
    cost_table.add_emit_count(bases[i], static_cast<cost_type>(1.0));
  }

  // state
  cost_table.add_state_count(MATCH, MATCH, 0.8);
  cost_table.add_state_count(MATCH, INSERT, 0.1);
  cost_table.add_state_count(MATCH, DELETE, 0.1);

  cost_table.add_state_count(INSERT, MATCH, 0.7);
  cost_table.add_state_count(INSERT, INSERT, 0.3);

  cost_table.add_state_count(DELETE, MATCH, 0.7);
  cost_table.add_state_count(DELETE, DELETE, 0.3);

  cost_table.add_state_count(L_START, MATCH, 0.01);
  cost_table.add_state_count(MATCH, L_END, 0.01);
  //cost_table.add_state_count(L_START, L_START, 0.005);
  //cost_table.add_state_count(L_END, L_END, 0.005);

  cost_table.add_state_count(MATCH, MATCH, MATCH, 1.0);

  cost_table.normalize();
}

int
run_em_training(int argc, const char** argv, const Param& param)
{
  // verbose
  int verbose = 0;
  if (param.get_bool("verbose")) verbose=3;
  search_type_t search_type=GLOBAL_ALIGNMENT;
  if (param.get_bool("local")) search_type=LOCAL_ALIGNMENT;
  if (param.get_bool("local-structure")) search_type=LOCAL_STRUCTURE;
#ifdef HAVE_MPI
  int my_rank = MPI::COMM_WORLD.Get_rank();
  if (my_rank!=0) verbose=0;
#endif

  // output
  std::ostream* output = &std::cout;
#ifdef HAVE_MPI
  if (my_rank==0) {
#endif
    if (strcmp(argv[0], "-")!=0) {
      std::ofstream* out = new std::ofstream(argv[0]);
      if (!out->is_open()) {
	perror(argv[0]);
	return 1;
      }
      output = out;
    }
#ifdef HAVE_MPI
  }
#endif

  // initialize the model
  std::string init_model = param.get_string("initial-model");
  LCostTable init;
  if (init_model.empty()) {
    set_default_parameters(init);
  } else {
    std::ifstream in(init_model.c_str());
    if (!in.is_open()) {
#ifdef HAVE_MPI
      if (my_rank==0) {
#endif
	perror(init_model.c_str());
#ifdef HAVE_MPI
      }
#endif
      return 1;
    }
    init.restore(in);
  }

  // load example sequences
  std::list<ExampleS> train;
  if (argc>=2) {		// load from files in args
    for (int i=1; i<argc; ++i) {
      std::ifstream in(argv[i]);
      if (!in.is_open()) {
#ifdef HAVE_MPI
	if (my_rank==0) {
#endif
	  perror(argv[i]);
#ifdef HAVE_MPI
	}
#endif
	return 1;
      }
      ExampleS::load_examples(in, train);
    }
  } else {			// load from stdin
    ExampleS::load_examples(std::cin, train);
  }

  if (verbose>0)
    std::cout << train.size() << " data can be used." << std::endl;

  // execute training procedures
  {
    float eps = param.get_real("conv-rate");
    int max_it = param.get_int("max-iteration");
    size_t it;
    float ll;
    struct timeval start_tm, finish_tm, calc_tm;
    char tm_buf[20];

    gettimeofday(&start_tm, NULL);
    LCostTable res = em_training(train, init, max_it, eps, verbose,
				 it, ll, search_type);
    gettimeofday(&finish_tm, NULL);
    timersub(&finish_tm, &start_tm, &calc_tm);
    sprintf(tm_buf, "%ld.%03lds", calc_tm.tv_sec, calc_tm.tv_usec/1000);
#ifdef HAVE_MPI
    if (my_rank==0) {
#endif
      *output << "# CostTable for PHMMTS" << std::endl
	      << "# Trained by \"" << param.cmd_line() << "\"" << std::endl
	      << "#   method: EM" << std::endl
	      << "#   time: " << tm_buf << std::endl
	      << "#   iteration: " << it << std::endl
	      << "#   loglikelihood: " << ll << std::endl;
      res.store(*output);
      std::cout << "# CostTable for PHMMTS" << std::endl
	      << "# Trained by \"" << param.cmd_line() << "\"" << std::endl
	      << "#   method: EM" << std::endl
	      << "#   time: " << tm_buf << std::endl
	      << "#   iteration: " << it << std::endl
	      << "#   loglikelihood: " << ll << std::endl;
#ifdef HAVE_MPI
    }
#endif
    if (output!=&std::cout) delete output;
  }

  return 0;
}

int
run_crf_training(int argc, const char** argv, const Param& param)
{
  // verbose
  int verbose = 0;
  if (param.get_bool("verbose")) verbose=3;
  search_type_t search_type=GLOBAL_ALIGNMENT;
  if (param.get_bool("local")) search_type=LOCAL_ALIGNMENT;
  if (param.get_bool("local-structure")) search_type=LOCAL_STRUCTURE;
#ifdef HAVE_MPI
  int my_rank = MPI::COMM_WORLD.Get_rank();
  if (my_rank!=0) verbose=0;
#endif

  // output
  std::ostream* output = &std::cout;
#ifdef HAVE_MPI
  if (my_rank==0) {
#endif
    if (strcmp(argv[0], "-")!=0) {
      std::ofstream* out = new std::ofstream(argv[0]);
      if (!out->is_open()) {
	perror(argv[0]);
	return 1;
      }
      output = out;
    }
#ifdef HAVE_MPI
  }
#endif

  // initialize the model
  std::string init_model = param.get_string("initial-model");
  std::vector<double> x(LCostTable::n_params);
  std::fill(x.begin(), x.end(), 0.0);
  if (!init_model.empty()) {
    LCostTable cf;
    std::ifstream in(init_model.c_str());
    if (!in.is_open()) {
#ifdef HAVE_MPI
      if (my_rank==0) {
#endif
      perror(init_model.c_str());
#ifdef HAVE_MPI
      }
#endif
      return 1;
    }
    cf.restore(in);
    for (size_t i=0; i<cf.size(); ++i) x[i] = log(cf[i]);
  }

  // load example sequences
  std::list<ExampleS> train;
  Expectation count_feature;
  if (argc>=2) {		// load from files in args
    for (int i=1; i<argc; ++i) {
      std::ifstream in(argv[i]);
      if (!in.is_open()) {
#ifdef HAVE_MPI
	if (my_rank==0) {
#endif
	perror(argv[i]);
#ifdef HAVE_MPI
	}
#endif
	return 1;
      }
      ExampleS::load_examples(in, train, count_feature);
    }
  } else {			// load from stdin
    ExampleS::load_examples(std::cin, train, count_feature);
  }
  if (verbose>0)
    std::cout << train.size() << " data can be used." << std::endl;

  // execute training procedures
  {
    float eps = param.get_real("conv-rate");
    int max_it = param.get_int("max-iteration");
    float sigma = param.get_int("sigma");
    size_t it;
    float ll;
    struct timeval start_tm, finish_tm, calc_tm;
    char tm_buf[20];

    gettimeofday(&start_tm, NULL);
    crf_training(train, x, count_feature, max_it, eps, sigma, verbose,
		 it, ll, search_type);
    gettimeofday(&finish_tm, NULL);
    timersub(&finish_tm, &start_tm, &calc_tm);
    sprintf(tm_buf, "%ld.%03lds", calc_tm.tv_sec, calc_tm.tv_usec/1000);
    
#ifdef HAVE_MPI
    if (my_rank==0) {
#endif
      LCostTable res;
      for (size_t i=0; i<x.size(); ++i) res[i].set_log_value(x[i]);
      *output << "# CostTable for PHMMTS" << std::endl
	      << "# Trained by \"" << param.cmd_line() << "\"" << std::endl
	      << "#   method: CRF" << std::endl
	      << "#   time: " << tm_buf << std::endl
	      << "#   iteration: " << it << std::endl
	      << "#   loglikelihood: " << ll << std::endl;
      res.store(*output, true);
#ifdef HAVE_MPI
    }
#endif
    if (output!=&std::cout) delete output;
  }

  return 0;
}

int
run_ml_training(int argc, const char** argv, const Param& param)
{
  // verbose
  int verbose = 0;
  if (param.get_bool("verbose")) verbose=3;
  bool l_opt = param.get_bool("local");
#ifdef HAVE_MPI
  int my_rank = MPI::COMM_WORLD.Get_rank();
  if (my_rank!=0) verbose=0;
#endif

  // output
  std::ostream* output = &std::cout;
#ifdef HAVE_MPI
  if (my_rank==0) {
#endif
    if (strcmp(argv[0], "-")!=0) {
      std::ofstream* out = new std::ofstream(argv[0]);
      if (!out->is_open()) {
	perror(argv[0]);
	return 1;
      }
      output = out;
    }
#ifdef HAVE_MPI
  }
#endif

  // load example sequences
  std::list<ExampleS> train;
  Expectation count_feature;
  if (argc>=2) {		// load from files in args
    for (int i=1; i<argc; ++i) {
      std::ifstream in(argv[i]);
      if (!in.is_open()) {
	perror(argv[i]);
	return 1;
      }
      ExampleS::load_examples(in, train, count_feature, l_opt);
    }
  } else {			// load from stdin
    ExampleS::load_examples(std::cin, train, count_feature, l_opt);
  }
  if (verbose>0)
    std::cout << train.size() << " data can be used." << std::endl;

  // execute training procedures
  {
#ifdef HAVE_MPI
    if (my_rank==0) {
#endif
      count_feature.normalize();
      *output << "# CostTable for PHMMTS" << std::endl
	      << "# Trained by \"" << param.cmd_line() << "\"" << std::endl
	      << "#   method: ML" << std::endl;
      count_feature.store(*output);
#ifdef HAVE_MPI
    }
#endif
    if (output!=&std::cout) delete output;
  }

  return 0;
}

int
main(int argc, const char** argv)
{
#ifdef HAVE_MPI
  MPI::Init(argc, const_cast<char**&>(argv));
  int my_rank = MPI::COMM_WORLD.Get_rank();
#endif
  // parse arguments
  Param param(opts, "output-model [training-seq ...]");
  if (!param.parse(argc, argv)) return 1;
  if (argc<1) {
#ifdef HAVE_MPI
    if (my_rank==0) param.usage();
    MPI::Finalize();
#endif
    return 1;
  }
  int ret=0;

  if (param.get_string("algorithm")=="EM")
    ret=run_em_training(argc, argv, param);
  else if (param.get_string("algorithm")=="CRF")
    ret=run_crf_training(argc, argv, param);
  else if (param.get_string("algorithm")=="ML")
    ret=run_ml_training(argc, argv, param);
  else {
#ifdef HAVE_MPI
    if (my_rank==0) {
#endif
      std::cout << "The algorithm \""
		<< param.get_string("algorithm")
		<< "\" is not supported" << std::endl;
#ifdef HAVE_MPI
    }
#endif
    ret=1;
  }
#ifdef HAVE_MPI
  MPI::Finalize();
#endif
  return ret;
}
